/* ----------------------------------------------------------------------------

m A C 3 d e c

	by Luke Dennis  (based on A52Lib and AC3DEC by Aaron Holtzmann)

Basically a funky MacOS Carbon app which hooks into UNIX AC-3 code.  This code
uses some snippets and ideas from the following most excellent open-source apps:
Movie2MPEG by Kenji Gotoh, MacMPEG2Decoder by VLich, and SNES9X by John Stiles.

MP3 encoder used is LAME, for which information may be found at www.mp3dev.org.

Any suggestions or bugs, please email me at lukifer@mail.com.  See the included
Read Me.txt file for more information.  Enjoy!

Oh yeah, by using this app/code you accept all risks, I take no responsibility, 
you agree to act within the law, you agree to the terms in the Read Me file, 
yada yada yada.

MPAA: Go to Hell.  You go to hell and you die.

Dolby: Likewise.

(Okay, okay, Thomas Dolby kicks ass.  Too bad it doesn't extend to his lawyers.)

----------------------------------------------------------------------------- */

// Anything wacky gets set with this pragma:
#define TEST 0


/* TODO:

Use scCompressionListType to remove incompatible AIFF codecs
(http://developer.apple.com/qa/qa2001/qa1062.html)
Make QuickTime and AAC separate options


TOTEST:

QT VBR MP3 decoding audio length
Check normalization for clipping


TODO EVENTUALLY:

Frame listing/actual vbr bitrate feedback during MP3
VOB parsing
MP2/MP3/MPG input

*/


// Includes
// --------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <altivec.h>
#include <QuickTimeComponents.h>
#include <PLstringfuncs.h>

#include "a52.h"

//#include "debug.h"
#include "mAC3dec.h"
#include "apple_events.h"
#include "ac3fix.h"


// Globals
// -------
Decoder*		Mac3dec;
Preferences*	Prefs;
EventRecord		Event;
AEAddressDesc	OurAddress;
MenuHandle		FileMenu, DecodeMenu;
Boolean			Done;

extern Boolean vob;


// Main Function
// -------------

void main(void)
{
	Done = false;
	
	// Create app
	Mac3dec = new Decoder;
	
	// Event loop
	Mac3dec->EventLoop();
	
	// We out
	delete Mac3dec;
}


inline void IfError(Str255 theError, unsigned long theErrorNumber) 
{	if (theErrorNumber != noErr)
		HandleError(theError, theErrorNumber, false);
}
inline void IfErrorFatal(Str255 theError, unsigned long theErrorNumber) 
{	if (theErrorNumber != noErr)
		HandleError(theError, theErrorNumber, true);
}

void HandleError(Str255 theError, unsigned long theErrorNumber, Boolean fatal)
{
	Str255 errorNumberString;
	DialogPtr theErrorDialog;
	short theItem;

	NumToString((unsigned long) theErrorNumber, errorNumberString);
	ParamText(theError, errorNumberString, 0L, 0L);
	theErrorDialog = GetNewDialog(rDialogErrorBox, nil, (WindowPtr) - 1);
	
	// Make the OK button the default choice
	SetDialogDefaultItem(theErrorDialog, 1);
	
	do {ModalDialog(nil, &theItem);} while (!theItem);	
	DisposeDialog(theErrorDialog);

	if (fatal)
	{	delete Mac3dec;
		ExitToShell();  // Buh-Bye.
	}
}


void HandleErrorFourCC(Str255 theError, OSType fourCC, Boolean fatal)
{
	Str255 errorNumberString;
	DialogPtr theErrorDialog;
	short theItem;

	errorNumberString[0] = 4;
	BlockMoveData(&fourCC, &errorNumberString[1], sizeof(OSType));
	ParamText(theError, errorNumberString, 0L, 0L);
	theErrorDialog = GetNewDialog(rDialogErrorBox, nil, (WindowPtr) - 1);
	
	// Make the OK button the default choice
	SetDialogDefaultItem(theErrorDialog, 1);
	
	do {ModalDialog(nil, &theItem);} while (!theItem);	
	DisposeDialog(theErrorDialog);

	if (fatal)
	{	delete Mac3dec;
		ExitToShell();  // Buh-Bye.
	}
}


pascal OSErr HandleDragTrack(DragTrackingMessage message, WindowRef theWindow, void *handlerRefCon, DragRef theDrag)
{
	static Boolean hilited;
	RgnHandle theRgn;
	Rect tempRect = {0};
	
	switch (message)
	{	case kDragTrackingEnterWindow:
			SetPortWindowPort(Mac3dec->window);
			tempRect.right  = Mac3dec->winRect.right  - Mac3dec->winRect.left;
			tempRect.bottom = Mac3dec->winRect.bottom - Mac3dec->winRect.top;
			RectRgn(theRgn = NewRgn(), &tempRect);
			ShowDragHilite(theDrag, theRgn, true);
			DisposeRgn(theRgn);
			hilited = true;
			break;

		case kDragTrackingLeaveWindow:
			if (hilited)
			{	// remove hiliting
				HideDragHilite(theDrag);
			}
			break;
	}
	return noErr;
}


pascal OSErr HandleDragReceive(WindowRef theWindow, void *handlerRefCon, DragRef theDrag)
{
	unsigned short items, numFlavors;
	ItemReference theItem;
	AEDescList aeList;
	long numAEItems = 0;
	OSErr theErr;

	HideDragHilite(theDrag);

	// Create an AEDescList to hold the filespecs
	theErr = AECreateList(NULL, 0, false, &aeList);
		if (theErr != noErr) HandleError("\pAECreateList", theErr, true);

	// Copy all of the filespec records into our AEDescList
	CountDragItems(theDrag, &items);

	for (unsigned short i=1; i<=items; i++)
	{
		Size dataSize = sizeof(HFSFlavor);
		FlavorType flavor;
		HFSFlavor fileData;
		unsigned long dataOffset = 0;
		Boolean notFile = false;

		theErr = GetDragItemReferenceNumber(theDrag, i, &theItem);
			if (theErr != noErr) HandleError("\pGetDragItemReferenceNumber", theErr, true);

		// Check whether drag contains files
		theErr = CountDragItemFlavors(theDrag, theItem, &numFlavors);
			if (theErr != noErr) HandleError("\pCountDragItemFlavors", theErr, true);
		for (unsigned short j = 1; j <= numFlavors; j++)
		{	theErr = GetFlavorType(theDrag, theItem, j, &flavor);
				if (theErr != noErr) HandleError("\pCountDragItemFlavors", theErr, true);
			if (flavor == flavorTypeHFS) break;
			notFile = true;
		} if (notFile) continue;

		GetFlavorData(theDrag, theItem, flavorTypeHFS, &fileData, &dataSize, dataOffset);
			if (theErr != noErr) HandleError("\pGetFlavorData", theErr, true);
		// Add to AEDescList
		theErr = AEPutPtr(&aeList, ++numAEItems, typeFSS, &fileData.fileSpec, sizeof(FSSpec));
			if (theErr != noErr) HandleError("\pAEPutPtr", theErr, true);
	}

	// Create an apple event, add the AEDescList of filespecs and send
	AppleEvent theEvent;
	AppleEvent aereply;

	theErr = AECreateAppleEvent(kCoreEventClass, kAEOpenDocuments, &OurAddress, kAutoGenerateReturnID, kAnyTransactionID, &theEvent);
		if (theErr != noErr) HandleError("\pAECreateAppleEvent", theErr, true);
	theErr = AEPutKeyDesc(&theEvent, keyDirectObject, &aeList);
		if (theErr != noErr) HandleError("\pAEPutParamDesc", theErr, true);
	theErr = AESend(&theEvent, &aereply, kAENoReply | kAECanInteract | kAEDontReconnect, kAENormalPriority, kAEDefaultTimeout, NULL, NULL);
		if (theErr != noErr) HandleError("\pAESend", theErr, true);
	return noErr;
}


pascal void NavServicesSuxEventProc(NavEventCallbackMessage callBackSelector, NavCBRecPtr callBackParms, NavCallBackUserData callBackUD)
{
	EventRecord *theEvent = callBackParms->eventData.eventDataParms.event;
	
	switch(callBackSelector)
	{	case kNavCBEvent:
			// I ain't gonna do a goddamn thing here
			break;
	}
}


static // Stolen from MoreFiles... aren't I a stinker?
OSErr FSpGetFullPath(const FSSpec *spec, StringPtr path)
{
	OSErr		result = noErr;
	OSErr		realResult = noErr;
	Str255		str;
	FSSpec		tempSpec;
	CInfoPBRec	pb;
	
	/* work around Nav Services "bug" (it returns invalid FSSpecs with empty names) */
	if (spec->name[0] == 0)
		result = FSMakeFSSpec(spec->vRefNum, spec->parID, spec->name, &tempSpec);
	else	/* Make a copy of the input FSSpec that can be modified */
		BlockMoveData(spec, &tempSpec, sizeof(FSSpec));
	
	if (result == noErr)
	{	if (tempSpec.parID == fsRtParID)
		{	/* The object is a volume */
			
			/* Add a colon to make it a full pathname */
			++tempSpec.name[0];
			tempSpec.name[tempSpec.name[0]] = ':';
			
			/* We're done */
			PLstrcpy(path, tempSpec.name);
		}
		else
		{	/* The object isn't a volume */
			
			/* Is the object a file or a directory? */
			pb.dirInfo.ioNamePtr = tempSpec.name;
			pb.dirInfo.ioVRefNum = tempSpec.vRefNum;
			pb.dirInfo.ioDrDirID = tempSpec.parID;
			pb.dirInfo.ioFDirIndex = 0;
			result = PBGetCatInfoSync(&pb);
			// Allow file/directory name at end of path to not exist.
			realResult = result;
			if ( (result == noErr) || (result == fnfErr) )
			{	/* if the object is a directory, append a colon so full pathname ends with colon */
				if ( (result == noErr) && (pb.hFileInfo.ioFlAttrib & kioFlAttribDirMask) != 0 )
				{	++tempSpec.name[0];
					tempSpec.name[tempSpec.name[0]] = ':';
				}
				
				/* Put the object name in first */
				PLstrcpy(path, tempSpec.name);
				
				/* Get the ancestor directory names */
				pb.dirInfo.ioNamePtr = tempSpec.name;
				pb.dirInfo.ioVRefNum = tempSpec.vRefNum;
				pb.dirInfo.ioDrParID = tempSpec.parID;
				do	/* loop until we have an error or find the root directory */
				{	pb.dirInfo.ioFDirIndex = -1;
					pb.dirInfo.ioDrDirID = pb.dirInfo.ioDrParID;
					result = PBGetCatInfoSync(&pb);
					if ( result == noErr )
					{	/* Append colon to directory name */
						++tempSpec.name[0];
						tempSpec.name[tempSpec.name[0]] = ':';
						
						/* Add directory name to beginning of fullPath */
						PLstrcpy(str,  tempSpec.name);
						PLstrcat(str,  path);
						PLstrcpy(path, str);
					}
				} while ((result == noErr) && (pb.dirInfo.ioDrDirID != fsRtDirID));
	}	}	}
	
	if (result == noErr)
	{	/* Return the length */
		result = realResult;	// return realResult in case it was fnfErr
	} else
	{	/* Dispose of the handle and return NULL and zero length */
		path[0] = 0;
	}
	
	// Terminate string
	path[path[0]+1] = 0;
	
	return result;
}


/* Only used for speed testing...

static void TicksToPercent(unsigned long ticks, unsigned long total, StringPtr str)
{
	Str255	tempStr;
	float	percent;
	long	val;
	
	percent = ((float) ticks) / ((float) total) * 100.0f;
	
	val = (long) percent;
	if (val < 10)	PLstrcpy(str, "\p ");
	else			PLstrcpy(str, "\p");
	NumToString(val, tempStr);
	PLstrcat(str, tempStr);
	
	percent *= 10.0f;
	val = ((long) percent) - (val * 10);
	NumToString(val, tempStr);
	PLstrcat(str, "\p.");
	PLstrcat(str, tempStr);
	PLstrcat(str, "\p %");
} */


static void TicksToTime(unsigned long ticks, StringPtr str, Boolean bTicks)
{
	unsigned long tempTicks, seconds, minutes, hours;
	Str255 tempStr;

	tempTicks = ticks;
	PLstrcpy(str, "\p");
	
	// Hours
	hours = tempTicks / (60 * 60 * 60);
	tempTicks -= (hours * 60 * 60 * 60);
	
	// Minutes
	minutes = tempTicks / (60 * 60);
	tempTicks -= (minutes * 60 * 60);
	
	// Seconds
	seconds = tempTicks / 60;

	// Write Hours
	if (hours > 0)
	{	NumToString(hours, str);
		if (minutes < 10)	PLstrcat(str, "\p:0");
		else				PLstrcat(str, "\p:");
	}

	// Write minutes
	NumToString(minutes, tempStr);
	PLstrcat(str, tempStr);
	if (seconds < 10)	PLstrcat(str, "\p:0");
	else				PLstrcat(str, "\p:");

	// Write seconds
	NumToString(seconds, tempStr);
	PLstrcat(str, tempStr);
	
	// Write ticks
	if (bTicks)
	{	tempTicks = ticks % 60;
		if (tempTicks < 10)	PLstrcat(str, "\p.0");
		else				PLstrcat(str, "\p.");
		NumToString(tempTicks, tempStr);
		PLstrcat(str, tempStr);
	}
}


// These seem to do nothing...
/*static void LameErrorProc(const char* format, va_list ap)
{	HandleError("\pLAME choked.  How very LAME.", noErr, true); }

static void LameDebugProc(const char* format, va_list ap)
{	HandleError("\pLAME must speak with the code-monkey.", noErr, true); }

static void LameMsgProc(const char* format, va_list ap)
{	HandleError("\pLAME's trying to tell us something, but buggered if I know what it is.", noErr, false); }*/



// Callback Function - A52Lib

static pascal Boolean AC3FillBufferProc(SoundComponentDataPtr *outData, void *refCon)
{
	Decoder* dec = (Decoder*) refCon;

	// a52lib
	int				length = 0;
	sample_t		level = 32767.0f;
	unsigned char*	start;
	float*			fltPtr = a52_samples(dec->a52_state);

	// Seek to sync word and parse sync info
	length = dec->SeekSync(&dec->flags, &dec->sample_rate, &dec->bit_rate);

	// Decode frame
	if (Prefs->a52dolby) dec->flags = A52_DOLBY;
	else				 dec->flags = A52_STEREO;
	dec->flags |= A52_ADJUST_LEVEL;
	dec->FillBuffer(&start, length); // Meesa like code re-use!
	if (dec->state <= kStateIdle) // We're done
		dec->moreSource = false;
	a52_frame(dec->a52_state, start, &dec->flags, &level, 0);

	// Disable Dynamic Range Compression
	if (Prefs->a52nodynrng)
		a52_dynrng(dec->a52_state, NULL, NULL);

	// Get decoded samples
	short*	shortPtr = (short*) dec->ac3CompData.desc.buffer;

	// ... and normalize
	if (Prefs->normalize && !dec->normPass)
	{
		float ng = dec->normGrow;
		
		for (int i=0; i<6; i++)
		{
			// Get sample data
			a52_block(dec->a52_state);
			
			// Get channels and interleave
			for (int j=0; j<256; j++)
			{
				// Left Channel
				//shortPtr[(i*256+j)*2]   = (short) (fltPtr[j]);
				*shortPtr++ = (short) (fltPtr[j] * ng);
	
				// Right channel
				//shortPtr[(i*256+j)*2+1] = (short) (fltPtr[j+256]);
				*shortPtr++ = (short) (fltPtr[j+256] * ng);
		}	}
	}
	
	// ... and don't normalize
	else
	{	for (int i=0; i<6; i++)
		{
			// Get sample data
			a52_block(dec->a52_state);
			
			// Get channels and interleave
			for (int j=0; j<256; j++)
			{
				// Left Channel
				//shortPtr[(i*256+j)*2]   = (short) (fltPtr[j]);
				*shortPtr++ = (short) (fltPtr[j]);
	
				// Right channel
				//shortPtr[(i*256+j)*2+1] = (short) (fltPtr[j+256]);
				*shortPtr++ = (short) (fltPtr[j+256]);
			}
	}	}

	dec->ac3CompData.bufferSize = 6144;
	*outData = (SoundComponentDataPtr) &dec->ac3CompData;

	return (dec->moreSource);
}


// Callback Function - QuickTime

static pascal Boolean QTInFillBufferProc(SoundComponentDataPtr *outData, void *refCon)
{
	OSErr theErr;
	Decoder* dec = (Decoder*) refCon;

	// QuickTime decoding
	long sourceBytesReturned, numberOfSamples;
	TimeValue	sourceReturnedTime, durationPerSample;
	
	// if after getting the last chunk of data the total time is over the duration, we're done
	if (dec->time >= dec->duration)
		dec->moreSource = false;

	if (dec->moreSource)
	{
		HUnlock(dec->sourceHandle);

		theErr = GetMediaSample(dec->media,					// specifies the media for this operation
							 	dec->sourceHandle,			// function returns the sample data into this handle
							 	dec->info.outputBytes,		// maximum number of bytes of sample data to be returned
							 	&sourceBytesReturned,		// the number of bytes of sample data returned
							 	dec->time,					// starting time of the sample to be retrieved (must be in Media's TimeScale)
							 	&sourceReturnedTime,		// indicates the actual time of the returned sample data
							 	&durationPerSample,			// duration of each sample in the media
							 	NULL,						// sample description corresponding to the returned sample data (NULL to ignore)
							 	NULL,						// index value to the sample description that corresponds to the returned sample data (NULL to ignore)
							 	0L,							// maximum number of samples to be returned (0 to use a value that is appropriate for the media)
							 	&numberOfSamples,			// number of samples it actually returned
							 	NULL);						// flags that describe the sample (NULL to ignore)
		
		HLock(dec->sourceHandle);

		if ((theErr != noErr) || (sourceBytesReturned == 0))
		{	dec->moreSource = false;
			if ((theErr != noErr) && (sourceBytesReturned > 0))
				HandleError("\pGetMediaSample - Failed in FillBufferDataProc", theErr, true);
		}
		
		dec->time = sourceReturnedTime + durationPerSample * numberOfSamples;
		dec->qtInCompData.bufferSize = sourceBytesReturned; 		
	}

	*outData = (SoundComponentDataPtr) &dec->qtInCompData;
	
	return (dec->moreSource);
}


// Callback function - Normalization for QuickTime

static pascal Boolean QTOutFillBufferProc(SoundComponentDataPtr *outData, void *refCon)
{
	OSErr theErr;
	Decoder* dec = (Decoder*) refCon;
	unsigned long outputFrames, actualOutputBytes, outputFlags;
	short* shortPtr = (short*) dec->decomBuffer;
	Boolean moreStuff = true;

	if (dec->moreSource)
	{	actualOutputBytes = 0;
		theErr = SoundConverterFillBuffer(	dec->qtInfo.outputSC,	// a sound converter
											dec->qtFillBufferUPP,	// the callback UPP
											refCon,					// refCon passed to FillDataProc
											dec->qtOutBuffer,		// the destination data buffer
											dec->qtInfo.outputBytes,// size of the destination buffer
											&actualOutputBytes,		// number of output bytes
											&outputFrames,			// number of output frames
											&outputFlags);			// FillBuffer retured advisor flags
			if (theErr != noErr)
				HandleError("\pAnd I never saw Venice.", theErr, true);

// FIXME: Possible premature EOF
	/*	if ((outputFlags & kSoundConverterHasLeftOverData == false) &&
			(actualOutputBytes <= 0))	moreStuff = false;*/
	//	else							dec->moreSource = true; 
	}
	else
	{	theErr = SoundConverterEndConversion(	dec->qtInfo.outputSC,
												dec->qtOutBuffer,
												&outputFrames,
												&actualOutputBytes);
		moreStuff = false;		
	}

	// Normalize if needed
	if (Prefs->normalize)
	{	float ng = dec->normGrow;
		for (int i=0; i<(actualOutputBytes/2); i++)
			shortPtr[i] = shortPtr[i] * ng;
	}

	dec->qtOutCompData.bufferSize = actualOutputBytes; 		
	*outData = (SoundComponentDataPtr) &dec->qtOutCompData;
	
	return (moreStuff);
}


// Quick and dirty URL launcher so that people can give me money...
static void Web(Str255 urlStr)
{
	OSErr		theErr = noErr;
    ICInstance	inst;
    long		startSel;
    long		endSel;

    theErr = ICStart(&inst, 'mAC3');
    
    if (theErr == noErr)
    {	startSel = 0;
		endSel = urlStr[0];
		theErr = ICLaunchURL(inst, "\p", (char *) &urlStr[1], urlStr[0], &startSel, &endSel);
	}
		
	theErr = ICStop(inst);
}



#pragma mark -


// Preferences
// -----------

Preferences::Preferences(void)
{
	// Set defaults
	channelSplit	= false;
	audioOutput		= kAudio44khz;
	wrapper			= kWrapperAIFF;
	chooseSave		= false;
	hasAltivec		= false;
	useAltivec		= false;
	osX				= false;
	autoDecode		= false;
	normalize		= false;
	normalizeGlobal = false;
	a52nodynrng		= false;
	useProgressBar	= false;
	BlockMoveData(&kRectWindowSize, &winSize, sizeof(Rect));
	
	mp3info.mono		= false;
	mp3info.joint		= true;
	mp3info.fast		= false;
	mp3info.high		= true;
	mp3info.vbr			= false;
	mp3info.bitrate		= 128;
	mp3info.minBitrate	= 80;
	mp3info.maxBitrate	= 320;
	mp3info.sample		= 44.100f;
	mp3info.vbrQuality	= 4;
	mp3info.nspsytune	= false;
	
	// Get us some memory for AIFF and MOV settings
	customHandle = 0L;
	movHandle = 0L;
	
	// Read preferences
	Read();
}


Preferences::~Preferences(void)
{
	// Write preferences before closing
	Write();
	
	// Cleanup memory
	DisposeHandle(customHandle);
	DisposeHandle(movHandle);
	DisposeHandle(compressionParams);
}


void Preferences::Dialog(void)
{
	DialogPtr	thePrefsDialog;
	short		theItem, nullItem;
	Handle		nullHandle;
	ControlHandle	controlDestHdl, controlAutoDecodeHdl, controlAltivecHdl;
	ControlHandle	controlDynRangeHdl, controlDolbyHdl, controlLameSampleHdl, controlProgressBar;
	RGBColor theColor = {0x0000, 0x0000, 0x0000};
	Rect tempRect;

	// Create dialog window
	thePrefsDialog = GetNewDialog(rDialogPrefs, nil, (WindowPtr) - 1);
		if (thePrefsDialog == nil) return;
	SetDialogDefaultItem(thePrefsDialog, rDialogPrefsOK);
	SetPortDialogPort(thePrefsDialog);

	// Get Handles to the controls
	GetDialogItemAsControl(thePrefsDialog, rDialogPrefsDest,		&controlDestHdl);
	GetDialogItemAsControl(thePrefsDialog, rDialogPrefsAutoDecode,	&controlAutoDecodeHdl);
	GetDialogItemAsControl(thePrefsDialog, rDialogPrefsUseAltivec,	&controlAltivecHdl);
	GetDialogItemAsControl(thePrefsDialog, rDialogPrefsDynRange,	&controlDynRangeHdl);
	GetDialogItemAsControl(thePrefsDialog, rDialogPrefsDolby,		&controlDolbyHdl);
	GetDialogItemAsControl(thePrefsDialog, rDialogPrefsLameSample,	&controlLameSampleHdl);
	GetDialogItemAsControl(thePrefsDialog, rDialogPrefsNoProgress,	&controlProgressBar);
	
	// For now, Altivec OFF
	SetControlValue(controlAltivecHdl,	false);
	DeactivateControl(controlAltivecHdl);

	// Set initial Preferences values
	SetControlValue(controlDestHdl,			!chooseSave);
	SetControlValue(controlAutoDecodeHdl,	autoDecode);
	SetControlValue(controlDynRangeHdl,		!a52nodynrng);
	SetControlValue(controlDolbyHdl,		a52dolby);
	SetControlValue(controlLameSampleHdl,	lameResample);	
	SetControlValue(controlProgressBar,		!useProgressBar);	

	// Show if Altivec is active or not
	GetDialogItem(thePrefsDialog, rDialogPrefsAltivecLbl, &nullItem, &nullHandle, &tempRect);
	if (osX)	MoveTo(tempRect.left + 82, tempRect.top + 13);
	else		MoveTo(tempRect.left + 82, tempRect.top + 12);
	if (hasAltivec)
	{	theColor.green = 0xCCCC;
		RGBForeColor(&theColor);
		TextSize(11);
		DrawString("\pPresent");
		TextSize(12);
		theColor.green = 0x0000;
		SetControlValue(controlAltivecHdl, Prefs->useAltivec);
	} else
	{	theColor.red = 0xFFFF;
		RGBForeColor(&theColor);
		TextSize(11);
		DrawString("\pNot Present");
		TextSize(12);
		theColor.red = 0x0000;
		DeactivateControl(controlAltivecHdl);
	} RGBForeColor(&theColor);


	// Dialog go laid!
	ShowWindow(GetDialogWindow(thePrefsDialog));

	do
	{	ModalDialog(nil, &theItem);
		
		switch (theItem)
		{
			case rDialogPrefsDest:
				SetControlValue(controlDestHdl, !(GetControlValue(controlDestHdl)));
				break;
			
			case rDialogPrefsAutoDecode:
				SetControlValue(controlAutoDecodeHdl, !(GetControlValue(controlAutoDecodeHdl)));
				break;

			case rDialogPrefsUseAltivec:
				//SetControlValue(controlAltivecHdl, !(GetControlValue(controlAltivecHdl)));
				break;

			case rDialogPrefsDynRange:
				SetControlValue(controlDynRangeHdl, !(GetControlValue(controlDynRangeHdl)));
				break;

			case rDialogPrefsDolby:
				SetControlValue(controlDolbyHdl, !(GetControlValue(controlDolbyHdl)));
				break;

			case rDialogPrefsLameSample:
				SetControlValue(controlLameSampleHdl, !(GetControlValue(controlLameSampleHdl)));
				break;

			case rDialogPrefsNoProgress:
				SetControlValue(controlProgressBar, !(GetControlValue(controlProgressBar)));
				break;
		}
	} while(theItem != rDialogPrefsOK);

	// Retrieve the settings
	autoDecode		=	GetControlValue(controlAutoDecodeHdl);
	chooseSave		= !(GetControlValue(controlDestHdl));
	a52nodynrng		= !(GetControlValue(controlDynRangeHdl));
	a52dolby		=	GetControlValue(controlDolbyHdl);
	lameResample	=	GetControlValue(controlLameSampleHdl);
	useProgressBar	= !(GetControlValue(controlProgressBar));

	if (hasAltivec)	useAltivec = GetControlValue(controlAltivecHdl);
	else			useAltivec = false;

	
	// Write Prefs to file
	Write();
	
	// Destroy Dialog
	DisposeDialog(thePrefsDialog);
	SetPortWindowPort(Mac3dec->window);
}


void Preferences::MP3Dialog(void)
{
	Str31			str;
	DialogPtr		dialog;
	short			theItem;
	ControlHandle	cNspsytune, cChannels, cQuality, cSamplingRate, cMethod, cBitrate, cBitrateEdit, 
							cVBRQuality, cMinBitrate, cMaxBitrate;
	Handle hdl; Rect rect; short item;

	// Create dialog window
	dialog = GetNewDialog(rDialogMP3Settings, nil, (WindowPtr)-1);
		if (dialog == nil) return;
	SetDialogDefaultItem(dialog, rDialogMP3OK);
	SetPortDialogPort(dialog);

	// Get control handles
	GetDialogItemAsControl(dialog, rDialogMP3nspsytune,		&cNspsytune);
	GetDialogItemAsControl(dialog, rDialogMP3Channels,		&cChannels);
	GetDialogItemAsControl(dialog, rDialogMP3Quality,		&cQuality);
	GetDialogItemAsControl(dialog, rDialogMP3SamplingRate,	&cSamplingRate);
	GetDialogItemAsControl(dialog, rDialogMP3Method,		&cMethod);
	GetDialogItemAsControl(dialog, rDialogMP3Bitrate,		&cBitrate);
	GetDialogItemAsControl(dialog, rDialogMP3BitrateEdit,	&cBitrateEdit);
	GetDialogItemAsControl(dialog, rDialogMP3VBRQuality,	&cVBRQuality);
	GetDialogItemAsControl(dialog, rDialogMP3MinBitrate,	&cMinBitrate);
	GetDialogItemAsControl(dialog, rDialogMP3MaxBitrate,	&cMaxBitrate);

	// Set initial values
	if (mp3info.nspsytune)	SetControlValue(cNspsytune, 1);

		 if (mp3info.mono)	SetControlValue(cChannels, 1);
	else if (mp3info.joint)	SetControlValue(cChannels, 3);
	else					SetControlValue(cChannels, 2);
	
		 if (mp3info.fast)	SetControlValue(cQuality, 3);
	else if (mp3info.high)	SetControlValue(cQuality, 1);
	else					SetControlValue(cQuality, 2);
	
		 if (mp3info.sample == 16000) SetControlValue(cSamplingRate, 1);
	else if (mp3info.sample == 22050) SetControlValue(cSamplingRate, 2);
	else if (mp3info.sample == 24000) SetControlValue(cSamplingRate, 3);
	else if (mp3info.sample == 32000) SetControlValue(cSamplingRate, 4);
	else if (mp3info.sample == 44100) SetControlValue(cSamplingRate, 5);
	else if (mp3info.sample == 48000) SetControlValue(cSamplingRate, 6);
	else							  SetControlValue(cSamplingRate, 5);
	
	if (mp3info.vbrQuality > 9 || mp3info.vbrQuality < 0) mp3info.vbrQuality = 4;
	SetControlValue(cVBRQuality, mp3info.vbrQuality + 1);

	switch (mp3info.bitrate)
	{	case 8:		SetControlValue(cBitrate, 1);	break;
		case 16:	SetControlValue(cBitrate, 2);	break;
		case 24:	SetControlValue(cBitrate, 3);	break;
		case 32:	SetControlValue(cBitrate, 4);	break;
		case 40:	SetControlValue(cBitrate, 5);	break;
		case 48:	SetControlValue(cBitrate, 6);	break;
		case 56:	SetControlValue(cBitrate, 7);	break;
		case 64:	SetControlValue(cBitrate, 8);	break;
		case 80:	SetControlValue(cBitrate, 9);	break;
		case 96:	SetControlValue(cBitrate, 10);	break;
		case 112:	SetControlValue(cBitrate, 11);	break;
		case 128:	SetControlValue(cBitrate, 12);	break;
		case 160:	SetControlValue(cBitrate, 13);	break;
		case 192:	SetControlValue(cBitrate, 14);	break;
		case 224:	SetControlValue(cBitrate, 15);	break;
		case 256:	SetControlValue(cBitrate, 16);	break;
		case 320:	SetControlValue(cBitrate, 17);	break;
		
		default:	SetControlValue(cBitrate, 12);	break;
	}

	switch (mp3info.minBitrate)
	{	case 8:		SetControlValue(cMinBitrate, 1);	break;
		case 16:	SetControlValue(cMinBitrate, 2);	break;
		case 24:	SetControlValue(cMinBitrate, 3);	break;
		case 32:	SetControlValue(cMinBitrate, 4);	break;
		case 40:	SetControlValue(cMinBitrate, 5);	break;
		case 48:	SetControlValue(cMinBitrate, 6);	break;
		case 56:	SetControlValue(cMinBitrate, 7);	break;
		case 64:	SetControlValue(cMinBitrate, 8);	break;
		case 80:	SetControlValue(cMinBitrate, 9);	break;
		case 96:	SetControlValue(cMinBitrate, 10);	break;
		case 112:	SetControlValue(cMinBitrate, 11);	break;
		case 128:	SetControlValue(cMinBitrate, 12);	break;
		case 160:	SetControlValue(cMinBitrate, 13);	break;
		case 192:	SetControlValue(cMinBitrate, 14);	break;
		case 224:	SetControlValue(cMinBitrate, 15);	break;
		case 256:	SetControlValue(cMinBitrate, 16);	break;
		case 320:	SetControlValue(cMinBitrate, 17);	break;
		
		default:	SetControlValue(cMinBitrate, 10);	break;
	}

	switch (mp3info.maxBitrate)
	{	case 8:		SetControlValue(cMaxBitrate, 1);	break;
		case 16:	SetControlValue(cMaxBitrate, 2);	break;
		case 24:	SetControlValue(cMaxBitrate, 3);	break;
		case 32:	SetControlValue(cMaxBitrate, 4);	break;
		case 40:	SetControlValue(cMaxBitrate, 5);	break;
		case 48:	SetControlValue(cMaxBitrate, 6);	break;
		case 56:	SetControlValue(cMaxBitrate, 7);	break;
		case 64:	SetControlValue(cMaxBitrate, 8);	break;
		case 80:	SetControlValue(cMaxBitrate, 9);	break;
		case 96:	SetControlValue(cMaxBitrate, 10);	break;
		case 112:	SetControlValue(cMaxBitrate, 11);	break;
		case 128:	SetControlValue(cMaxBitrate, 12);	break;
		case 160:	SetControlValue(cMaxBitrate, 13);	break;
		case 192:	SetControlValue(cMaxBitrate, 14);	break;
		case 224:	SetControlValue(cMaxBitrate, 15);	break;
		case 256:	SetControlValue(cMaxBitrate, 16);	break;
		case 320:	SetControlValue(cMaxBitrate, 17);	break;
		
		default:	SetControlValue(cMaxBitrate, 16);	break;
	}

		 if (mp3info.vbr)	SetControlValue(cMethod, 3);
	else if (mp3info.abr)	SetControlValue(cMethod, 2);
	else					SetControlValue(cMethod, 1);

	GetDialogItem(dialog, rDialogMP3BitrateEdit, &item, &hdl, &rect);
	NumToString(mp3info.bitrate, str);
	SetDialogItemText(hdl, str);

	// Dialog Go Laid!
	ShowWindow(GetDialogWindow(dialog));
	theItem = rDialogMP3Method;

	do
	{	switch (theItem)
		{
			case rDialogMP3nspsytune:
				SetControlValue(cNspsytune, !GetControlValue(cNspsytune));
			case rDialogMP3Channels:
				break;			
			case rDialogMP3Quality:
				break;
			case rDialogMP3SamplingRate:
				break;
			case rDialogMP3Method:
				HideDialogItem(dialog, 22);
				switch (GetControlValue(cMethod))
				{	default:
					case 1: // CBR
						ShowDialogItem(dialog, rDialogMP3BitrateLabel);
						ShowDialogItem(dialog, rDialogMP3Bitrate);
						HideDialogItem(dialog, rDialogMP3BitrateEdit);
						HideDialogItem(dialog, rDialogMP3VBRQuality);
						HideDialogItem(dialog, rDialogMP3VBRQLabel);
						HideDialogItem(dialog, rDialogMP3MinLabel);
						HideDialogItem(dialog, rDialogMP3MaxLabel);
						HideDialogItem(dialog, rDialogMP3MinBitrate);
						HideDialogItem(dialog, rDialogMP3MaxBitrate);
						break;
					case 2: // ABR
						ShowDialogItem(dialog, rDialogMP3BitrateLabel);
						HideDialogItem(dialog, rDialogMP3Bitrate);
						ShowDialogItem(dialog, rDialogMP3BitrateEdit);
						HideDialogItem(dialog, rDialogMP3VBRQuality);
						HideDialogItem(dialog, rDialogMP3VBRQLabel);
						ShowDialogItem(dialog, rDialogMP3MinLabel);
						HideDialogItem(dialog, rDialogMP3MaxLabel);
						ShowDialogItem(dialog, rDialogMP3MinBitrate);
						HideDialogItem(dialog, rDialogMP3MaxBitrate);
						break;
					case 3: // VBR
						HideDialogItem(dialog, rDialogMP3BitrateLabel);
						HideDialogItem(dialog, rDialogMP3Bitrate);
						HideDialogItem(dialog, rDialogMP3BitrateEdit);
						ShowDialogItem(dialog, rDialogMP3VBRQuality);
						ShowDialogItem(dialog, rDialogMP3VBRQLabel);
						ShowDialogItem(dialog, rDialogMP3MinLabel);
						ShowDialogItem(dialog, rDialogMP3MaxLabel);
						ShowDialogItem(dialog, rDialogMP3MinBitrate);
						ShowDialogItem(dialog, rDialogMP3MaxBitrate);
						break;
				}
				ShowDialogItem(dialog, 22);					
				break;
			case rDialogMP3Bitrate:
				break;
			case rDialogMP3VBRQuality:
				break;
			case rDialogMP3MinBitrate:
				if (GetControlValue(cMaxBitrate) < GetControlValue(cMinBitrate))
					SetControlValue(cMaxBitrate, GetControlValue(cMinBitrate));
				break;
			case rDialogMP3MaxBitrate:
				if (GetControlValue(cMaxBitrate) < GetControlValue(cMinBitrate))
					SetControlValue(cMinBitrate, GetControlValue(cMaxBitrate));
				break;
		}
	
		ModalDialog(nil, &theItem);
	} while (theItem != rDialogMP3OK);

	// Get settings from controls
	if (GetControlValue(cNspsytune))	mp3info.nspsytune = true;
	else								mp3info.nspsytune = false;
	
	switch (GetControlValue(cChannels))
	{	case 1: mp3info.mono  = true;
				mp3info.joint = false;	break;
		case 2:	mp3info.mono  = false;
				mp3info.joint = false;	break;
		case 3: mp3info.mono  = false;
				mp3info.joint = true;	break;
	}

	switch (GetControlValue(cQuality))
	{	case 1: mp3info.high  = true;
				mp3info.fast  = false;	break;
		case 2:	mp3info.high  = false;
				mp3info.fast  = false;	break;
		case 3: mp3info.high  = false;
				mp3info.fast  = true;	break;
	}
	
	switch (GetControlValue(cSamplingRate))
	{	case 1:	mp3info.sample = 16000; break;
		case 2: mp3info.sample = 22050; break;
		case 3: mp3info.sample = 24000; break;
		case 4: mp3info.sample = 32000; break;
		case 5: mp3info.sample = 44100; break;
		case 6: mp3info.sample = 48000; break;
	}
	
	mp3info.abr			= (GetControlValue(cMethod) == 2);
	mp3info.vbr			= (GetControlValue(cMethod) == 3);
	mp3info.vbrQuality	= (GetControlValue(cVBRQuality) - 1);

	if (mp3info.abr)
	{	long temp;
		GetDialogItem(dialog, rDialogMP3BitrateEdit, &item, &hdl, &rect);
		GetDialogItemText(hdl, str);
		if (str[0] > 0)
		{	StringToNum(str, &temp);
			mp3info.bitrate = (int) temp;
			if (mp3info.bitrate > 320 || mp3info.bitrate < 8) mp3info.bitrate = 128;
		} else mp3info.bitrate = 128;
	}
	else
	{	switch (GetControlValue(cBitrate))
		{	case 1: case 2: case 3: case 4: case 5: case 6: case 7: case 8: 
				mp3info.bitrate = GetControlValue(cBitrate) * 8;	break;
			case 9: mp3info.bitrate = 80;	break;
			case 10:mp3info.bitrate = 96;	break;
			case 11:mp3info.bitrate = 112;	break;
			case 12:mp3info.bitrate = 128;	break;
			case 13:mp3info.bitrate = 160;	break;
			case 14:mp3info.bitrate = 192;	break;
			case 15:mp3info.bitrate = 224;	break;
			case 16:mp3info.bitrate = 256;	break;
			case 17:mp3info.bitrate = 320;	break;
	}	}

	switch (GetControlValue(cMinBitrate))
	{	case 1: case 2: case 3: case 4: case 5: case 6: case 7: case 8:
			mp3info.minBitrate = GetControlValue(cMinBitrate) * 8;	break;
		case 9: mp3info.minBitrate = 80;	break;
		case 10:mp3info.minBitrate = 96;	break;
		case 11:mp3info.minBitrate = 112;	break;
		case 12:mp3info.minBitrate = 128;	break;
		case 13:mp3info.minBitrate = 160;	break;
		case 14:mp3info.minBitrate = 192;	break;
		case 15:mp3info.minBitrate = 224;	break;
		case 16:mp3info.minBitrate = 256;	break;
		case 17:mp3info.minBitrate = 320;	break;
	}

	switch (GetControlValue(cMaxBitrate))
	{	case 1: case 2: case 3: case 4: case 5: case 6: case 7: case 8:
			mp3info.maxBitrate = GetControlValue(cMaxBitrate) * 8;	break;
		case 9: mp3info.maxBitrate = 80;	break;
		case 10:mp3info.maxBitrate = 96;	break;
		case 11:mp3info.maxBitrate = 112;	break;
		case 12:mp3info.maxBitrate = 128;	break;
		case 13:mp3info.maxBitrate = 160;	break;
		case 14:mp3info.maxBitrate = 192;	break;
		case 15:mp3info.maxBitrate = 224;	break;
		case 16:mp3info.maxBitrate = 256;	break;
		case 17:mp3info.maxBitrate = 320;	break;
	}
	

	// Write Prefs to file
	Write();
	
	// Destroy Dialog
	DisposeDialog(dialog);
	SetPortWindowPort(Mac3dec->window);
}


OSErr Preferences::Write(void)
{
	OSErr theErr;
	SInt16 volRefNum;
	long directoryID, bytesWritten;
	FSSpec fileSSpec;
	SInt16 fileRefNum;
	Str255 writeBuffer;
	
	/* Preferences Layout:
		1st Char:		Prefs Version (kPrefsVersion)
		2nd Char:		audioOutput ('0', '1', '2')
		3rd Char:		wrapper ('0', '1', '2')
		4th Char: 		channelSplit ('0', '1')
		5th Char:		normalize ('0', '1')
		6th Char:		chooseSave ('0', '1')
		7th Char:		Disable Progress Bar ('3', '4')
		8th Char:		autoDecode ('0', '1')
		9th Char:		useAltivec ('0', '1')
		10th Char:		MP3 mono ('0', '1')
		11th Char:		MP3 joint stereo ('0', '1')
		12th Char:		MP3 fast mode ('0', '1')
		13th Char:		MP3 high quality ('0', '1')
		14th Char:		MP3 Sampling Rate ('0' - '5')
		15th Char:		MP3 VBR ('0', '1')
		16th Char:		MP3 VBR Quality ('0' - '9')
		17th Char:		MP3 Bitrate ('0' - 'G', 'Z')
		18th Char: 		MP3 Minimum Bitrate ('0' - 'G')
		19th Char:		MP3 Maximum Bitrate ('0' - 'G')
		20th Char:		MP3 ABR ('0', '1')
		21st Char:		Global Normalize ('0', '1')
		22nd Char:		MP3 Use --nspsytune ('0', '1')
		23rd-26th:		MP3 Custom Bitrate ('Z' above, short)
		27th Char:		A52 Dynamic Range Compression ('0', '1')
		28th Char:		A52 Dolby Downmixing
		29th Char:		LAME Resampling
						Window Size
						AIFF Stuff
						MOV Stuff
	*/
	
	PLstrcpy(writeBuffer, "\p0000103000101404B9G0000000110"); // defaults
	bytesWritten = 30;
	
	// Find Preferences File
	theErr = FindFolder(kOnSystemDisk, kPreferencesFolderType, kDontCreateFolder, &volRefNum, &directoryID);
		if (theErr != noErr) return theErr;
	theErr = FSMakeFSSpec(volRefNum, directoryID, "\pmAC3dec Preferences", &fileSSpec);
		if (theErr == fnfErr)
		{	theErr = FSpCreate(&fileSSpec, '????', 'pref', smSystemScript);
				if (theErr != noErr) return theErr;
		}
		else if (theErr != noErr && theErr != fnfErr) return theErr;
	theErr = FSpOpenDF(&fileSSpec, fsWrPerm, &fileRefNum);
		if (theErr != noErr) return theErr;
	SetEOF(fileRefNum, 0);

	/***************************************/
	
	// Write prefs version (1)
	writeBuffer[1] = kPrefsVersion;

	// Write sample rate (2)
	     if (audioOutput == kAudio44khz)  writeBuffer[2] = '0';
	else if (audioOutput == kAudio48khz)  writeBuffer[2] = '1';
	else if (audioOutput == kAudioCustom) writeBuffer[2] = '2';

	// Write file format (3)
		 if (wrapper == kWrapperAIFF)	writeBuffer[3] = '0';
		 if (wrapper == kWrapperWAV)	writeBuffer[3] = '3';
	else if (wrapper == kWrapperPCM)	writeBuffer[3] = '1';
	else if (wrapper == kWrapperMOV)	writeBuffer[3] = '4';
	else if (wrapper == kWrapperMP3)	writeBuffer[3] = '2';
	else						writeBuffer[3] = '0';
	
	// Write channel split (4)
	if (channelSplit)	writeBuffer[4] = '1';

	// Write normalize (5)
	if (!normalize)		writeBuffer[5] = '0';

	// Write destination (6)
	if (chooseSave)		writeBuffer[6] = '1';

	// Write destination (7)
	if (!useProgressBar)	writeBuffer[7] = '4';
		 
	// Write auto-decode (8)
	if (autoDecode)	writeBuffer[8] = '1';
	
	// Write Altivec (9)
	if (useAltivec && hasAltivec)		writeBuffer[9] = '1';
	
	// Write MP3 Mono (10)
	if (mp3info.mono)		writeBuffer[10] = '1';
	
	// Write MP3 Joint Stereo (11)
	if (!mp3info.joint)		writeBuffer[11] = '0';
	
	// Write MP3 Fast Mode (12)
	if (mp3info.fast)		writeBuffer[12] = '1';
	
	// Write MP3 High Quality (13)
	if (!mp3info.high)		writeBuffer[13] = '0';
	
	// Write MP3 Sampling Rate (14)
		 if (mp3info.sample == 16000)	writeBuffer[14] = '0';
	else if (mp3info.sample == 22050) writeBuffer[14] = '1';
	else if (mp3info.sample == 24000) writeBuffer[14] = '2';
	else if (mp3info.sample == 32000) writeBuffer[14] = '3';
	else if (mp3info.sample == 44100) writeBuffer[14] = '4';
	else if (mp3info.sample == 48000) writeBuffer[14] = '5';
	else {	 mp3info.sample =  44100;	writeBuffer[14] = '4';	}
	
	// Write MP3 VBR (15)
	if (mp3info.vbr)		writeBuffer[15] = '1';
	
	// Write MP3 VBR Quality (16)
	switch (mp3info.vbrQuality)
	{	case 0: writeBuffer[16] = '0'; break;
		case 1: writeBuffer[16] = '1'; break;
		case 2: writeBuffer[16] = '2'; break;
		case 3: writeBuffer[16] = '3'; break;
		case 4: writeBuffer[16] = '4'; break;
		case 5: writeBuffer[16] = '5'; break;
		case 6: writeBuffer[16] = '6'; break;
		case 7: writeBuffer[16] = '7'; break;
		case 8: writeBuffer[16] = '8'; break;
		case 9: writeBuffer[16] = '9'; break;
		
		default: mp3info.vbrQuality = 4; writeBuffer[16] = '4'; break;
	}
	
	// Write MP3 Bitrate (17)
	switch (mp3info.bitrate)
	{	case 8:		writeBuffer[17] = '0'; break;
		case 16:	writeBuffer[17] = '1'; break;
		case 24:	writeBuffer[17] = '2'; break;
		case 32:	writeBuffer[17] = '3'; break;
		case 40:	writeBuffer[17] = '4'; break;
		case 48:	writeBuffer[17] = '5'; break;
		case 56:	writeBuffer[17] = '6'; break;
		case 64:	writeBuffer[17] = '7'; break;
		case 80:	writeBuffer[17] = '8'; break;
		case 96:	writeBuffer[17] = '9'; break;
		case 112:	writeBuffer[17] = 'A'; break;
		case 128:	writeBuffer[17] = 'B'; break;
		case 160:	writeBuffer[17] = 'C'; break;
		case 192:	writeBuffer[17] = 'D'; break;
		case 224:	writeBuffer[17] = 'E'; break;
		case 256:	writeBuffer[17] = 'F'; break;
		case 320:	writeBuffer[17] = 'G'; break;
		
		default: 
			if (mp3info.bitrate < 4 || mp3info.bitrate > 512)
			{	mp3info.bitrate = 128; writeBuffer[17] = 'B'; break;	}
			else					   writeBuffer[17] = 'Z';
	}

	// Write MP3 Minimum Bitrate (18)
	switch (mp3info.minBitrate)
	{	case 8:		writeBuffer[18] = '0'; break;
		case 16:	writeBuffer[18] = '1'; break;
		case 24:	writeBuffer[18] = '2'; break;
		case 32:	writeBuffer[18] = '3'; break;
		case 40:	writeBuffer[18] = '4'; break;
		case 48:	writeBuffer[18] = '5'; break;
		case 56:	writeBuffer[18] = '6'; break;
		case 64:	writeBuffer[18] = '7'; break;
		case 80:	writeBuffer[18] = '8'; break;
		case 96:	writeBuffer[18] = '9'; break;
		case 112:	writeBuffer[18] = 'A'; break;
		case 128:	writeBuffer[18] = 'B'; break;
		case 160:	writeBuffer[18] = 'C'; break;
		case 192:	writeBuffer[18] = 'D'; break;
		case 224:	writeBuffer[18] = 'E'; break;
		case 256:	writeBuffer[18] = 'F'; break;
		case 320:	writeBuffer[18] = 'G'; break;
		
		default: mp3info.minBitrate = 96; writeBuffer[18] = '8'; break;
	}

	// Write MP3 Maximum Bitrate (19)
	switch (mp3info.maxBitrate)
	{	case 8:		writeBuffer[19] = '0'; break;
		case 16:	writeBuffer[19] = '1'; break;
		case 24:	writeBuffer[19] = '2'; break;
		case 32:	writeBuffer[19] = '3'; break;
		case 40:	writeBuffer[19] = '4'; break;
		case 48:	writeBuffer[19] = '5'; break;
		case 56:	writeBuffer[19] = '6'; break;
		case 64:	writeBuffer[19] = '7'; break;
		case 80:	writeBuffer[19] = '8'; break;
		case 96:	writeBuffer[19] = '9'; break;
		case 112:	writeBuffer[19] = 'A'; break;
		case 128:	writeBuffer[19] = 'B'; break;
		case 160:	writeBuffer[19] = 'C'; break;
		case 192:	writeBuffer[19] = 'D'; break;
		case 224:	writeBuffer[19] = 'E'; break;
		case 256:	writeBuffer[19] = 'F'; break;
		case 320:	writeBuffer[19] = 'G'; break;
		
		default: mp3info.maxBitrate = 256; writeBuffer[19] = 'G'; break;
	}

	// Write MP3 ABR (20)
	if (mp3info.abr)		writeBuffer[20] = '1';

	// Write global normalize (21)
	if (normalizeGlobal)	writeBuffer[21] = '1';

	// Write MP3 nspsytune (22)
	if (mp3info.nspsytune)	writeBuffer[22] = '1';

	// Write MP3 custom bitrate (23-26)
	if (writeBuffer[17] == 'Z')
		BlockMoveData(&mp3info.bitrate, &writeBuffer[23], 4);

	// Write A52 Dynamic Range Compression (27)
	if (a52nodynrng)		writeBuffer[27] = '0';
	
	// Write A52 Dolby Downmixing (28)
	if (!a52dolby)			writeBuffer[28] = '0';
	
	// Write LAME Resampling (29)
	if (lameResample)		writeBuffer[29] = '1';
	
	/***************************************/

	// Write the data to the preferences file
	theErr = (FSWrite(fileRefNum, &bytesWritten, &writeBuffer));
		if (theErr != noErr) return theErr;
		
	// Write window size
	bytesWritten = sizeof(Rect);
	theErr = (FSWrite(fileRefNum, &bytesWritten, &Mac3dec->winRect));
		if (theErr != noErr) return theErr;
	
	// Write Custom AIFF settings
	/*bytesWritten = sizeof(SoundComponentData);
	theErr = (FSWrite(fileRefNum, &bytesWritten, &customSettings));
		if (theErr != noErr) return theErr;*/
	long hSize = 0;
	HLock(customHandle);
	hSize = GetHandleSize(customHandle);
	bytesWritten = sizeof(hSize);
	theErr = (FSWrite(fileRefNum, &bytesWritten, &hSize));
		if (theErr != noErr) return theErr;
	if (hSize > 0)
	{	bytesWritten = hSize;
		theErr = (FSWrite(fileRefNum, &bytesWritten, *customHandle));
			if (theErr != noErr) return theErr;
	}
	HUnlock(customHandle);

	// Write QuickTime settings
	/*bytesWritten = sizeof(SoundComponentData);
	theErr = (FSWrite(fileRefNum, &bytesWritten, &movSettings));
		if (theErr != noErr) return theErr;*/
	hSize = 0;
	HLock(movHandle);
	hSize = GetHandleSize(movHandle);
	bytesWritten = sizeof(hSize);
	theErr = (FSWrite(fileRefNum, &bytesWritten, &hSize));
		if (theErr != noErr) return theErr;
	if (hSize > 0)
	{	bytesWritten = hSize;
		theErr = (FSWrite(fileRefNum, &bytesWritten, *movHandle));
			if (theErr != noErr) return theErr;
	}
	HUnlock(movHandle);


	// Close the File
	FSClose(fileRefNum);

	return noErr;
}


OSErr Preferences::Read(void)
{
	OSErr theErr;
	SInt16 volRefNum;
	long directoryID, bytesRead, tempNum, hSize;
	FSSpec fileSSpec;
	SInt16 fileRefNum;
	Str255 buf;
	static BitMap screenBits;
	Rect boundsRect = GetQDGlobalsScreenBits(&screenBits)->bounds;
	Boolean moreToRead = true;

	/* Preferences Layout:
		1st Char:		Prefs Version (kPrefsVersion)
		2nd Char:		audioOutput ('0', '1', '2')
		3rd Char:		wrapper ('0', '1', '2')
		4th Char: 		channelSplit ('0', '1')
		5th Char:		normalize ('0', '1')
		6th Char:		chooseSave ('0', '1')
		7th Char:		Disable Progress Bar ('3', '4')
		8th Char:		autoDecode ('0', '1')
		9th Char:		useAltivec ('0', '1')
		10th Char:		MP3 mono ('0', '1')
		11th Char:		MP3 joint stereo ('0', '1')
		12th Char:		MP3 fast mode ('0', '1')
		13th Char:		MP3 high quality ('0', '1')
		14th Char:		MP3 Sampling Rate ('0' - '5')
		15th Char:		MP3 VBR ('0', '1')
		16th Char:		MP3 VBR Quality ('0' - '9')
		17th Char:		MP3 Bitrate ('0' - 'G', 'Z')
		18th Char: 		MP3 Minimum Bitrate ('0' - 'G')
		19th Char:		MP3 Maximum Bitrate ('0' - 'G')
		20th Char:		MP3 ABR ('0', '1')
		21st Char:		Global Normalize ('0', '1')
		22nd Char:		MP3 Use --nspsytune ('0', '1')
		23rd-26th:		MP3 Custom Bitrate ('Z' above, short)
		27th Char:		A52 Dynamic Range Compression ('0', '1')
		28th Char:		A52 Dolby Downmixing
		29th Char:		LAME Resampling
						Window Size
						AIFF Stuff
						MOV Stuff
	*/
	
	bytesRead = 30;

	// Find prefs file
	theErr = FindFolder(kOnSystemDisk, kPreferencesFolderType, kDontCreateFolder, &volRefNum, &directoryID);
		if (theErr != noErr) return theErr;	
	theErr = FSMakeFSSpec(volRefNum, directoryID, "\pmAC3dec Preferences", &fileSSpec);
		if (theErr == fnfErr) {Write(); return noErr;} // Write defaults
		else if (theErr != noErr && theErr != fnfErr) return theErr;
	theErr = FSpOpenDF(&fileSSpec, fsRdPerm, &fileRefNum);
		if (theErr != noErr) return theErr;
	theErr = FSRead(fileRefNum, &bytesRead, &buf);
		if (theErr != noErr && theErr != eofErr) return theErr;

	// Read window size
	moreToRead = (bytesRead == 30);
	if (moreToRead)
	{	tempNum = sizeof(Rect);
		theErr = FSRead(fileRefNum, &tempNum, &winSize);
		if (theErr != noErr)
			BlockMoveData(&kRectWindowSize, &winSize, sizeof(Rect));
		else
		{	if (winSize.top		<=	19
			 || winSize.top		>=	boundsRect.bottom
			 || winSize.left	>=	boundsRect.right
			 || winSize.bottom	<=	boundsRect.top
			 || winSize.right	<=	boundsRect.left )
			 	BlockMoveData(&kRectWindowSize, &winSize, sizeof(Rect));
		}
		
		moreToRead = (theErr == noErr && tempNum == sizeof(Rect));
	}
		
	// Read AIFF Custom Settings
	/*if (moreToRead)
	{	tempNum = sizeof(SoundComponentData);
		theErr = FSRead(fileRefNum, &tempNum, &customSettings);
		if (theErr != noErr)
		{	customSettings.format = kSoundNotCompressed;
			customSettings.numChannels = 2;
			customSettings.sampleSize = 16;
			customSettings.sampleRate = rate44khz;
		}
		moreToRead = (theErr == noErr && tempNum == sizeof(SoundComponentData));
	}*/	
	if (moreToRead)
	{	tempNum = sizeof(hSize);
		theErr = FSRead(fileRefNum, &tempNum, &hSize);
		moreToRead = (theErr == noErr && tempNum == sizeof(hSize));
		if (moreToRead && hSize > 0)
		{	tempNum = hSize;
			if (customHandle == 0L)		customHandle = NewHandle(tempNum);
			else						SetHandleSize(customHandle, tempNum);
			HLock(customHandle);
			theErr = FSRead(fileRefNum, &tempNum, *customHandle);
			moreToRead = (theErr == noErr && tempNum == hSize);
			
			if (moreToRead)
			{	HUnlock(customHandle);
			
				// Extract the info we need
				ComponentInstance theComponent = OpenDefaultComponent(StandardCompressionType, StandardCompressionSubTypeSound);
					if (theComponent == 0L) HandleError("\pCould not open Sound Components in GetOutputFormat", noErr, true);
				SCSetInfo(theComponent, scSettingsStateType, &customHandle);
				
				// Setup everything from the component data
				theErr = SCGetInfo(theComponent, scSoundCompressionType, &customSettings.format);
					IfError("\pCould not set compression type", theErr);
				theErr = SCGetInfo(theComponent, scSoundChannelCountType, &customSettings.numChannels);
					IfError("\pCould not set number of channels", theErr);
				theErr = SCGetInfo(theComponent, scSoundSampleRateType, &customSettings.sampleRate);
					IfError("\pCould not set sample rate", theErr);
				theErr = SCGetInfo(theComponent, scSoundSampleSizeType, &customSettings.sampleSize);
					IfError("\pCould not set sample size", theErr);
				if (customSettings.format == k8BitOffsetBinaryFormat)
					customSettings.format = kSoundNotCompressed;	
				customSettings.sampleCount = 0;
				customSettings.buffer = nil;
				customSettings.reserved = 0;
				customSettings.flags = kNoRealtimeProcessing;
				
				CloseComponent(theComponent);
			}
			else // If the handle doesn't look right, fuhgeddaboudit.
			{	DisposeHandle(customHandle);
				customHandle = 0L;
			}
		}
	}

	// Read QuickTime Settings
	/*if (moreToRead)
	{	tempNum = sizeof(SoundComponentData);
		theErr = FSRead(fileRefNum, &tempNum, &movSettings);
		if (theErr != noErr)
		{	movSettings.format = kSoundNotCompressed;
			movSettings.numChannels = 2;
			movSettings.sampleSize = 16;
			movSettings.sampleRate = rate44khz;
		}
		moreToRead = (theErr == noErr && tempNum == sizeof(SoundComponentData));
	}*/
	if (moreToRead)
	{	tempNum = sizeof(hSize);
		theErr = FSRead(fileRefNum, &tempNum, &hSize);
		moreToRead = (theErr == noErr && tempNum == sizeof(hSize));
		if (moreToRead && hSize > 0)
		{	tempNum = hSize;
			if (movHandle == 0L)	movHandle = NewHandle(tempNum);
			else					SetHandleSize(movHandle, tempNum);
			HLock(movHandle);
			theErr = FSRead(fileRefNum, &tempNum, *movHandle);
			moreToRead = (theErr == noErr && tempNum == hSize);
			
			if (moreToRead)
			{	HUnlock(movHandle);
				
				// Use what we have learned
				// Extract the info we need
				ComponentInstance theComponent = OpenDefaultComponent(StandardCompressionType, StandardCompressionSubTypeSound);
					if (theComponent == 0L) HandleError("\pCould not open Sound Components in GetOutputFormat", noErr, true);
				SCSetInfo(theComponent, scSettingsStateType, &movHandle);
				
				// Setup everything from the component data
				theErr = SCGetInfo(theComponent, scSoundCompressionType, &movSettings.format);
					IfError("\pCould not set compression type", theErr);
				theErr = SCGetInfo(theComponent, scSoundChannelCountType, &movSettings.numChannels);
					IfError("\pCould not set number of channels", theErr);
				theErr = SCGetInfo(theComponent, scSoundSampleRateType, &movSettings.sampleRate);
					IfError("\pCould not set sample rate", theErr);
				theErr = SCGetInfo(theComponent, scSoundSampleSizeType, &movSettings.sampleSize);
					IfError("\pCould not set sample size", theErr);
				if (movSettings.format == k8BitOffsetBinaryFormat)
					movSettings.format = kSoundNotCompressed;	
				movSettings.sampleCount = 0;
				movSettings.buffer = nil;
				movSettings.reserved = 0;
				movSettings.flags = kNoRealtimeProcessing;

				/* We need to get the appropriate codec settings for the compressor
				/  these 'magic' settings need to be added to the Sound Description */
				SCGetInfo(theComponent, scCodecSettingsType, &compressionParams);
		
				CloseComponent(theComponent);
			}
			else // If the handle doesn't look right, fuhgeddaboudit.
			{	DisposeHandle(movHandle);
				movHandle = 0L;
			}
		}
	}


	// Read prefs version, and delete if prefs are old (1)
	if (bytesRead < 2 || buf[1] != kPrefsVersion)
	{	SetEOF(fileRefNum, 0);
		FSClose(fileRefNum);
		Write();
		return noErr;
	}

	// Close file, as it's no longer needed
	FSClose(fileRefNum);
	
	/***************************************/

	// Read audio format (2)
	if (bytesRead < 3) {Write(); return noErr;}	
	     if (buf[2] == '1')  audioOutput = (short) kAudio48khz;
	else if (buf[2] == '2')  audioOutput = (short) kAudioCustom;
	else   /*buf[2] == '0'*/ audioOutput = (short) kAudio44khz;

	// Read file format (3)
	if (bytesRead < 4) {Write(); return noErr;}
		 if (buf[3] == '1')		wrapper = kWrapperPCM;
	else if (buf[3] == '3')		wrapper = kWrapperWAV;
	else if (buf[3] == '0')		wrapper = kWrapperAIFF;
	else if (buf[3] == '4')		wrapper = kWrapperMOV;
	else /*  buf[3] == '2'*/	wrapper = kWrapperMP3;
	
	// Read channel split (4)
	if (bytesRead < 5) {Write(); return noErr;}
		if (buf[4] == '1')		channelSplit = true;
	else /* buf[4] == '0'*/		channelSplit = false;
	
	// Read normalize (5)
	if (bytesRead < 6) {Write(); return noErr;}
		if (buf[5] == '0')		normalize = false;
	else /* buf[5] == '1'*/		normalize = true;
		
	// Read output destination (6)
	if (bytesRead < 7) {Write(); return noErr;}
		if (buf[6] == '0')  	chooseSave = false;
	else  /*buf[6] == '1'*/ 	chooseSave = true;

	// Read progress bar disablification (7)
	if (bytesRead < 8) {Write(); return noErr;}
		if (buf[7] == '4')		useProgressBar = false;
	else /* buf[7] == '3' */	useProgressBar = true;

	// Read auto-decode (8)
	if (bytesRead < 9) {Write(); return noErr;}
		if (buf[8] == '1')		autoDecode = true;
	else /* buf[8] == '0' */	autoDecode = false;
	
	// Read altivec (9)
	if (bytesRead < 10) {Write(); return noErr;}
		if (buf[9] == '1')	useAltivec = true;
	else /* buf[9] == '0'*/	useAltivec = false;
	
		/* Right now, altivec OFF: */ useAltivec = false;

	// Read MP3 Mono (10)
	if (bytesRead < 11) {Write(); return noErr;}
		if (buf[10] == '1')		mp3info.mono = true;
	else /* buf[10] == '0' */	mp3info.mono = false;
	
	// Read MP3 Joint Stereo (11)
	if (bytesRead < 12) {Write(); return noErr;}
		if (buf[11] == '0')		mp3info.joint = false;
	else /* buf[11] == '1' */	mp3info.joint = true;
	
	// Read MP3 Fast Mode (12)
	if (bytesRead < 13) {Write(); return noErr;}
		if (buf[12] == '1')		mp3info.fast = true;
	else /* buf[12] == '0' */	mp3info.fast = false;

	// Read MP3 High Quality (13)
	if (bytesRead < 14) {Write(); return noErr;}
		if (buf[13] == '0')		mp3info.high = false;
	else /* buf[13] == '1' */	mp3info.high = true;
	
	// Read MP3 Sampling Rate (14)
	if (bytesRead < 15) {Write(); return noErr;}
	switch (buf[14])
	{	case '0': mp3info.sample = 16000; break;
		case '1': mp3info.sample = 22050; break;
		case '2': mp3info.sample = 24000; break;
		case '3': mp3info.sample = 32000; break;
		case '4': mp3info.sample = 44100; break;
		case '5': mp3info.sample = 48000; break;
		
		default:  mp3info.sample = 44.100f; break;
	}

	// Read MP3 VBR (15)
	if (bytesRead < 16) {Write(); return noErr;}
		if (buf[15] == '1')		mp3info.vbr = true;
	else /* buf[15] == '0' */	mp3info.vbr = false;

	// Read MP3 VBR Quality (16)
	if (bytesRead < 17) {Write(); return noErr;}
	switch (buf[16])
	{	case '0': mp3info.vbrQuality = 0; break;
		case '1': mp3info.vbrQuality = 1; break;
		case '2': mp3info.vbrQuality = 2; break;
		case '3': mp3info.vbrQuality = 3; break;
		case '4': mp3info.vbrQuality = 4; break;
		case '5': mp3info.vbrQuality = 5; break;
		case '6': mp3info.vbrQuality = 6; break;
		case '7': mp3info.vbrQuality = 7; break;
		case '8': mp3info.vbrQuality = 8; break;
		case '9': mp3info.vbrQuality = 9; break;
		
		default:  mp3info.vbrQuality = 4; break;
	}
	
	// Read MP3 Bitrate (17)
	if (bytesRead < 18) {Write(); return noErr;}
	switch (buf[17])
	{	case '0': mp3info.bitrate = 8;	break;
		case '1': mp3info.bitrate = 16;	break;
		case '2': mp3info.bitrate = 24; break;
		case '3': mp3info.bitrate = 32;	break;
		case '4': mp3info.bitrate = 40;	break;
		case '5': mp3info.bitrate = 48; break;
		case '6': mp3info.bitrate = 56; break;
		case '7': mp3info.bitrate = 64; break;
		case '8': mp3info.bitrate = 80; break;
		case '9': mp3info.bitrate = 96; break;
		case 'A': mp3info.bitrate =112; break;
		case 'B': mp3info.bitrate =128; break;
		case 'C': mp3info.bitrate =160; break;
		case 'D': mp3info.bitrate =192; break;
		case 'E': mp3info.bitrate =224; break;
		case 'F': mp3info.bitrate =256; break;
		case 'G': mp3info.bitrate =320; break;
		
		default:  mp3info.bitrate =128; break;
	}

	// Read MP3 Minimum Bitrate (18)
	if (bytesRead < 19) {Write(); return noErr;}
	switch (buf[18])
	{	case '0': mp3info.minBitrate = 8;  break;
		case '1': mp3info.minBitrate = 16; break;
		case '2': mp3info.minBitrate = 24; break;
		case '3': mp3info.minBitrate = 32; break;
		case '4': mp3info.minBitrate = 40; break;
		case '5': mp3info.minBitrate = 48; break;
		case '6': mp3info.minBitrate = 56; break;
		case '7': mp3info.minBitrate = 64; break;
		case '8': mp3info.minBitrate = 80; break;
		case '9': mp3info.minBitrate = 96; break;
		case 'A': mp3info.minBitrate =112; break;
		case 'B': mp3info.minBitrate =128; break;
		case 'C': mp3info.minBitrate =160; break;
		case 'D': mp3info.minBitrate =192; break;
		case 'E': mp3info.minBitrate =224; break;
		case 'F': mp3info.minBitrate =256; break;
		case 'G': mp3info.minBitrate =320; break;
		
		default:  mp3info.minBitrate = 80; break;
	}

	// Read MP3 Maximum Bitrate (19)
	if (bytesRead < 20) {Write(); return noErr;}
	switch (buf[19])
	{	case '0': mp3info.maxBitrate = 8;  break;
		case '1': mp3info.maxBitrate = 16; break;
		case '2': mp3info.maxBitrate = 24; break;
		case '3': mp3info.maxBitrate = 32; break;
		case '4': mp3info.maxBitrate = 40; break;
		case '5': mp3info.maxBitrate = 48; break;
		case '6': mp3info.maxBitrate = 56; break;
		case '7': mp3info.maxBitrate = 64; break;
		case '8': mp3info.maxBitrate = 80; break;
		case '9': mp3info.maxBitrate = 96; break;
		case 'A': mp3info.maxBitrate =112; break;
		case 'B': mp3info.maxBitrate =128; break;
		case 'C': mp3info.maxBitrate =160; break;
		case 'D': mp3info.maxBitrate =192; break;
		case 'E': mp3info.maxBitrate =224; break;
		case 'F': mp3info.maxBitrate =256; break;
		case 'G': mp3info.maxBitrate =320; break;
		
		default:  mp3info.maxBitrate =320; break;
	}

	// Read MP3 ABR (20)
	if (bytesRead < 21) {Write(); return noErr;}
		if (buf[20] == '1')		mp3info.abr = true;
	else /* buf[20] == '0' */	mp3info.abr = false;
		
	// Read global normalize (21)
	if (bytesRead < 22) {Write(); return noErr;}
		if (buf[21] == '1')		normalizeGlobal = true;
	else /* buf[21] == '0' */	normalizeGlobal = false;

	// Read MP3 nspsytune (22)
	if (bytesRead < 23) {Write(); return noErr;}
		if (buf[22] == '1')		mp3info.nspsytune = true;
	else /* buf[22] == '0' */	mp3info.nspsytune = false;

	// Read MP3 custom bitrate (23-26)
	if (bytesRead < 27) {Write(); return noErr;}
	if (buf[17] == 'Z')
	{	BlockMoveData(&buf[23], &mp3info.bitrate, 4);
		if (mp3info.bitrate < 4 || mp3info.bitrate > 512)
			mp3info.bitrate = 128;
	}
	
	// Read Dynamic Range Compression (27)
	if (bytesRead < 28) {Write(); return noErr;}
		if (buf[27] == '0')		a52nodynrng = true;
	else /* buf[27] == '1' */	a52nodynrng = false;
	
	// Read Dolby Downmixing (28)
	if (bytesRead < 29) {Write(); return noErr;}
		if (buf[28] == '0')		a52dolby = false;
	else /* buf[28] == '1' */	a52dolby = true;

	// Read Lame Resampling (29)
	if (bytesRead < 30) {Write(); return noErr;}
		if (buf[29] == '1')		lameResample = true;
	else /* buf[29] == '0' */	lameResample = false;
	

	/***************************************/

	return noErr;
}



#pragma mark -

// Constructor/Destructor
// -----------------------

Decoder::Decoder(void)
{
	OSErr	theErr;
	long	gestaltResponse;
	Handle	menubarHdl;
	Boolean	itemDefault;
	Str255	theString;
	Point	cellSize;
	Rect	gridRect;
	ControlFontStyleRec bigRight, littleLeft, littleRight;
	ProcessSerialNumber	ourProcess = {0, kCurrentProcess};

	// Initialize
	InitCursor();
	FlushEvents(everyEvent, 0);
	active = 	true;
	state =		kStateIdle;
	numItems =	0;
	inputRef =	0;
	output =	0;

	// Allocate memory
	MoreMasters();	MoreMasters();
	inputBuffer = (unsigned char*) vec_malloc(kInputBufferSize);
		if (MemError() || inputBuffer == nil)  HandleError("\pOut of Memory", MaxBlock(), true);
	outputBuffer = (char*) vec_malloc(kOutputBufferSize);
		if (MemError() || inputBuffer == nil)  HandleError("\pOut of Memory", MaxBlock(), true);
	a52Buffer = (char*) vec_malloc(kA52BufferSize);
		if (MemError() || a52Buffer == nil)  HandleError("\pOut of Memory", MaxBlock(), true);

	// Setup QuickTime
	EnterMovies();

	// Load Navigation Services
	theErr = NavLoad();
		if (theErr != noErr || NavServicesAvailable() == false) HandleError("\pNavigation Services is not present", theErr, true);

	// Now Install the Apple Event Handlers
	InstallAppleEvents();

	// Setup Appearance Manager so we look pretty
	RegisterAppearanceClient();

	// Load preferences
	Prefs = new Preferences;

	// Check for Altivec:
	theErr = Gestalt(gestaltPowerPCProcessorFeatures, &gestaltResponse);
	if (theErr == noErr) Prefs->hasAltivec = ((1 << gestaltPowerPCHasVectorInstructions) & gestaltResponse) != 0;
	else HandleError("\pCould not detect processor features", theErr, false);
	if (Prefs->hasAltivec == false) Prefs->useAltivec = false;

	// Check to see if we're using OS X
	theErr = Gestalt(gestaltSystemVersion, &gestaltResponse);
	if (theErr == noErr) Prefs->osX = (gestaltResponse >= 0x00001000);
	else HandleError("\pCould not detect operating system", theErr, false);
	
	// Check for pre-OS9, meaning no FSRefs
	Prefs->useRefs = (gestaltResponse >= 0x00000900);
		
	// Get the address of our process
	theErr = AECreateDesc(typeProcessSerialNumber, &ourProcess, sizeof(ProcessSerialNumber), &OurAddress);
		if (theErr != noErr) HandleError("\pError getting process address", theErr, true);

	// Get the menu bar, the X one if we're using OS X.
	if (Prefs->osX)
	{	if (!(menubarHdl = GetNewMBar(rMenubarX)))
			HandleError("\pCould not set up menu bar", noErr, true);
		SetMenuBar(menubarHdl);
		if (!(FileMenu = GetMenuHandle(rMenuFileX)))
			HandleError("\pCould not get handle to Decode Menu", noErr, true);
	} else
	{	if (!(menubarHdl = GetNewMBar(rMenubar)))
			HandleError("\pCould not set up menu bar", noErr, true);
		SetMenuBar(menubarHdl);
		if (!(FileMenu = GetMenuHandle(rMenuFile)))
			HandleError("\pCould not get handle to Decode Menu", noErr, true);
	}
	if (!(DecodeMenu = GetMenuHandle(rMenuDecode)))
		HandleError("\pCould not get handle to Decode Menu", noErr, true);

	DrawMenuBar();
	EnableMenuCommand(NULL, kHICommandPreferences);
	DisableMenuItem(DecodeMenu, rMenuDecodePause);

	// Create window and interface
	BlockMoveData(&Prefs->winSize, &winRect, sizeof(Rect));
	window = NULL;
	window = NewCWindow(NULL, &winRect, "\p m  A  C  3  d  e  c ", false, kWindowGrowDocumentProc, (WindowPtr)-1, false, 0);

	// Create batch list
	BlockMoveData(&kRectList, &listRect, sizeof(Rect));
	SetRect(&gridRect, 0, 0, 1, 0);
	SetPt(&cellSize, 1024, 16);
	list = LNew(&listRect, &gridRect, cellSize, 0, window, 1, 0, 0, 1);
	LSetDrawingMode(true, list);
	SetWRefCon(window, (SInt32) list);

	// Create controls
	startButton		= NewControl(window, &kRectStartButton,	"\pStart", true, 0, 0, 1, kControlPushButtonProc, 0);
	progressBar		= NewControl(window, &kRectProgressBar,	"\p", true, 0, 0, kProgressBarMax, kControlProgressBarProc, 0);
	splitToggle		= NewControl(window, &kRectChannelSplit,"\pSplit Channels", true, Prefs->channelSplit == true, 0, 1, kControlCheckBoxProc, 0);
	separator		= NewControl(window, &kRectSeparator,	"\p", true, 0, 0, 1, kControlSeparatorLineProc, 0);
	sampleRate		= NewControl(window, &kRectSampleRate,	"\p", Prefs->wrapper < kWrapperMP3, 0, 1128, 0, popupMenuProc, 0);
	wrapper			= NewControl(window, &kRectWrapper,		"\p", true, 0, 1129, 0, popupMenuProc, 0);
	mp3Settings		= NewControl(window, &kRectMP3Settings, "\pMP3 Options", Prefs->wrapper >= kWrapperMP3, 0, 0, 1, kControlBevelButtonSmallBevelProc, 0);
	normToggle		= NewControl(window, &kRectNormToggle,	"\pNormalize", true, (Prefs->normalize), 0, 1, kControlCheckBoxProc, 0);
	normGlobalToggle= NewControl(window, &kRectNormGToggle,	"\pGlobally", true, (Prefs->normalizeGlobal), 0, 1, kControlCheckBoxProc, 0);
	decodeText		= NewControl(window, &kRectDecodeText,	"\p", true, 0, 0, 0, kControlStaticTextProc, 0);
	timeText		= NewControl(window, &kRectTimeText,	"\p", true, 0, 0, 0, kControlStaticTextProc, 0);
	amountText		= NewControl(window, &kRectAmountText,	"\p", true, 0, 0, 0, kControlStaticTextProc, 0);
	detailsText		= NewControl(window, &kRectDetailsText,	"\p", true, 0, 0, 0, kControlStaticTextProc, 0);
	sampleRateLabel	= NewControl(window, &kRectSampleRateText,"\p", Prefs->wrapper < kWrapperMP3, 0, 0, 0, kControlStaticTextProc, 0);
	wrapperLabel	= NewControl(window, &kRectWrapperText,	"\p", true, 0, 0, 0, kControlStaticTextProc, 0);

	// Set the font and such for text boxes and make the window PRETTY!
	bigRight.flags	= littleLeft.flags	= littleRight.flags = (kControlUseFontMask | kControlUseSizeMask | kControlUseJustMask);
	littleLeft.size	= littleLeft.font	= littleRight.font	= littleRight.size = kControlFontSmallSystemFont;
	bigRight.font	= bigRight.size		= kControlFontBigSystemFont;
	littleLeft.just	= teFlushLeft;
	bigRight.just	= littleRight.just = teFlushRight;

	SetControlFontStyle(timeText,			&bigRight);
	
	SetControlFontStyle(mp3Settings,		&littleLeft);
	SetControlFontStyle(amountText,			&littleLeft);
	SetControlFontStyle(sampleRateLabel,	&littleLeft);
	SetControlFontStyle(wrapperLabel,		&littleLeft);
	SetControlFontStyle(normToggle,			&littleLeft);
	SetControlFontStyle(normGlobalToggle,	&littleLeft);
	SetControlFontStyle(splitToggle,		&littleLeft);
	
	SetControlFontStyle(detailsText,		&littleRight);
	SetControlFontStyle(sampleRate,			&littleRight);
	SetControlFontStyle(wrapper,			&littleRight);

	// Correct text labels
	PLstrcpy(theString, "\pSample Rate");
	SetControlData(sampleRateLabel, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	PLstrcpy(theString, "\pFormat");
	SetControlData(wrapperLabel, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	
	// Make Start Button Default
	itemDefault = !Prefs->osX;
	SetControlData(startButton, kControlNoPart, kControlPushButtonDefaultTag, sizeof (itemDefault), Ptr (&itemDefault));
	
	// Set initial control values
	Remove(-1);
	DisableMenuItem(FileMenu,		rMenuFileRemove);
	DisableMenuItem(DecodeMenu,		rMenuDecodeFix);
	SetControlValue(sampleRate, 	Prefs->audioOutput + 1);
	SetControlValue(wrapper,		Prefs->wrapper + 1);
	SetControlValue(normToggle,		Prefs->normalize);
	SetControlValue(normGlobalToggle,Prefs->normalizeGlobal);
	SetControlValue(splitToggle,	Prefs->channelSplit);
	
	if (!Prefs->normalize || !GetControlValue(normToggle))
		DeactivateControl(normGlobalToggle);
	if (Prefs->wrapper >= kWrapperMP3)
	{	SetControlValue(splitToggle, 0);
		Prefs->channelSplit = false;
		DeactivateControl(splitToggle);
			 if (Prefs->wrapper == kWrapperMP3)
			SetControlTitle(mp3Settings, "\pMP3 Options");
		else if (Prefs->wrapper == kWrapperMOV)
			SetControlTitle(mp3Settings, "\pQT Options");
	}
	else if (Prefs->channelSplit)
	{	SetControlValue(normToggle, 0);
		SetControlValue(normGlobalToggle, 0);
		Prefs->normalize = false;
		Prefs->normalizeGlobal = false;
		DeactivateControl(sampleRate);
		DeactivateControl(sampleRateLabel);
		DeactivateControl(normToggle);
		DeactivateControl(normGlobalToggle);
	}

	// Initialize Drag and Drop
	InstallTrackingHandler(NewDragTrackingHandlerUPP(HandleDragTrack), window, 0L);
	InstallReceiveHandler(NewDragReceiveHandlerUPP(HandleDragReceive), window, 0L);

	// Draw Window
	Resize(winRect.right - winRect.left, winRect.bottom - winRect.top);
	SetThemeWindowBackground(window, kThemeBrushModelessDialogBackgroundActive, true);
	ShowWindow(window);
	DrawControls(window);
	DrawThemeListBoxFrame(&listRect, kThemeStateActive);
	SetPortWindowPort(window);
}


Decoder::~Decoder(void)
{
	// Free memory
	vec_free(inputBuffer);
	vec_free(outputBuffer);
	vec_free(a52Buffer);

	// Delete prefs; the latest prefs will be updated by the destructor
	delete Prefs;
}


void Decoder::FillBuffer(unsigned char **start, long amount)
{
	OSErr theErr;
	unsigned long bufferSize;

	*start = inputBuffer;

	if (Prefs->useRefs)
		theErr = FSReadFork(inputRef, fsAtMark, 0, amount, inputBuffer, &bufferSize);
	else
	{	long bizzytes = amount;
		theErr = FSRead(input, &bizzytes, inputBuffer);
		bufferSize = bizzytes;
	}

	// First see if we're done (at end of file)
	if (bufferSize < amount)
		state = kStateDone;
	else // Otherwise, error-handling as normal
		if (theErr != noErr) HandleError("\pCould not read to input buffer", theErr, false);

	if (Prefs->useRefs)	FSGetForkPosition(inputRef, &inputEOF);
	else {long x;		GetFPos(input, &x);			 inputEOF = (long long) x;}
	

	//*end = *start + bufferSize;
}





#pragma mark -


// Event handling

void Decoder::EventLoop(void)
{
	while (!Done)
	{
		if (active)
		{	for (int i = 0; i < 8; i++)
			{	if (WaitNextEvent(everyEvent, &Event, 1, nil))
					HandleEvent(&Event);
		}	}
		else
		{	for (int i = 0; i < 8; i++)
			{	if (WaitNextEvent(everyEvent, &Event, 20, nil))
					HandleEvent(&Event);
		}	}

		if (state == kStateStart && Prefs->autoDecode)
		{	state = kStateIdle;
			Start();
		}
	}
}


void Decoder::HandleEvent(EventRecord* event)
{
	unsigned char charCode;
	RgnHandle theRgn;
	Rect tempRect = {0, 0, 0, 0};

	// Menu shortcuts
	if (event->what == keyDown && (event->modifiers & cmdKey) != 0)
	{	charCode = event->message & charCodeMask;
		HandleMenuChoice(MenuKey(charCode));
		return;
	}

	switch(event->what)
	{
		case mouseDown:
			HandleMouseDown(event);
			break;
		
		case autoKey:
		case keyDown:
			HandleKeyDown(event->message, (event->modifiers & 512) != 0);
			break;
			
		case kHighLevelEvent:
			AEProcessAppleEvent(event); // Do the Apple Event in question
			break;
			
		case updateEvt:
			if ((WindowPtr) event->message == window)
			{	BeginUpdate(window);
				DrawControls(window);
				tempRect.right	= winRect.right  - winRect.left;
				tempRect.bottom	= winRect.bottom - winRect.top;
				RectRgn(theRgn = NewRgn(), &tempRect);
				DrawThemeListBoxFrame(&listRect, active ? kThemeStateActive : kThemeStateInactive);
		    	LUpdate(theRgn, list);
				LActivate(active, list);
		    	DisposeRgn(theRgn);
				EndUpdate(window);
			}
			break;

		case activateEvt:
			tempRect.right	= winRect.right  - winRect.left;
			tempRect.bottom	= winRect.bottom - winRect.top;
			RectRgn(theRgn = NewRgn(), &tempRect);
			LUpdate(theRgn, list);
			if (event->modifiers & activeFlag)
			{	active = true;
				LActivate(true, list);
				ActivateControl(progressBar);
				DrawThemeListBoxFrame(&listRect, kThemeStateActive);
				if (state <= kStateIdle)
				{	ActivateControl(wrapper);
					ActivateControl(wrapperLabel);
					ActivateControl(mp3Settings);
					if (!Prefs->channelSplit)
					{	ActivateControl(normToggle);
						ActivateControl(sampleRate);
						ActivateControl(sampleRateLabel);
						if (Prefs->normalize)
							ActivateControl(normGlobalToggle);
					}
					if (Prefs->wrapper <= kWrapperMP3)
						ActivateControl(splitToggle);
				}
				if (numItems > 0 || state > kStateIdle)
					ActivateControl(startButton);
			}
			else
			{	active = false;
				LActivate(false, list);
				DrawThemeListBoxFrame(&listRect, kThemeStateInactive);
				DeactivateControl(progressBar);
				DeactivateControl(sampleRate);
				DeactivateControl(wrapper);
				DeactivateControl(sampleRateLabel);
				DeactivateControl(wrapperLabel);
				DeactivateControl(mp3Settings);
				DeactivateControl(startButton);
				DeactivateControl(normToggle);
				DeactivateControl(normGlobalToggle);
				DeactivateControl(splitToggle);
			}
			DisposeRgn(theRgn);
			break;
			
		case osEvt:
			HiliteMenu(0);
			break;
	}
}


void Decoder::HandleMouseDown(EventRecord* event)
{
	WindowPtr			windowPtr;
	short				partCode;
	short				i;
	Point				thePt, startPoint;
	static BitMap		screenBits;
	Rect				boundsRect = GetQDGlobalsScreenBits(&screenBits)->bounds;
	
	SetPortWindowPort(window);
	
	partCode = FindWindow(event->where, &windowPtr);
	
	if (partCode == inMenuBar)
	{	HandleMenuChoice(MenuSelect(event->where)); return;	}
	
	if (windowPtr != window) return;

	switch (partCode)
	{
		case inContent:
			if (FrontWindow() != window)
			{	SelectWindow(window);
				return;
			}
			HandleContentClick(event);
			break;
		
		case inDrag:
			if (FrontWindow() != window)
			{	SelectWindow(window);
				return;
			}
			if (!(event->modifiers & cmdKey) && FrontWindow() != window)
				SelectWindow(window);
			else
			{	// Drag Window
				if (IsWindowCollapsed(window) || Prefs->osX)
				{	DragWindow(window, event->where, &boundsRect);
				}
				else
				{	i = 0;
					startPoint = event->where;
				
					boundsRect.bottom	-= (winRect.bottom	- winRect.top);
					boundsRect.right	-= (winRect.right	- winRect.left);
					
					while (event->what != mouseUp)
					{	WaitNextEvent(everyEvent, event, nil, nil);
						thePt.h = event->where.h - startPoint.h + winRect.left;
						thePt.v = event->where.v - startPoint.v + winRect.top;
						
						if (thePt.v <= 40)	thePt.v = 40;

						MoveWindow(window, thePt.h, thePt.v, 0);
						
						if (event->what == updateEvt)
						{	i++;
							if (i == 10)
							{	HandleEvent(event);
								i = 0;
						}	}
					}
				}
				GetWindowBounds(window, kWindowContentRgn, &winRect);
			}
			break;

		case inGrow:
			Rect growSize = (boundsRect);
			growSize.left = kMinWindowWidth; growSize.top = kMinWindowHeight;
			long newSize = GrowWindow(window, event->where, &growSize);
			Resize(LoWord(newSize), HiWord(newSize));

			break;
	}
}


void Decoder::HandleKeyDown(unsigned char charCode, Boolean shiftDown)
{
	int i = 0, j = 0;
	Cell theCell = {0, 0};
	Boolean cellSelected;
	
	// No key activity allowed if a decode is active
	if (state != kStateIdle) return;

	switch (charCode)
	{
		case kUpArrow:
			cellSelected = LGetSelect(true, &theCell, list);
			
			if (cellSelected && theCell.v != 0)
			{	if (!shiftDown)
				{	// Deselect all list items
					Cell tempCell = {0, 0};
					for (j = 0; j < (numItems); j++, tempCell.v++)
						LSetSelect(false, tempCell, list);
				}
				theCell.v--;
				LSetSelect(true, theCell, list);
			}
			else
			{	LSetSelect(true, theCell, list);
			}
			LAutoScroll(list);
			Details(0, 0, 0, 0, false);
			break;
		
		case kDownArrow:
			cellSelected = LGetSelect(true, &theCell, list);
			
			// Get the last selected item of a continuous list
			while (LGetSelect(true, &theCell, list)) theCell.v++; theCell.v--;

			if (cellSelected && theCell.v < (numItems - 1))
			{	if (!shiftDown)
				{	// Deselect all list items
					Cell tempCell = {0, 0};
					for (j = 0; j < (numItems - 1); j++, tempCell.v++)
						LSetSelect(false, tempCell, list);
				}
				theCell.v++;
				LSetSelect(true, theCell, list);
			}
			else
			{	LSetSelect(true, theCell, list);
			}
			LAutoScroll(list);
			Details(0, 0, 0, 0, false);
			break;
		
		case kDeleteKey:
		case kOtherDeleteKey:
			while(LGetSelect(true, &theCell, list))
				Remove(theCell.v);
			Details(0, 0, 0, 0, false);
			break;

		case kReturnKey:
			if (IsControlActive(startButton) && state == kStateIdle)
			{	HiliteControl(startButton, true);
				Delay(8, NULL);
				HiliteControl(startButton, false);
				
				HandleMenuChoice((rMenuDecode << 16) | rMenuDecodeStart);
			}
			break;
		
		default:
			//HandleError("\pThis key:", charCode, false);
			break;

	}

	if (LGetSelect(true, &theCell, list))
	{	EnableMenuItem(FileMenu,	rMenuFileRemove);
		EnableMenuItem(DecodeMenu,	rMenuDecodeFix);
	}
	else
	{	DisableMenuItem(FileMenu,	rMenuFileRemove);
		DisableMenuItem(DecodeMenu,	rMenuDecodeFix);
	}
}


void Decoder::HandleMenuChoice(long menuChoice)
{
	OSErr	theErr = noErr;
	SInt16	menuID, menuItem;
	Cell theCell = {0, 0};

	menuID = HiWord(menuChoice);
	menuItem = LoWord(menuChoice);

	if (menuID == 0) return;
	
	switch (menuID)
	{
		case rMenuApple:
			if (menuItem == rMenuAppleAbout)
				AboutBox();
			break;
		
		case rMenuFileX:
			if (menuItem >= rMenuFileMP3Settings)
				menuItem++;
		case rMenuFile:
			switch (menuItem)
			{
				case rMenuFileQuit:
					Done = true;
					state = kStateCancel;
					break;
				
				case rMenuFileOpen:
					GetFileWithNav();
					break;
				
				case rMenuFileRemove:
					while(LGetSelect(true, &theCell, list))
						Remove(theCell.v);
					break;
				
				case rMenuFilePrefs:
					if (Prefs->osX)
						Prefs->MP3Dialog();
					else
						Prefs->Dialog();
					break;
				
				case rMenuFileMP3Settings:
					Prefs->MP3Dialog();
					break;
				
				case rMenuFileMOVSettings:
					theErr = GetQuickTimeOutputFormat(&Prefs->movSettings);
					break;
			}
			break;
		
		case rMenuDecode:
			switch (menuItem)
			{
				case rMenuDecodeStart:
					Start();
					break;
				
				case rMenuDecodePause:
					// Can't pause nothing
					if (state < kStateIdle) break;
					
					// Unpause
					if (state == kStatePaused)
					{	state = kStateDecoding;
						SetMenuItemText(DecodeMenu, rMenuDecodePause, "\pPause");
						break;
					}
					else
					{	SetMenuItemText(DecodeMenu, rMenuDecodePause, "\pUnpause");
						HiliteMenu(0);
						// Stay here until user deselects pause
						state = kStatePaused;
						while (state == kStatePaused)
						{	if (WaitNextEvent(everyEvent, &Event, 100, nil))
								HandleEvent(&Event);
						}
					}
					break;
					
				case rMenuDecodeFix:
					FixAC3();
					state = kStateIdle;
					break;
			}
			break;
	}
	
	HiliteMenu(0);
}


void Decoder::HandleContentClick(EventRecord* event)
{
	OSErr				theErr = noErr;
	Cell				theCell = {0};
	Point				thePt = event->where;
	ControlPartCode		hitPart;
	ControlHandle		hitControl;

	SetPortWindowPort(window);
	GlobalToLocal(&thePt);
	listRect.right += 16;
	if (PtInRect(thePt, &listRect) && state == kStateIdle)
	{	listRect.right -= 16;
		LClick(thePt, event->modifiers, list);
		if (LGetSelect(true, &theCell, list))
		{	EnableMenuItem(FileMenu,	rMenuFileRemove);
			EnableMenuItem(DecodeMenu,	rMenuDecodeFix);
		}
		else
		{	DisableMenuItem(FileMenu,	rMenuFileRemove);
			DisableMenuItem(DecodeMenu,	rMenuDecodeFix);
		}
		Details(0, 0, 0, 0, false);
		return;
	}	listRect.right -= 16;
	
	hitPart = FindControl(thePt, window, &hitControl);
	
	if (hitPart != kControlNoPart)
	{
		if (TrackControl(hitControl, thePt, (ControlActionUPP) -1) == hitPart)
		{
			if (hitControl == startButton)
				HandleMenuChoice((rMenuDecode << 16) | rMenuDecodeStart);

			else if (hitControl == sampleRate)
			{	Prefs->audioOutput = GetControlValue(sampleRate)-1;
				if (Prefs->audioOutput == kAudioCustom)
				{	theErr = GetOutputFormat(&Prefs->customSettings, 0); // Ask for output sound format
					if (theErr != noErr)
					{	SetControlValue(sampleRate, kAudio44khz+1);
						Prefs->audioOutput = kAudio44khz;
			}	}	}

			else if (hitControl == wrapper)
			{	Prefs->wrapper = GetControlValue(wrapper)-1;

				// Cool Interface Shite				
				if (Prefs->wrapper <= kWrapperPCM)
				{	ActivateControl(splitToggle);
					if (!GetControlValue(splitToggle))
					{	ActivateControl(sampleRate);
						ActivateControl(sampleRateLabel);
				}	}
				else
				{	DeactivateControl(splitToggle);
					SetControlValue(splitToggle, false);
					Prefs->channelSplit = false;
				}
				
				if (Prefs->wrapper == kWrapperMP3)
				{	SetControlValue(sampleRate, 1);
					HideControl(sampleRate);
					HideControl(sampleRateLabel);
					ShowControl(mp3Settings);
					SetControlValue(splitToggle, 0);
					ActivateControl(normToggle);
					if (GetControlValue(normToggle)) ActivateControl(normGlobalToggle);
					SetControlTitle(mp3Settings, "\pMP3 Options");
				}
				else if (Prefs->wrapper == kWrapperMOV)
				{	SetControlValue(sampleRate, 1);
					HideControl(sampleRate);
					HideControl(sampleRateLabel);
					ShowControl(mp3Settings);
					SetControlValue(splitToggle, 0);
					ActivateControl(normToggle);
					if (GetControlValue(normToggle)) ActivateControl(normGlobalToggle);
					SetControlTitle(mp3Settings, "\pQT Options");
				}
				else
				{	HideControl(mp3Settings);
					ShowControl(sampleRate);
					ShowControl(sampleRateLabel);
				}
			}

			else if (hitControl == mp3Settings)
			{	if (Prefs->wrapper == kWrapperMP3)
					Prefs->MP3Dialog();
					
				else if (Prefs->wrapper == kWrapperMOV)
					theErr = GetQuickTimeOutputFormat(&Prefs->movSettings); // Ask for output sound format
			}
			
			else if (hitControl == normToggle)
			{	if (GetControlValue(normToggle))
				{	SetControlValue(normToggle, 0);
					SetControlValue(normGlobalToggle, 0);
					DeactivateControl(normGlobalToggle);
					Prefs->normalizeGlobal = false;
				} else
				{	SetControlValue(normToggle, 1);
					ActivateControl(normGlobalToggle);
				}
				Prefs->normalize = GetControlValue(normToggle);
			}
			
			else if (hitControl == normGlobalToggle)
			{	SetControlValue(normGlobalToggle, !GetControlValue(normGlobalToggle));
				Prefs->normalizeGlobal = GetControlValue(normGlobalToggle);
			}
			
			else if (hitControl == splitToggle)
			{	if (GetControlValue(splitToggle))
				{	SetControlValue(splitToggle, 0);
					ActivateControl(normToggle);
					ActivateControl(sampleRate);
					ActivateControl(sampleRateLabel);
				}
				else
				{	SetControlValue(splitToggle, 1);
					SetControlValue(normToggle, 0);
					SetControlValue(normGlobalToggle, 0);
					DeactivateControl(normToggle);
					DeactivateControl(normGlobalToggle);
					DeactivateControl(sampleRate);
					DeactivateControl(sampleRateLabel);
					Prefs->normalize = false;
					Prefs->normalizeGlobal = false;
				}
				
				Prefs->channelSplit = GetControlValue(splitToggle);
			}
		}
	}
}


void Decoder::AboutBox(void)
{
	static BitMap screenBits;
	Rect boundsRect = GetQDGlobalsScreenBits(&screenBits)->bounds;
	DialogPtr dlog;
	short theItem;

	dlog = GetNewDialog(rAlertNewAboutBox, nil, (WindowPtr) - 1);
	MoveWindow(GetDialogWindow(dlog), (boundsRect.right - boundsRect.left - 230) / 2, (boundsRect.bottom - boundsRect.top - 292) / 2, 0);
	SetDialogDefaultItem(dlog, 1);
	
	do
	{	ModalDialog(nil, &theItem);
		
		switch(theItem)
		{
			case 2:
				theItem = 1;
				Web("\phttps://www.paypal.com/cgi-bin/webscr?cmd=_xclick&business=lukifer@mail.com&noshipping=1");
				break;
			
			case 3:
				theItem = 1;
				Web("\phttp://s1.amazon.com/exec/varzea/pay/T1PTRDVCYXJQR4");
				break;
		}
	} while (theItem != 1);
	
	DisposeDialog(dlog);
}


void Decoder::Resize(short h, short v)
{
	long h2, v2;
	Rect invalRect = {0};
	
	SizeWindow(window, h, v, false);
	GetWindowBounds(window, kWindowContentRgn, &winRect);
	
	h2 = (winRect.right  - winRect.left) - (kRectWindowSize.right  - kRectWindowSize.left);
	v2 = (winRect.bottom - winRect.top)  - (kRectWindowSize.bottom - kRectWindowSize.top);

	// Resize list
	listRect.right	= kRectList.right  + h2;
	listRect.bottom	= kRectList.bottom + v2;
	LSize(listRect.right - listRect.left, listRect.bottom - listRect.top, list);

	// Resize controls
	SizeControl(progressBar, kRectProgressBar.right - kRectProgressBar.left + h2,
		kRectProgressBar.bottom - kRectProgressBar.top);

	SizeControl(separator, kRectSeparator.right - kRectSeparator.left + h2,
		kRectSeparator.bottom - kRectSeparator.top);
	
	SizeControl(decodeText, kRectDecodeText.right - kRectDecodeText.left + h2,
		kRectDecodeText.bottom - kRectDecodeText.top);
	
	SizeControl(detailsText, kRectDetailsText.right - kRectDetailsText.left + h2,
		kRectDetailsText.bottom - kRectDetailsText.top);
	
	// Move controls
	MoveControl(startButton,		kRectStartButton.left + h2,		kRectStartButton.top + v2);
	MoveControl(splitToggle,		kRectChannelSplit.left + h2,	kRectChannelSplit.top);
	MoveControl(normToggle,			kRectNormToggle.left + h2,		kRectNormToggle.top);
	MoveControl(normGlobalToggle,	kRectNormGToggle.left + h2,		kRectNormGToggle.top);
	MoveControl(wrapper,			kRectWrapper.left + h2,			kRectWrapper.top);
	MoveControl(sampleRate,			kRectSampleRate.left + h2,		kRectSampleRate.top);
	MoveControl(mp3Settings,		kRectMP3Settings.left + h2,		kRectMP3Settings.top);
	MoveControl(timeText,			kRectTimeText.left + h2,		kRectTimeText.top);
	MoveControl(sampleRateLabel,	kRectSampleRateText.left + h2,	kRectSampleRateText.top);
	MoveControl(wrapperLabel,		kRectWrapperText.left + h2,		kRectWrapperText.top);
	
	// Redraw
	invalRect.right  = winRect.right  - winRect.left;
	invalRect.bottom = winRect.bottom - winRect.top;
	SetPortWindowPort(window);
	EraseRect(&invalRect);
	InvalWindowRect(window, &invalRect);
	DrawControls(window);
}


void Decoder::Details(int sampleIn, int sampleOut, int channelsIn, int channelsOut, Boolean fullInfo)
{
	Str255 theString;
	Cell theCell = {0, 0};
	Boolean tempBool = Prefs->channelSplit;
	int audioLength = 0; // in ticks
	long long fileSize = 0;

	// Clear time field
	PLstrcpy(theString, "\p");
	SetControlData(timeText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 		

	if (fullInfo == false && LGetSelect(true, &theCell, list))
	{	theCell.v++;
		if (!(LGetSelect(true, &theCell, list)) && Prefs->useRefs)
		{	if (CheckQT(&items[theCell.v-1]))
			{	PLstrcpy(theString, "\pQuickTime File");
				SetControlData(detailsText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
				DrawControls(window);
	 			return;
	 		}
			if (noErr == FSpMakeFSRef(&items[theCell.v-1], &inputBig))
			{	if (noErr == FSOpenFork(&inputBig, 0, 0, fsRdPerm, &inputRef))
				{
					// AC-3 audio length
					int originalState = state; // I'm a cheater, I am
					state = kStateDecoding;
					bit_rate = 0;
					while(0 >= SeekSync(&flags, &sample_rate, &bit_rate))
					{	if (WaitNextEvent(everyEvent, &Event, 1, nil))
							HandleEvent(&Event);
					}
					state = originalState;

					FSGetForkSize(inputRef, &fileSize);

					//HandleError("\pBit Rate", bit_rate, false);
					//HandleError("\pFileSize", fileSize, false);

					if (bit_rate > 0)
					{	audioLength = (fileSize * 8 * 60) / bit_rate; //((fileSize / 1024) * 8 * 60) / (bit_rate / 1000);
						TicksToTime(audioLength, theString, false);
						SetControlData(timeText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
					}
					
					sampleIn = sample_rate;
					channelsIn = flags & A52_CHANNEL_MASK;
		}	}	}
		else
		{	PLstrcpy(theString, "\p");
			SetControlData(detailsText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
			DrawControls(window);
 			return;
 		}
		
		if (Prefs->useRefs)	FSCloseFork(inputRef);
		else				FSClose(input);
	}
	else if (fullInfo == false)
	{	PLstrcpy(theString, "\p");
		SetControlData(detailsText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
		DrawControls(window);
 		return;
 	}
 	
 	if (qt)
 		PLstrcpy(theString, "\pQuickTime");
 	else
	{	switch (sampleIn)
		{	case 32000: PLstrcpy(theString, "\p32kHz "); break;
			case 44100: PLstrcpy(theString, "\p44.1kHz "); break;
			case 48000: PLstrcpy(theString, "\p48kHz "); break;
			default: PLstrcpy(theString, "\p"); break;
		}
	
		switch (channelsIn)
		{	case A52_3F2R:		if (flags & A52_LFE)
								PLstrcat(theString, "\p5.1ch AC-3");
						else	PLstrcat(theString, "\p5ch AC-3");
								break;
			case A52_2F2R:
			case A52_3F1R:		if (flags & A52_LFE)
								PLstrcat(theString, "\p4.1ch AC-3");
						else	PLstrcat(theString, "\p4ch AC-3");
								break;
			case A52_2F1R:
			case A52_3F:		if (flags & A52_LFE)
								PLstrcat(theString, "\p3.1ch AC-3");
						else	PLstrcat(theString, "\p3ch AC-3");
								break;
			case A52_STEREO:	PLstrcat(theString, "\pStereo AC-3"); break;
			case A52_DOLBY:		PLstrcat(theString, "\pDolby Stereo AC-3"); break;
			case A52_MONO:		PLstrcat(theString, "\pMono AC-3"); break;
			case A52_CHANNEL:	PLstrcat(theString, "\pDual Mono AC-3"); break;
			default: PLstrcat(theString, "\pAC-3"); break;
	}	}
	
	if (fullInfo)
	{	if (Prefs->wrapper <= kWrapperPCM)
		{	switch (sampleOut)
			{	case rate44khz:		PLstrcat(theString, "\p  ->  44.1kHz "); break;
				case rate48khz:		PLstrcat(theString, "\p  ->  48kHz "); break;
				case rate32khz:		PLstrcat(theString, "\p  ->  32kHz "); break;
				case rate22050hz:	PLstrcat(theString, "\p  ->  22050Hz "); break;
				case rate22khz:		PLstrcat(theString, "\p  ->  22kHz "); break;
				case rate16khz:		PLstrcat(theString, "\p  ->  16kHz "); break;
				case rate11khz:		PLstrcat(theString, "\p  ->  11kHz "); break;
				case rate11025hz:	PLstrcat(theString, "\p  ->  11025Hz "); break;
				case rate8khz:		PLstrcat(theString, "\p  ->  8kHz "); break;
				default:			PLstrcat(theString, "\p  ->  ");
		}	}
		else
		{	switch (Prefs->mp3info.sample)
			{	case 44100:		PLstrcat(theString, "\p  ->  44.1kHz "); break;
				case 48000:		PLstrcat(theString, "\p  ->  48kHz "); break;
				case 32000:		PLstrcat(theString, "\p  ->  32kHz "); break;
				case 24000:		PLstrcat(theString, "\p  ->  24kHz "); break;
				case 22050:		PLstrcat(theString, "\p  ->  22kHz "); break;
				case 16000:		PLstrcat(theString, "\p  ->  16kHz "); break;
				case 11025:		PLstrcat(theString, "\p  ->  11kHz "); break;
				case 80000:		PLstrcat(theString, "\p  ->  8kHz "); break;
				default:		PLstrcat(theString, "\p  ->  ");
		}	}
	
		if (Prefs->wrapper == kWrapperAIFF)
		{	switch (channelsOut)
			{	case 2: PLstrcat(theString, "\pStereo AIFF"); break;
				case 1: PLstrcat(theString, "\pMono AIFF"); break;
				default: PLstrcat(theString, "\pAIFF"); break;
		}	}
		else if (Prefs->wrapper == kWrapperWAV)
		{	switch (channelsOut)
			{	case 2: PLstrcat(theString, "\pStereo WAV"); break;
				case 1: PLstrcat(theString, "\pMono WAV"); break;
				default: PLstrcat(theString, "\pWAV"); break;
		}	}
		else if (Prefs->wrapper == kWrapperPCM)
		{	switch (channelsOut)
			{	case 2: PLstrcat(theString, "\pStereo PCM"); break;
				case 1: PLstrcat(theString, "\pMono PCM"); break;
				default: PLstrcat(theString, "\pPCM"); break;
		}	}
		else if (Prefs->wrapper == kWrapperMP3)
		{	Str31 tempstr;

				 if (Prefs->mp3info.joint)	PLstrcat(theString, "\pJStereo ");
			else if (Prefs->mp3info.mono)	PLstrcat(theString, "\pMono ");
			else							PLstrcat(theString, "\pStereo ");
			PLstrcat(theString, "\pMP3 (");

			if (Prefs->mp3info.vbr)
			{	PLstrcat(theString, "\pVBR:");
				NumToString(Prefs->mp3info.vbrQuality, tempstr);
				PLstrcat(theString, tempstr);
				PLstrcat(theString, "\p)");
			}
			else
			{	NumToString(Prefs->mp3info.bitrate, tempstr);
				PLstrcat(theString, tempstr);
				PLstrcat(theString, "\p kb");
				if (Prefs->mp3info.abr) PLstrcat(theString, "\p ABR");
				PLstrcat(theString, "\p)");
		}	}
		else if (Prefs->wrapper == kWrapperMOV)
		{	switch (channelsOut)
			{	case 2: PLstrcat(theString, "\pStereo QuickTime"); break;
				case 1: PLstrcat(theString, "\pMono QuickTime"); break;
				default: PLstrcat(theString, "\pQuickTime"); break;
	}	}	}

	SetControlData(detailsText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	DrawControls(window);
}


void Decoder::Feedback(Boolean norm)
{
	Str255 theString;

		 if (norm)						PLstrcpy(theString, "\pNormalizing ");
	else if (Prefs->wrapper <= kWrapperPCM)	PLstrcpy(theString, "\pDecoding ");
	else								PLstrcpy(theString, "\pTranscoding ");
	PLstrcat(theString, inputSpec.name);
	PLstrcpy(titleText, theString);
	
	SetControlData(decodeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);

	if (!norm)
	{	int temp = flags & A52_CHANNEL_MASK;
		int channels = temp;
		/*switch (temp)
		{	case A52_MONO:
			case A52_STEREO:
			case A52_3F:		channels = temp;	break;
			case A52_2F1R:		channels = 3;		break;
			case A52_3F1R:
			case A52_2F2R:		channels = 4;		break;
			case A52_3F2R: 		channels = 5;		break;
		}*/
		if (qt)		Details(0,	info.outputFormat.sampleRate, 2, info.outputFormat.numChannels, true);
		else 		Details(	sample_rate, 					info.outputFormat.sampleRate,
								channels,			 			info.outputFormat.numChannels, true);
	}

	DrawControls(window);
}


void Decoder::Time(unsigned long theTicks, long amountDone)
{
	Str31 theString;
	unsigned long ticksLeft;
	float val;
	
	if (amountDone < 25)
	{	DrawControls(window);
		return;
	}
	
	if (amountDone >= kProgressBarMax)
	{	PLstrcpy(theString, "\p");
		DrawControls(window);
		return;
	}
	
	val = theTicks;
	ticksLeft = (long) (((kProgressBarMax * val) / amountDone) - theTicks);
	
	TicksToTime(ticksLeft, theString, false);
	SetControlData(timeText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	

	DrawControls(window);
}



#pragma mark -

void Decoder::Add(FSSpec* file)
{
	Point cell;
	
	if (numItems < 1)
	{	ActivateControl(startButton);
		EnableMenuItem(DecodeMenu, rMenuDecodeStart);
		numItems = 0; // reduncancy, reduncancy, reduncancy... reduncancy
	}

	if (numItems >= 1024)
	{	HandleError("\pIf you are trying to decode over 1024 files, you are truly sick in the head.", noErr, false);
		return;
	}

	// Add to list
	LAddRow(1, numItems, list);
	SetPt(&cell, 0, numItems);
	LSetCell(file->name+1, file->name[0], cell, list);
	
	// Allocate memory and store
	BlockMoveData(file, &items[numItems], sizeof(FSSpec));
	numItems++;
}


void Decoder::Remove(short itemNum)
{
	if (itemNum > -1)
	{	LDelRow(1, itemNum, list);

		// Move up to fill in gap
		for (int i = itemNum; i < (numItems - 1); i++)
			BlockMoveData(&items[i+1], &items[i], sizeof(FSSpec));

		if (numItems > 0) numItems--;
	}
	
	if (numItems < 1)
	{	DeactivateControl(startButton);
		DisableMenuItem(DecodeMenu, rMenuDecodeStart);
	}
}


void Decoder::GetFolderWithNav(FSSpec* theSpec, Str255 message)
{
	// Used to choose a destination folder for output files
	
	OSErr theErr;
	NavReplyRecord reply;
	NavDialogOptions dialogOptions;
	AEDesc resultDesc;
    NavEventUPP eventProc = NewNavEventUPP(NavServicesSuxEventProc);
	
	NavGetDefaultDialogOptions(&dialogOptions);
	dialogOptions.dialogOptionFlags |= kNavAllFiles;
	dialogOptions.dialogOptionFlags |= kNavNoTypePopup;
	dialogOptions.dialogOptionFlags ^= kNavAllowPreviews;

	BlockMoveData("\pmAC3dec", dialogOptions.clientName, 8);
	BlockMoveData(message, dialogOptions.message, message[0]+1);
	theErr = NavChooseFolder(nil, &reply, &dialogOptions, eventProc, nil, nil);
	DisposeNavEventUPP(eventProc);
	
	if (theErr == noErr && reply.validRecord)
	{	AEKeyword	theKeyword;
		
		theErr = AEGetNthDesc(&(reply.selection), 1, typeFSS, &theKeyword, &resultDesc);
			if (theErr != noErr) {HandleError("\pFailed to extract file from Apple Event", theErr, true); return; }
		
		AEGetDescData(&resultDesc, theSpec, sizeof(FSSpec));
	}

	NavDisposeReply(&reply);
	return;
}


void Decoder::GetFileWithNav(void)
{
	OSErr theErr = noErr;
	NavReplyRecord theReply;
	NavDialogOptions dialogOptions;
	NavTypeListHandle openTypeList = NULL;
	NavEventUPP eventProc = NewNavEventUPP(NavServicesSuxEventProc);
	AEDesc resultDesc;
	AEKeyword aKeyWord;
	FSSpec inputFile;
	long count, index;

	theErr = NavGetDefaultDialogOptions(&dialogOptions);
	dialogOptions.dialogOptionFlags |= kNavAllFiles;
	dialogOptions.dialogOptionFlags |= kNavNoTypePopup;
	dialogOptions.dialogOptionFlags ^= kNavAllowPreviews;

	BlockMoveData("\pmAC3dec", dialogOptions.clientName, 8);
	BlockMoveData("\pDecode AC-3", dialogOptions.windowTitle, 12);
	openTypeList = nil; // User can force-feed any file type they like
	theErr = NavChooseFile(nil, &theReply, &dialogOptions, nil, nil, nil, openTypeList, nil);

	if (theReply.validRecord && theErr == noErr)
	{
		theErr = AECoerceDesc(&theReply.selection, typeFSS, &resultDesc);
		
		if (theErr == errAECoercionFail) theErr = AECoerceDesc(&theReply.selection, typeAEList, &resultDesc);

		if (theErr == noErr)
		{		
			if ((theErr = AECountItems(&(theReply.selection), &count)) == noErr)
				for (index=1; index<=count; index++ )
					if ((theErr = AEGetNthDesc(&(theReply.selection),index, typeFSS, &aKeyWord, &resultDesc)) == noErr)
					{
						theErr = AEGetDescData(&resultDesc, &inputFile, sizeof(FSSpec));
							if (theErr != noErr) HandleError("\pCould not get data from Apple Event", theErr, true);
						
						Add(&inputFile);

						AEDisposeDesc(&resultDesc);
					}
		}
	}

	NavDisposeReply(&theReply);
	DisposeNavEventUPP(eventProc);
}


Boolean Decoder::PutFileWithNav(StringPtr theOutputString, FSSpec* theFile, short fileType)
{
    OSErr theErr = noErr;
    NavReplyRecord reply;
    NavDialogOptions dialogOptions;
    OSType fileTypeToSave = 'AIFF';
    OSType creatorType = 'TVOD';
    NavEventUPP eventProc = NewNavEventUPP(NavServicesSuxEventProc);
    AEDesc resultDesc;

    theErr = NavGetDefaultDialogOptions(&dialogOptions);
    dialogOptions.dialogOptionFlags ^= kNavAllowStationery;

    if (theErr == noErr)
    {
		BlockMoveData(theOutputString, dialogOptions.savedFileName, theOutputString[0]+1);
		BlockMoveData("\pmAC3dec", dialogOptions.clientName, 8);
		if (fileType == kWrapperAIFF)
	       	BlockMoveData("\pSave AIFF File As...", dialogOptions.message, 21);
	    else if (fileType == kWrapperWAV)
	    {   BlockMoveData("\pSave WAV Audio File As...", dialogOptions.message, 26);
			fileTypeToSave = 'WAVE';
		}	    	
	    else if (fileType == kWrapperPCM)
	    {   BlockMoveData("\pSave PCM Audio File As...", dialogOptions.message, 26);
			fileTypeToSave = 'PCM ';
			creatorType = '????';		
		}	    	
		else if (fileType == kWrapperMP3)
		{	BlockMoveData("\pSave MP3 File As...", dialogOptions.message, 20);
			fileTypeToSave = 'MP3 ';
		}
		else if (fileType == kWrapperMOV)
		{	BlockMoveData("\pSave QuickTime File As...", dialogOptions.message, 29);
			fileTypeToSave = 'MooV';
		}
	    else if (fileType == kWrapperAC3)
	    {	BlockMoveData("\pSave Fixed AC-3 File As...", dialogOptions.message, 27);	    	
			fileTypeToSave = 'AC-3';
			creatorType = 'mAC3';			
		}

		theErr = NavPutFile(nil, &reply, &dialogOptions, eventProc, fileTypeToSave, creatorType, nil);
        
        if (theErr == noErr && reply.validRecord)
        {
            AEKeyword   theKeyword;
                        
            theErr = AEGetNthDesc(&(reply.selection), 1, typeFSS, &theKeyword, &resultDesc);
            	if (theErr != noErr) HandleError("\pFailed to, um, do some things", theErr, true);

            AEGetDescData(&resultDesc, theFile, sizeof(FSSpec));

			if (reply.replacing == false)
			{	theErr = FSpCreate(theFile, creatorType, fileTypeToSave, smSystemScript);
					if (theErr != 0) HandleError("\pFSpCreate failed on Save", theErr, true);
			}
			else
			{	short refNum;
				theErr = FSpOpenDF(theFile, fsWrPerm, &refNum);
					if (theErr != 0) HandleError("\pCould not open data fork of output file", theErr, true);
				theErr = (SetEOF(refNum, 0));
					if (theErr != noErr) HandleError("\pSetEOF Failed", theErr, true);
				FSClose(refNum);
			}
            
            NavCompleteSave(&reply, kNavTranslateInPlace);
	        NavDisposeReply(&reply);
        }
        else
        {	NavDisposeReply(&reply);
        	return false; // If user cancelled then [clean up memory and*] bail				* Thank you Stux
		}
    }    
    
    return true;
}


void Decoder::GetName(StringPtr theString, int fileType)
{
	// theFileType == -1: We already have a file by that name
	// theFileType == 0: AIFF Output
	// theFileType == 1: WAV Output
	// theFileType == 2: PCM Output
	// theFileType == 4: MP3 Output
	// theFileType == 5: OGG Output
	// theFileType == 9: AC-3 fixed Output
	// theFileType == 10: Left Channel
	// theFileType == 11: Right Channel
	// theFileType == 12: Center Channel
	// theFileType == 13: Left Surround
	// theFileType == 14: Right Surround
	// theFileType == 15: LFE

	int j, k, len;
	long tempNum = 1;
	Str31 tempStr;
	Ptr tempPtr;
	
	// If this isn't a duplicate file, remove all numbers and extensions from the name
	if (fileType != -1)
	{	for (k = theString[0]; k > 0; k--)
		{	if (theString[k] == ' ')
			{	k--;	theString[0] = k;	}
			if (theString[k] < '0' || theString[k] > '9')
				break;
		}
		
		tempPtr = PLstrstr(theString, "\p.AC3");
		if (tempPtr == NULL)
			tempPtr = PLstrstr(theString, "\p.ac3");
		if (tempPtr != NULL)
			theString[0] -= 4;
		
		tempPtr = PLstrstr(theString, "\p.MP3");
		if (tempPtr == NULL)
			tempPtr = PLstrstr(theString, "\p.mp3");
		if (tempPtr != NULL)
			theString[0] -= 4;
		
		tempPtr = PLstrstr(theString, "\p.MOV");
		if (tempPtr == NULL)
			tempPtr = PLstrstr(theString, "\p.mov");
		if (tempPtr != NULL)
			theString[0] -= 4;
		
		tempPtr = PLstrstr(theString, "\p.WAV");
		if (tempPtr == NULL)
			tempPtr = PLstrstr(theString, "\p.wav");
		if (tempPtr != NULL)
			theString[0] -= 4;
		
		tempPtr = PLstrstr(theString, "\p.AIFF");
		if (tempPtr == NULL)
			tempPtr = PLstrstr(theString, "\p.aiff");
		if (tempPtr != NULL)
			theString[0] -= 5;
			
		tempPtr = PLstrstr(theString, "\p.AIF");
		if (tempPtr == NULL)
			tempPtr = PLstrstr(theString, "\p.aif");
		if (tempPtr != NULL)
			theString[0] -= 4;
	}
	
	// Generate the name
	switch (fileType)
	{
		case -1:
			// Search for a space followed by numbers
			for (k = theString[0]; (theString[k] >= '0' && theString[k] <= '9') || k <= 0; k--)
			{}
			
			if (theString[k] != ' ')
			{	// This is the first duplicate file; add a 1
				if (theString[0] > 29)	theString[0] = 29;
			
				PLstrcat(theString, "\p 1");			
			}
			else
			{	// We have a number file, increment by 1
				k++; tempStr[0] = 0;
				for (j = 1, len = theString[0]; k <= len; k++, j++)
				{	tempStr[0]++;
					tempStr[j] = theString[k];
					theString[0]--;
				}

				StringToNum(tempStr, &tempNum);
				tempNum++;
				NumToString(tempNum, tempStr);
				
				if (tempStr[0] + theString[0] > 31)
					theString[0] -= (tempStr[0] + theString[0] - 31);

				PLstrcat(theString, tempStr);
			}
			break;
			
		case kWrapperAIFF:
			if (theString[0] > 26)
				theString[0] = 26;
			
			PLstrcat(theString, "\p.aiff");
			break;
		
		case kWrapperWAV:
			if (theString[0] > 27)
				theString[0] = 27;
			
			PLstrcat(theString, "\p.wav");	
			break;
		
		case kWrapperPCM:
			if (theString[0] > 27)
				theString[0] = 27;
			
			PLstrcat(theString, "\p.pcm");	
			break;

		case kWrapperMP3:
			if (theString[0] > 27)
				theString[0] = 27;
			
			PLstrcat(theString, "\p.mp3");	
			break;
		
		case kWrapperMOV:
			if (theString[0] > 27)
				theString[0] = 27;
			
			PLstrcat(theString, "\p.mov");	
			break;
	
		case kWrapperAC3:
			if (theString[0] > 21)
				theString[0] = 21;

			PLstrcat(theString, "\p fixed.ac3");
			break;

		case 10 + kLeft:
			if (theString[0] > 24)
				theString[0] = 24;
			
			PLstrcat(theString, "\p L");		
			break;

		case 10 + kRight:
			if (theString[0] > 24)
				theString[0] = 24;
			
			PLstrcat(theString, "\p R");
			break;

		case 10 + kLeftS:
			if (theString[0] > 23)
				theString[0] = 23;
			
			PLstrcat(theString, "\p LS");
			break;

		case 10 + kRightS:
			if (theString[0] > 23)
				theString[0] = 23;
			
			PLstrcat(theString, "\p RS");
			break;

		case 10 + kCenter:
			if (theString[0] > 24)
				theString[0] = 24;
			
			PLstrcat(theString, "\p C");
			break;
			
		case 10 + kLFE:
			if (theString[0] > 22)
				theString[0] = 22;
			
			PLstrcat(theString, "\p LFE");
			break;

		default:
			SetString(&theString, "\pUntitled");
			break;
	}
}


void Decoder::GetInputFile(int num)
{
	OSErr theErr;
	Str255 fileName;

	// Get the input file
	BlockMoveData(&items[num], &inputSpec, sizeof(FSSpec));
	BlockMoveData(inputSpec.name, fileName, inputSpec.name[0]+1);

	// Open
	if (Prefs->useRefs)
	{	// Convert to a 2GB+ happy FSRef
		theErr = FSpMakeFSRef(&inputSpec, &inputBig);
			if (theErr != noErr) HandleError("\pCould not convert to FSRef", theErr, true);
		if (inputRef != 0) FSCloseFork(inputRef);
		theErr = FSOpenFork(&inputBig, 0, 0, fsRdPerm, &inputRef);
			if (theErr != noErr) HandleError("\pCould not open data fork of input file", theErr, true);
	}
	else
	{	theErr = FSpOpenDF(&inputSpec, fsRdPerm, &input);
	}

	// Get input file size (for use by progress bar)
	if (Prefs->useRefs)	theErr = FSGetForkSize(inputRef, &inputFileSize);
	else {long zerk; theErr = GetEOF(input, &zerk); inputFileSize = (long long) zerk;}
		if (theErr != 0) HandleError("\pGetEOF Failed", theErr, true);
	inputFileSize /= kProgressBarMax; // Shrink by the size of the Progress Bar
}


void Decoder::GetOutputFile(Str63 theOutputString)
{
	OSErr theErr = noErr;
	long outputFileSize;
	OSType type, creator;

	aiffEOF = 0;

	if (Prefs->chooseSave == false) // Auto-Save
	{	do
		{	switch(Prefs->wrapper)
			{
				case kWrapperAIFF:
					type = 'AIFF';	creator = 'TVOD';	break;
				
				case kWrapperWAV:
					type = 'WAVE';	creator = 'TVOD';	break;
				
				case kWrapperPCM:
					type = 'PCM ';	creator = '????';	break;
				
				case kWrapperMP3:
					type = 'MPG3';	creator = 'TVOD';	break;
				
				case kWrapperMOV:
					type = 'MooV';	creator = 'TVOD';	break;
				
				default:
					type = creator = '????';			break;
			}
			
			FSMakeFSSpec(inputSpec.vRefNum, inputSpec.parID, theOutputString, &outputSpec);
			
			theErr = FSpCreate(&outputSpec, creator, type, smSystemScript);
			
			if (theErr == vLckdErr || theErr == wrPermErr || theErr == dirNFErr)
			{	if (folderSpec.name[0] == 0)
					GetFolderWithNav(&folderSpec, "\pYour desination folder is locked. Save files where?");
				
				FSMakeFSSpec(folderSpec.vRefNum, folderSpec.parID, theOutputString, &outputSpec);

				theErr = FSpCreate(&outputSpec, creator, type, smSystemScript);
			}
			
			if (theErr == dupFNErr)
			{
				OSErr theErrTwo = noErr; // so the value of theErr does not get corrupted
				
				// If file is less than 55 bytes (either 0 or just the AIFF header) we can overwrite it.
				FSClose(output);
				theErrTwo = FSpOpenDF(&outputSpec, fsWrPerm, &output);
				outputFileSize = 1024;
				if (theErrTwo == noErr)
				{	GetEOF(output, &outputFileSize);
					theErrTwo = FSClose(output);
						if (theErrTwo != noErr) HandleError("\pCould not close data fork of file to overwrite", theErrTwo, true);	
				}
				if (outputFileSize < 55)
				{	theErrTwo = FSpDelete(&outputSpec);
						if (theErrTwo != noErr) HandleError("\pCould not overwrite existing output file", theErrTwo, true);
				
							 if (Prefs->wrapper == kWrapperAIFF)	theErr = FSpCreate(&outputSpec, 'TVOD', 'AIFF', smSystemScript);
						else if (Prefs->wrapper == kWrapperWAV)		theErr = FSpCreate(&outputSpec, 'TVOD', 'WAVE', smSystemScript);
						else if (Prefs->wrapper == kWrapperPCM)		theErr = FSpCreate(&outputSpec, '????', 'PCM ', smSystemScript);
						else if (Prefs->wrapper == kWrapperMP3)		theErr = FSpCreate(&outputSpec, 'TVOD', 'MPG3', smSystemScript);
						else if (Prefs->wrapper == kWrapperMOV)		theErr = FSpCreate(&outputSpec, 'TVOD', 'MooV', smSystemScript);
						else										theErr = FSpCreate(&outputSpec, '????', '????', smSystemScript);

					if (theErrTwo != noErr && theErrTwo != dupFNErr)
						HandleError("\pCould not create output file", theErrTwo, true);
					else
					{	// Overwrite the worthless file and zero it out.
						theErrTwo = FSpOpenDF(&outputSpec, fsWrPerm, &output);
							if (theErrTwo != noErr) HandleError("\pCould open data fork of output file", theErrTwo, true);
						theErrTwo = SetEOF(output, 0);
							if (theErrTwo != noErr) HandleError("\pCould not zero out output file", theErrTwo, true);
						FSClose(output);
					}
				}
			}
			else if (theErr != noErr)
				HandleError("\pCould not create output file", theErr, true);
			
			GetName(theOutputString, -1); // This is called even if we don't need it; this is OK, since we only use the results before this point.
		} while(theErr == dupFNErr);
	}
	else // Choose Save
	{	if (!PutFileWithNav(theOutputString, &outputSpec, Prefs->wrapper))
		{	state = kStateCancel;
			return; // If user cancelled, time to leave.
	}	}

	// Open Output File
	if (output	 != 0) FSClose    (output);
	if (!Prefs->channelSplit)
	{	theErr = FSpOpenDF(&outputSpec, fsWrPerm, &output);
			if (theErr != noErr) HandleError("\pCould not open data fork of output file", theErr, true);
	}
}

/*
** THIS IS WHERE I LEFT OFF: 2/3/03
*/


#pragma mark -

void Decoder::Start(void)
{
	Point	theCell = {0, 0};
	Str255	theString;

	/*	If something is currently happening, simply switch to idle, and processes
		should be smart enough to do any cleanup work on their own */
		
	if (state != kStateIdle)
	{	state  = kStateCancel;
		return;
	}
						
	// Otherwise, start decoding
	state = kStateDecoding;
	folderSpec.name[0] = 0;

	// Alter controls to our liking
	SetControlTitle(startButton, "\pCancel");
	SetMenuItemText(DecodeMenu, rMenuDecodeStart, "\pCancel");
	EnableMenuItem(DecodeMenu, rMenuDecodePause);
	DeactivateControl(sampleRate);
	DeactivateControl(wrapper);
	DeactivateControl(normToggle);
	DeactivateControl(normGlobalToggle);
	DeactivateControl(splitToggle);
	DeactivateControl(mp3Settings);
	DisableMenuItem(FileMenu,	rMenuFileRemove);
	DisableMenuItem(FileMenu,	rMenuFilePrefs);
	DisableMenuItem(FileMenu,	rMenuFileMP3Settings);
	DisableMenuItem(DecodeMenu, rMenuDecodeFix);

	// Global Normalize
	normGrow = 32767.0f;
	if (Prefs->normalize && Prefs->normalizeGlobal)
		normGrow = NormalizeGlobal();	
	
	// Decode Loop Begin
	while (numItems > 0 && state != kStateCancel)
	{
		// Non-global Normalize
		if (Prefs->normalize && !Prefs->normalizeGlobal)
			normGrow = Normalize(0);

		if (state == kStateCancel) break;

		// Transcode
		Transcode(0);
		
		// Remove
		Remove(0);
	}

	// Interface Cleanup
	state = kStateIdle;
	SetControlTitle(startButton, "\pStart");
	SetMenuItemText(DecodeMenu, rMenuDecodeStart, "\pStart");
	PLstrcpy(theString, "\p");
	SetControlData(amountText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(detailsText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(timeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(decodeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlValue(progressBar, 0);
	ActivateControl(sampleRate);
	ActivateControl(wrapper);
	ActivateControl(normToggle);
	ActivateControl(normGlobalToggle);
	ActivateControl(splitToggle);
	ActivateControl(mp3Settings);
	DrawControls(window);
	EnableMenuItem(FileMenu,	rMenuFilePrefs);
	EnableMenuItem(FileMenu,	rMenuFileMP3Settings);
	DisableMenuItem(DecodeMenu, rMenuDecodePause);
	if (LGetSelect(true, &theCell, list))
	{	EnableMenuItem(FileMenu,	rMenuFileRemove);
		EnableMenuItem(DecodeMenu,	rMenuDecodeFix);
	}
	
	state = kStateIdle;
	folderSpec.name[0] = 0;
}


void Decoder::Transcode(int itemNum)
{
	OSErr theErr = noErr;
	Str255 theString;
	unsigned long actualOutputBytes, outputFlags;

	// Init and get files
	state = kStateDecoding;
	normPass = false;

	// File I/O
	GetInputFile(itemNum);
	qt = CheckQT(&inputSpec);
	if (Prefs->channelSplit)
	{	if (qt)
		{	FSClose(input);
			return;
		}
		SplitAC3();
		return;
	}
	PLstrcpy(theString, inputSpec.name);
	GetName(theString, Prefs->wrapper);
	GetOutputFile(theString);

	// Leave if user cancelled or error occurred
	if (state <= kStateIdle)
		return;
	
	// Initialize audio input
	if (qt)		InitInputQT();
	else		InitInputAC3();

	// Leave if user cancelled or error occurred
	if (state <= kStateIdle)
		return;

	// Initialize audio output
	switch(Prefs->wrapper)
	{
		case kWrapperAIFF:
		case kWrapperWAV:
		case kWrapperPCM:	InitOutputAIFF();					break;			
		case kWrapperMP3:	InitOutputMP3(&Prefs->mp3info);		break;
		case kWrapperMOV:	InitOutputMOV();					break;

		default: HandleError("\pI'm afraid I can't do that, Dave.", Prefs->wrapper, false); break;
	}
	Feedback(false);

	// Leave if user cancelled or error occurred
	if (state <= kStateIdle)
		return;	

	// Transcode Begin
	ticksStart = TickCount();

	// Loop and Transcode All Frames (All work is now handled by SoundConverterFillBuffer and its FillBufferProc)
	while(state > kStateIdle)
	{	
		actualOutputBytes = 0;
		theErr = SoundConverterFillBuffer(	info.outputSC,		// a sound converter
											fillBufferUPP,		// the callback UPP
											this,				// refCon passed to FillDataProc
											decomBuffer,		// the destination data buffer
											info.outputBytes,	// size of the destination buffer
											&actualOutputBytes,	// number of output bytes
									   &info.actualOutputFrames,// number of output frames
											&outputFlags);		// FillBuffer retured advisor flags

		if (theErr != noErr)
		{	HandleError("\pI'll be damned. This is funny.", theErr, false);		break;	}
		//if ((outputFlags & kSoundConverterHasLeftOverData) == false)
		if ((outputFlags & kSoundConverterHasLeftOverData) == false &&
			(actualOutputBytes == 0))
			state = kStateDone;
		
		// Write Frame; this is now where MP3 encoding takes place
		if (actualOutputBytes > 0)
			WriteFrame((unsigned char*) decomBuffer, actualOutputBytes);
	}

	// SoundConverter Cleanup
	theErr = SoundConverterEndConversion(info.outputSC, decomBuffer, &info.actualOutputFrames, &info.actualOutputSize);
	if (info.actualOutputSize > 0) WriteFrame((unsigned char*) decomBuffer, info.actualOutputSize);
				
	// Take a Wild Guess
	Cleanup();
}


void Decoder::FixAC3(void)
{
	OSErr theErr = noErr;
	FSSpec fixedFSSpec;
	short fixed;
	Str255 outputString, theString;
	Cell theCell = {0, 0};
	
	state = kStateDecoding;

	while(LGetSelect(true, &theCell, list) && state > kStateIdle)
	{
		// Open Data Fork of input file
		BlockMoveData(&items[theCell.v], &inputSpec, sizeof(FSSpec));
		theErr = FSpMakeFSRef(&inputSpec, &inputBig);
			if (theErr != noErr) HandleError("\pCould not convert to FSRef", theErr, true);
		theErr = FSOpenFork(&inputBig, 0, 0, fsRdPerm, &inputRef);
			if (theErr != noErr) HandleError("\pCould not open data fork of input file", theErr, true);

		BlockMoveData(inputSpec.name, outputString, inputSpec.name[0]+1);
		BlockMoveData(&inputSpec, &fixedFSSpec, sizeof(FSSpec));
		GetName(outputString, kWrapperAC3);

		// ac3fix setup
		if (PutFileWithNav(outputString, &fixedFSSpec, kWrapperAC3) == false)
			break;
		theErr = FSpOpenDF(&fixedFSSpec, fsWrPerm, &fixed);
			if (theErr != noErr) HandleError("\pCould not open data fork of output file", theErr, true);
		
		PLstrcpy(theString, "\p");
		SetControlData(timeText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
		SetControlData(detailsText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
		PLstrcpy(theString, "\pFixing ");
		PLstrcat(theString, inputSpec.name);
		SetControlData(decodeText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
		SetControlTitle(startButton, "\pCancel");
		SetMenuItemText(DecodeMenu, rMenuDecodeStart, "\pCancel");
		DrawControls(window);
		
		// Fix the AC-3
		checkFile(inputRef, fixed);
		
		theCell.v++;
	}
	
	// ac3fix cleanup
	PLstrcpy(theString, "\p");
	SetControlData(amountText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	SetControlData(timeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	SetControlData(decodeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	SetControlData(detailsText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	SetControlValue(progressBar,	0);
	state = kStateIdle;
	
	SetControlTitle(startButton, "\pStart");
	SetMenuItemText(DecodeMenu, rMenuDecodeStart, "\pStart");

	DrawControls(window);
}


float Decoder::Normalize(int itemNum)
{
	GetInputFile(itemNum);
	qt = CheckQT(&inputSpec);

	if (qt)		return NormalizeQT();
	else		return NormalizeAC3();
}


float Decoder::NormalizeAC3(void)
{
	OSErr theErr;
	Str255	theString;
	unsigned long outputFrames, actualOutputBytes, outputFlags;

	// Init values
	normPass = true;
	normVal  = 0;

	// Setup
	InitInputAC3();

	// Setup SoundConverter
	info.outputFormat.sampleSize	= 16;
	info.outputFormat.format		= k16BitBigEndianFormat;
	info.outputFormat.sampleRate	= info.inputFormat.sampleRate;
	info.outputFormat.numChannels	= info.inputFormat.numChannels;
	info.outputFormat.buffer		= 0L;
	info.outputFormat.sampleCount	= 0;
	info.outputFormat.reserved		= 0;
	info.outputFormat.flags 		= kNoRealtimeProcessing;
	
	// Open SounderConverter (Duh)
	theErr = SoundConverterOpen(&info.inputFormat, &info.outputFormat, &info.outputSC); // Open SoundConverter (Duh)
		if (theErr != noErr) HandleError("\pCould not open Sound Converter", theErr, true);

	// Allocate memory
	info.outputBytes = 6144;

	decomBuffer	= NewPtr(info.outputBytes);
		if (MemError()) HandleError("\pOut of Memory", MemError(), true);

	// Initialize Total Bytes to Zero
	info.totalBytesWritten = 0; 

	// Init converters
	theErr = SoundConverterBeginConversion(info.outputSC);
		if (theErr != noErr) HandleError("\pCould not begin sound conversion", theErr, true);

	// Begin
	ticksStart = TickCount();
	state = kStateDecoding;
	Feedback(true);

	while (state > kStateIdle)
	{
		actualOutputBytes = 0;
		theErr = SoundConverterFillBuffer(	info.outputSC,		// a sound converter
											fillBufferUPP,		// the callback UPP
											this,				// refCon passed to FillDataProc
											decomBuffer,		// the destination data buffer
											info.outputBytes,	// size of the destination buffer
											&actualOutputBytes,	// number of output bytes
											&outputFrames,		// number of output frames
											&outputFlags);		// FillBuffer retured advisor flags
		if (theErr != noErr)
		{	HandleError("\pI'll be damned. This is funny.", theErr, false);		break;	}
		if ((outputFlags & kSoundConverterHasLeftOverData) == false &&
			(actualOutputBytes == 0))
			state = kStateDone;
		
		// Test for norm value
	/*	if (Prefs->useAltivec)	NormTestG4(outputBuffer, actualOutputBytes/2);
		else	*/				NormTest((short*) decomBuffer, actualOutputBytes/2);
		Progress((long) (inputEOF/1024));
	}
	
	// Just for fun...
	theErr = SoundConverterEndConversion(info.outputSC, decomBuffer, &info.actualOutputFrames, &info.actualOutputSize);
	if (info.actualOutputSize > 0) NormTest((short*) decomBuffer, info.actualOutputSize/2);

	// Cleanup
	if (Prefs->useRefs)	FSCloseFork(inputRef);
	else				FSClose(input);
	normPass = false;
	PLstrcpy(theString, "\p");
	SetControlData(amountText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	SetControlData(timeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	SetControlValue(progressBar, 0);

	SoundConverterClose(info.outputSC);
	DisposePtr(decomBuffer);

	return ((float) kNormMax) / ((float) normVal / 32767.0f);
}


float Decoder::NormalizeQT(void)
{
	OSErr theErr;
	Str255	theString;
	unsigned long outputFrames, actualOutputBytes, outputFlags;

	// Init values
	normPass = true;
	normVal  = 0;

	// Setup
	InitInputQT();

	// Work waaaay different for qt normalize
	info.inputFormat = qtInfo.inputFormat;
	DisposeSoundConverterFillBufferDataUPP(fillBufferUPP);

	// Setup SoundConverter
	qtInfo.outputFormat.sampleSize	= 16;
	qtInfo.outputFormat.format		= k16BitBigEndianFormat;
	qtInfo.outputFormat.sampleRate	= info.inputFormat.sampleRate;
	qtInfo.outputFormat.numChannels	= info.inputFormat.numChannels;
	qtInfo.outputFormat.buffer		= 0L;
	qtInfo.outputFormat.sampleCount	= 0;
	qtInfo.outputFormat.reserved	= 0;
	qtInfo.outputFormat.flags 		= kNoRealtimeProcessing;

	// Setup component data
	qtInCompData.desc = qtInfo.inputFormat;
	qtInCompData.desc.buffer = (Byte *) *sourceHandle;
	qtInCompData.desc.flags = kExtendedSoundData;
	qtInCompData.recordSize = sizeof(ExtendedSoundComponentData);
	qtInCompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
	qtInCompData.bufferSize = 0;

	// Open SounderConverter (Duh)
	theErr = SoundConverterOpen(&qtInfo.inputFormat, &qtInfo.outputFormat, &qtInfo.outputSC); // Open SoundConverter (Duh)
		if (theErr != noErr) HandleError("\pCould not open Sound Converter", theErr, true);

	// Pass conversion atom to sound converter
	SoundConverterSetInfo(qtInfo.outputSC, scSoundVBRCompressionOK, Ptr(true));
	theErr = SoundConverterSetInfo(qtInfo.outputSC, siDecompressionParams, decomAtom);
	if (theErr != siUnknownInfoType && theErr != noErr)
	{	HandleError("\pProblem while setting up conversion atom", theErr, false);
	}

	// Allocate memory
	long targetBytes = 6144;
	do
	{	theErr = SoundConverterGetBufferSizes(qtInfo.outputSC, targetBytes, &qtInfo.inputBufferFrames, &qtInfo.inputBufferBytes, &qtInfo.outputBytes);
		targetBytes *= 2;
	} while (theErr == notEnoughBufferSpace && targetBytes < (MaxBlock() / 4));
	qtInfo.outputBytes = 6144;

	decomBuffer	= NewPtr(qtInfo.outputBytes);
		if (MemError()) HandleError("\pOut of Memory", MemError(), true);

	// Initialize Total Bytes to Zero
	info.totalBytesWritten = 0; 

	// Init converters
	theErr = SoundConverterBeginConversion(qtInfo.outputSC);
		if (theErr != noErr) HandleError("\pCould not begin sound conversion", theErr, true);

	// Begin
	ticksStart = TickCount();
	state = kStateDecoding;
	Feedback(true);

	while (state > kStateIdle)
	{
		actualOutputBytes = 0;
		theErr = SoundConverterFillBuffer(	qtInfo.outputSC,	// a sound converter
											qtFillBufferUPP,	// the callback UPP
											this,				// refCon passed to FillDataProc
											decomBuffer,		// the destination data buffer
											qtInfo.outputBytes,	// size of the destination buffer
											&actualOutputBytes,	// number of output bytes
											&outputFrames,		// number of output frames
											&outputFlags);		// FillBuffer retured advisor flags
		if (theErr != noErr)
		{	HandleError("\pI'll be damned. This is funny.", theErr, false);		break;	}
		if ((outputFlags & kSoundConverterHasLeftOverData) == false &&
			(actualOutputBytes == 0))
			state = kStateDone;
		
		// Test for norm value
	/*	if (Prefs->useAltivec)	NormTestG4(outputBuffer, actualOutputBytes/2);
		else	*/				NormTest((short*) decomBuffer, actualOutputBytes/2);
		Progress((long) (inputEOF/1024));
	}
	
	// Just for fun...
	theErr = SoundConverterEndConversion(qtInfo.outputSC, decomBuffer, &qtInfo.actualOutputFrames, &qtInfo.actualOutputSize);
	if (info.actualOutputSize > 0) NormTest((short*) decomBuffer, qtInfo.actualOutputSize/2);

	// Cleanup
	if (Prefs->useRefs)	FSCloseFork(inputRef);
	else				FSClose(input);
	normPass = false;
	PLstrcpy(theString, "\p");
	SetControlData(amountText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	SetControlData(timeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
	SetControlValue(progressBar, 0);

	DisposeTrackMedia(media);
	DisposeMovieTrack(track);
	DisposeMovie(movie);
	CloseMovieFile(qtRef);
	DisposeHandle(sourceHandle);
	SoundConverterClose(qtInfo.outputSC);
	DisposePtr(decomBuffer);
	
	return ((float) kNormMax) / ((float) normVal / 32767.0f);
}


float Decoder::NormalizeGlobal(void)
{
	float resultVal;
	float compareVal = 32767.0f;
	
	for (int i = 0; i < numItems && state != kStateCancel; i++)
	{	resultVal = Normalize(i);
		
		if (resultVal < compareVal)
			compareVal = resultVal;
	}

	return compareVal;
}


void Decoder::NormTest(short* data, long dataLen)
{
	short neg = -1 * normVal;
	
	for(int i=0; i<dataLen; i++)
	{	if (data[i] > normVal)
		{	normVal = data[i];
			neg = -1 * normVal;
		}
		if (data[i] < neg)
		{	neg = data[i];
			normVal = -1 * neg;
	}	}
}


/*void Decoder::NormTestG4(vector short* dataV, long dataLen)
{
	if (dataLen % 8 != 0)
		NormTest((short*) dataV, dataLen);
}*/


void Decoder::NormGrow(short* data, long dataLen)
{
	for (int i=0; i<dataLen; i++)
		data[i] = (short) (data[i] * normGrow);
}


#pragma mark -


OSErr Decoder::GetOutputFormat(SoundComponentData* theSCD, Boolean channelSplit)
{
	ComponentInstance theComponent;
	OSErr theErr;

	// Set the default output settings
	theSCD->flags = 0;
	theSCD->sampleCount = 0;
	theSCD->buffer = nil;
	theSCD->reserved = 0;

	if (Prefs->audioOutput != kAudioCustom)
	{	theSCD->format = kSoundNotCompressed;
		theSCD->numChannels = 2;
		theSCD->sampleSize = 16;
		theSCD->sampleRate = rate44khz;
	}
	
	if (Prefs->audioOutput == kAudio48khz)
		theSCD->sampleRate = rate48khz;

	if (Prefs->channelSplit)
		theSCD->numChannels = 1;

	// Custom sound settings
	if (Prefs->audioOutput == kAudioCustom)
	{
		// Create default to start
		theComponent = OpenDefaultComponent(StandardCompressionType, StandardCompressionSubTypeSound);
		if (theComponent == 0L) HandleError("\pCould not open Sound Components in GetOutputFormat", noErr, true);

		// Then get our remembered settings from prefs
		SCSetInfo(theComponent, scSettingsStateType, &Prefs->customHandle); 
		
		// Ask for the Sound Component settings the user would like to use
		theErr = SCRequestImageSettings(theComponent);
		if (theErr != noErr && theErr != userCanceledErr) HandleError("\pCould not open Sound Components dialog box", theErr, true);		
		
		if (theErr == userCanceledErr) // If user cancelled, give default settings:
		{	theSCD->format = kSoundNotCompressed;
			theSCD->numChannels = 2;
			theSCD->sampleSize = 16;
			theSCD->sampleRate = rate44khz;
			theSCD->flags |= kNoRealtimeProcessing;
			
			return userCanceledErr;
		}
		else // Then the real settings
		{	theErr = SCGetInfo(theComponent, scSoundCompressionType, &theSCD->format);
				if (theErr != noErr) HandleError("\pCould not set compression type", theErr, true);
			theErr = SCGetInfo(theComponent, scSoundChannelCountType, &theSCD->numChannels);
				if (theErr != noErr) HandleError("\pCould not set number of channels", theErr, true);
			theErr = SCGetInfo(theComponent, scSoundSampleRateType, &theSCD->sampleRate);
				if (theErr != noErr) HandleError("\pCould not set sample rate", theErr, true);
			theErr = SCGetInfo(theComponent, scSoundSampleSizeType, &theSCD->sampleSize);
				if (theErr != noErr) HandleError("\pCould not set sample size", theErr, true);

			// Store the component in preferences
			SCGetInfo(theComponent, scSettingsStateType, &Prefs->customHandle);
	}	}

	// Since compression setting of "None" causes mysterious not-on-my-end problems, we're going to cheat
	// and force the compression from 'raw ' into the nearly-identical 'NONE'.  Mwa-ha-ha-ha!
	if (theSCD->format == k8BitOffsetBinaryFormat) theSCD->format = kSoundNotCompressed;
	
	theSCD->flags |= kNoRealtimeProcessing;  // work in non-realtime, don't ask me why

	// Close it up
	CloseComponent(theComponent);
	
	return noErr;
}


OSErr Decoder::GetQuickTimeOutputFormat(SoundComponentData* theSCD)
{
	/* Fake flag means that we are not asking for settings, rather using the
	** existing SoundComponentData to generate nifty QuickTime stuff we need.
	** This allows us to save QT settings to disk.
	*/

	ComponentInstance theComponent;
	Boolean trueBoolean = true;
	OSErr theErr;

	// Set the default output settings
	/*theSCD->format = kSoundNotCompressed;
	theSCD->numChannels = 2;
	theSCD->sampleSize = 16;
	theSCD->sampleRate = rate44khz;*/
	theSCD->sampleCount = 0;
	theSCD->buffer = nil;
	theSCD->reserved = 0;
	theSCD->flags = 0;
	
	// Create default to start
	theComponent = OpenDefaultComponent(StandardCompressionType, StandardCompressionSubTypeSound);
		if (theComponent == 0L) HandleError("\pCould not open Sound Components in GetOutputFormat", noErr, true);

	// Then get our remembered settings from prefs
	SCSetInfo(theComponent, scSettingsStateType, &Prefs->movHandle); 

	// VBR friendly
	SCSetInfo(theComponent, scSoundVBRCompressionOK, &trueBoolean);
	SCSetInfo(theComponent, scSoundSampleRateChangeOK, &trueBoolean);

	// Ask for the Sound Component settings the user would like to use
	theErr = SCRequestImageSettings(theComponent);
	if (theErr != noErr && theErr != userCanceledErr) HandleError("\pCould not open Sound Components dialog box", theErr, true);		
		
	if (theErr == userCanceledErr) // If user cancelled, give default settings:
	{	theSCD->format = kSoundNotCompressed;
		theSCD->numChannels = 2;
		theSCD->sampleSize = 16;
		theSCD->sampleRate = rate44khz;
		theSCD->flags |= kNoRealtimeProcessing;
		
		return userCanceledErr;
	}
	else // Then the real settings
	{	theErr = SCGetInfo(theComponent, scSoundCompressionType, &theSCD->format);
			if (theErr != noErr) HandleError("\pCould not set compression type", theErr, true);
		theErr = SCGetInfo(theComponent, scSoundChannelCountType, &theSCD->numChannels);
			if (theErr != noErr) HandleError("\pCould not set number of channels", theErr, true);
		theErr = SCGetInfo(theComponent, scSoundSampleSizeType, &theSCD->sampleSize);
			if (theErr != noErr) HandleError("\pCould not set sample size", theErr, true);
	 	theErr = SCGetInfo(theComponent, scSoundSampleRateType, &theSCD->sampleRate);
	 		if (theErr != noErr) HandleError("\pCould not set sample rate", theErr, true);

		// Store the component in preferences
		SCGetInfo(theComponent, scSettingsStateType, &Prefs->movHandle);
	}

	/* We need to get the appropriate codec settings for the compressor
	/  these 'magic' settings need to be added to the Sound Description */
	SCGetInfo(theComponent, scCodecSettingsType, &Prefs->compressionParams);

	// Since compression setting of "None" causes mysterious not-on-my-end problems, we're going to cheat
	// and force the compression from 'raw ' into the nearly-identical 'NONE'.  Mwa-ha-ha-ha!
	if (theSCD->format == k8BitOffsetBinaryFormat) theSCD->format = kSoundNotCompressed;
	
	theSCD->flags |= kNoRealtimeProcessing;  // work in non-realtime, don't ask me why

	// Close it up
	CloseComponent(theComponent);
	
	return noErr;
}


Boolean Decoder::CheckQT(FSSpec *file)
{
	OSErr	theErr = noErr;
	Movie	theMovie = 0L;
	Track	theTrack = 0L;
	short	ref, resID = 0;
	Boolean	wasChanged;
	Boolean isQT = false;

	// Check to see if file can be read by QT
	theErr = OpenMovieFile(file, &ref, fsRdPerm);
		if (theErr != noErr) { HandleError("\pCould not open file with QT", theErr, false); return false; }
	theErr = NewMovieFromFile(&theMovie, ref, &resID, NULL, newMovieActive, &wasChanged);
	if (theErr == noErr)
		theTrack = GetMovieIndTrackType(theMovie, 1, SoundMediaType, movieTrackMediaType);
	
	// If we have an audio track, by jove, this is QuickTime
	if (theTrack != NULL) isQT = true;

	// Cleanup
	DisposeMovieTrack(theTrack);
	DisposeMovie(theMovie);
	CloseMovieFile(ref);

	return isQT;
}


void Decoder::InitInputAC3(void)
{
	// A52Lib setup
	a52_state = a52_init(0);

	// We use a callback for AC-3s now, so we have to fill in the blanks...
	fillBufferUPP = NewSoundConverterFillBufferDataUPP(AC3FillBufferProc);

	// Get the AC-3 sample rate
	(void) SeekSync(&flags, &sample_rate, &bit_rate);

	// Enter in the input settings
	info.inputFormat.flags			= 0;
	info.inputFormat.format			= k16BitBigEndianFormat;
	info.inputFormat.numChannels	= 2;
	info.inputFormat.sampleSize		= 16;
	info.inputFormat.sampleCount	= 0;
	info.inputFormat.buffer			= NULL;
	info.inputFormat.reserved		= 0;
	switch(sample_rate)
	{	case 48000:	info.inputFormat.sampleRate	= rate48khz;	break;
		case 44100:	info.inputFormat.sampleRate	= rate44khz;	break;
		case 32000: info.inputFormat.sampleRate = rate32khz;	break;
	}
	
	ac3CompData.desc = info.inputFormat;
	ac3CompData.desc.buffer = (Byte*) a52Buffer;
	ac3CompData.desc.flags = kExtendedSoundData;
	ac3CompData.recordSize = sizeof(ExtendedSoundComponentData);
	ac3CompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
	ac3CompData.bufferSize = 0;

	moreSource = true;
}


void Decoder::InitInputQT(void)
{
	OSErr	theErr = noErr;
	short	resID = 0;
	Boolean	wasChanged;
	unsigned long	targetBytes = 3072;
	
	SoundDescriptionV1Handle	description;
	Handle						extension;
	Size						size;

	// Zero stuff out
	movie = 0L;
	track = 0L;
	media = 0L;
	outputFormatIsVBR = false;

	// Create movie from input file
	theErr = OpenMovieFile(&inputSpec, &qtRef, fsRdPerm);
		if (theErr != noErr) HandleError("\pCould not open file with QT", theErr, true);
	theErr = NewMovieFromFile(&movie, qtRef, &resID, NULL, newMovieActive, &wasChanged);
		if (theErr != noErr) HandleError("\pCould create movie from file", theErr, true);
	track = GetMovieIndTrackType(movie, 1, SoundMediaType, movieTrackMediaType);
		if (track == NULL) HandleError("\pCould not get track from movie", GetMoviesError(), true);

	// get and return the sound track media
	media = GetTrackMedia(track);
		if (media == NULL) HandleError("\pCould not get track media", noErr, true);

	// get the description of the sample data
	description = (SoundDescriptionV1Handle) NewHandle(0);
	GetMediaSampleDescription(media, 1, (SampleDescriptionHandle) description);
		if (GetMoviesError()) HandleError("\pCould not get media sample description", theErr, true);

	// Set up the decompression atom
	extension = NewHandle(0);
	theErr = GetSoundDescriptionExtension((SoundDescriptionHandle) description, &extension, siDecompressionParams);
	if (theErr == noErr)
	{	size = GetHandleSize(extension);
		HLock(extension);
		decomAtom = (AudioCompressionAtom*) NewPtr(size);
		theErr = MemError();
		// copy the atom data to our buffer...
		BlockMoveData(*extension, decomAtom, size);
		HUnlock(extension);
	}
	else // if it doesn't have an atom, that's ok
		theErr = noErr;

	// Free Memory
	DisposeHandle(extension);

	// Create a sound converter for QT decoding
	qtInfo.inputFormat.flags		= 0;
	qtInfo.inputFormat.format		= (*description)->desc.dataFormat;
	qtInfo.inputFormat.numChannels	= (*description)->desc.numChannels;
	qtInfo.inputFormat.sampleSize	= (*description)->desc.sampleSize;
	qtInfo.inputFormat.sampleRate	= (*description)->desc.sampleRate;
	qtInfo.inputFormat.sampleCount= 0;
	qtInfo.inputFormat.buffer		= NULL;
	qtInfo.inputFormat.reserved	= 0;
	
	// Free more memory
	DisposeHandle((Handle) description);

	// Allocate memory for source media
	sourceHandle = NewHandle(kInputBufferSize);
		if (MemError()) HandleError("\pOut of Memory", MemError(), true);

	// Set up callbacks
	qtFillBufferUPP = NewSoundConverterFillBufferDataUPP(QTInFillBufferProc);
	fillBufferUPP	= NewSoundConverterFillBufferDataUPP(QTOutFillBufferProc);

	duration = GetMediaDuration(media);

	qtInCompData.desc = info.inputFormat;
	qtInCompData.desc.buffer = (Byte *) *sourceHandle;
	qtInCompData.desc.flags = kExtendedSoundData;
	qtInCompData.recordSize = sizeof(ExtendedSoundComponentData);
	qtInCompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
	qtInCompData.bufferSize = 0;
	time = 0;
	moreSource = true;
}


void Decoder::InitOutputAIFF(void)
{
	// This actually handles PCM and WAV as well...

	OSErr theErr = noErr;
	
	if (Prefs->audioOutput == kAudioCustom)
		info.outputFormat = Prefs->customSettings;
	else
	{	theErr = GetOutputFormat(&info.outputFormat, Prefs->channelSplit);
			if (theErr == userCanceledErr) {state = kStateIdle; return;}
	}
	
	// Compression is unsupported for WAV and PCM
	if (Prefs->wrapper == kWrapperPCM)
		info.outputFormat.format = k16BitBigEndianFormat;
	else if (Prefs->wrapper == kWrapperWAV)
		info.outputFormat.format = k16BitLittleEndianFormat;

	// Create dummy AIFF/WAV header	
	switch (Prefs->wrapper)
	{
		case kWrapperAIFF:
			theErr = 
			SetupAIFFHeader(output,	info.outputFormat.numChannels,	info.outputFormat.sampleRate,
									info.outputFormat.sampleSize,	info.outputFormat.format, 0, 0);
				if (theErr != noErr && theErr != fnOpnErr) HandleError("\pCould not set up AIFF header", theErr, true);		
			break;
			
		case kWrapperWAV:
			WAVHeader(output,	info.outputFormat.numChannels == 2, 	info.outputFormat.sampleRate,
								info.outputFormat.sampleSize,			false);
			break;
	}
	
	if (qt)
	{	// Output of QT Initial fillbuff get 16-bit PCM, same sample rate and channels
		
		// Hardcode intermediary settings
		qtInfo.outputFormat.format		= k16BitBigEndianFormat;
		qtInfo.outputFormat.sampleSize	= 16;
		qtInfo.outputFormat.numChannels	= info.outputFormat.numChannels;
		qtInfo.outputFormat.sampleRate	= info.outputFormat.sampleRate;
		qtInfo.outputFormat.flags		= kNoRealtimeProcessing;
		qtInfo.outputFormat.sampleCount	= 0;
		qtInfo.outputFormat.buffer		= 0L;
		qtInfo.outputFormat.reserved	= 0;
		
		info.inputFormat.format			= k16BitBigEndianFormat;
		info.inputFormat.sampleSize		= 16;
		info.inputFormat.numChannels	= info.outputFormat.numChannels;
		info.inputFormat.sampleRate		= info.outputFormat.sampleRate;
		info.inputFormat.flags			= 0;
		info.inputFormat.sampleCount	= 0;
		info.inputFormat.buffer			= 0L;
		info.inputFormat.reserved		= 0;

		// Setup component data
		qtInCompData.desc = qtInfo.inputFormat;
		qtInCompData.desc.buffer = (Byte *) *sourceHandle;
		qtInCompData.desc.flags = kExtendedSoundData;
		qtInCompData.recordSize = sizeof(ExtendedSoundComponentData);
		qtInCompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
		qtInCompData.bufferSize = 0;

		qtOutCompData.desc = info.inputFormat;
		qtOutCompData.desc.flags = kExtendedSoundData;
		qtOutCompData.recordSize = sizeof(ExtendedSoundComponentData);
		qtOutCompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
		qtOutCompData.bufferSize = 0;

		theErr = SoundConverterOpen(&qtInfo.inputFormat, &qtInfo.outputFormat, &qtInfo.outputSC); // Open SoundConverter (Duh)
			if (theErr != noErr) HandleError("\pCould not open Sound Converter", theErr, true);

		// Tell the sound converter we're cool with VBR formats
	//	SoundConverterSetInfo(qtInfo.outputSC, siClientAcceptsVBR, Ptr(true));															
		SoundConverterSetInfo(qtInfo.outputSC, scSoundVBRCompressionOK, Ptr(true));

		theErr = SoundConverterBeginConversion(qtInfo.outputSC);
			if (theErr != noErr) HandleError("\pCould not begin sound conversion", theErr, true);

		// Pass conversion atom to sound converter
		theErr = SoundConverterSetInfo(qtInfo.outputSC, siDecompressionParams, decomAtom);
			if (theErr != siUnknownInfoType && theErr != noErr)
			{	HandleError("\pProblem while setting up conversion atom", theErr, false);
				return;
			}
	}
	
	// Open SounderConverter (Duh)
	theErr = SoundConverterOpen(&info.inputFormat, &info.outputFormat, &info.outputSC); // Open SoundConverter (Duh)
		if (theErr != noErr) HandleError("\pCould not open Sound Converter", theErr, true);

	// Allocate memory
	if (qt)
	{	long targetBytes = 6144;
		do
		{	theErr = SoundConverterGetBufferSizes(info.outputSC, targetBytes, &info.inputBufferFrames, &info.inputBufferBytes, &info.outputBytes);
			targetBytes *= 2;
		} while (theErr == notEnoughBufferSpace && targetBytes < (MaxBlock() / 4));
	} else
	info.outputBytes = 6144;
	qtInfo.outputBytes = info.outputBytes;

	decomBuffer	= NewPtr(info.outputBytes);
	if (qt)
	{	qtOutBuffer	= NewPtr(info.outputBytes);
		qtOutCompData.desc.buffer = (Byte *) qtOutBuffer;
	}
	if (MemError()) HandleError("\pOut of Memory", MemError(), true);

	// Initialize Total Bytes to Zero
	info.totalBytesWritten = 0; 

	// Init converters
	theErr = SoundConverterBeginConversion(info.outputSC);
		if (theErr != noErr) HandleError("\pCould not begin sound conversion", theErr, true);
}



void Decoder::InitOutputMP3(MP3Info *mp3)
{
	OSErr theErr = noErr;

	gfp = lame_init();

	// Set SoundConverter output
	info.outputFormat.flags = 0;
	info.outputFormat.format = k16BitBigEndianFormat;
	info.outputFormat.numChannels = 2;
	info.outputFormat.sampleSize = 16;
	info.outputFormat.sampleCount = 0;
	info.outputFormat.buffer = nil;
	info.outputFormat.reserved = 0;

	// Resample with LAME
	if (Prefs->lameResample)
	{	info.outputFormat.sampleRate = qtInfo.inputFormat.sampleRate;

			 if (info.inputFormat.sampleRate == rate48khz)		lame_set_in_samplerate(gfp, 48000);
		else if (info.inputFormat.sampleRate == rate44khz)		lame_set_in_samplerate(gfp, 44100);
		else if (info.inputFormat.sampleRate == rate32khz)		lame_set_in_samplerate(gfp, 32000);
	//	else if (info.inputFormat.sampleRate == rate24khz)		lame_set_in_samplerate(gfp, 24000);
		else if (info.inputFormat.sampleRate == rate22050hz)	lame_set_in_samplerate(gfp, 22050);
		else if (info.inputFormat.sampleRate == rate16khz)		lame_set_in_samplerate(gfp, 16000);
		else if (info.inputFormat.sampleRate == rate11025hz)	lame_set_in_samplerate(gfp, 11025);
		else if (info.inputFormat.sampleRate == rate8khz)		lame_set_in_samplerate(gfp,  8000);
	}
	
	// Resample with QuickTime
	else
	{	switch(mp3->sample)
		{	case 48000:	info.outputFormat.sampleRate = rate48khz;	break;
			case 44100:	info.outputFormat.sampleRate = rate44khz;	break;
			case 32000:	info.outputFormat.sampleRate = rate32khz;	break;
			case 22500:	info.outputFormat.sampleRate = rate22050hz;	break;
			case 16000:	info.outputFormat.sampleRate = rate16khz;	break;
			case 11000:	info.outputFormat.sampleRate = rate11025hz;	break;		
			case 8000:	info.outputFormat.sampleRate = rate8khz;	break;
		
			default:	info.outputFormat.sampleRate = rate44khz;	mp3->sample = 44100;	break;
		}

		// Already resampled by QT
		lame_set_in_samplerate(gfp, mp3->sample);
	}

	// Set MP3 flags
	lame_set_num_channels(gfp, 2);
	lame_set_compression_ratio(gfp, 0);
	lame_set_out_samplerate(gfp, mp3->sample);
		
		 if (mp3->fast) lame_set_quality(gfp, 7);
	else if (mp3->high) lame_set_quality(gfp, 2);
	else				lame_set_quality(gfp, 5);
		
		 if (mp3->mono) 	lame_set_mode(gfp, MONO);
	else if (mp3->joint)	lame_set_mode(gfp, JOINT_STEREO);
	else					lame_set_mode(gfp, STEREO);
		
		 if (mp3->vbr)	lame_set_VBR(gfp, vbr_mtrh);
	else if (mp3->abr)	lame_set_VBR(gfp, vbr_abr);
	else				lame_set_VBR(gfp, vbr_off);
	
	if (mp3->vbr)
	{	lame_set_VBR_min_bitrate_kbps(gfp, mp3->minBitrate);
		lame_set_VBR_max_bitrate_kbps(gfp, mp3->maxBitrate);
		lame_set_VBR_q(gfp, mp3->vbrQuality);
	}
	else if (mp3->abr)
	{	lame_set_VBR_mean_bitrate_kbps(gfp, mp3->bitrate);
		lame_set_VBR_min_bitrate_kbps (gfp, mp3->minBitrate);
		lame_set_VBR_q(gfp, mp3->vbrQuality);
	}
	else
	{	lame_set_brate(gfp, Prefs->mp3info.bitrate);
	}

	if (mp3->nspsytune == true)		lame_set_exp_nspsytune(gfp, 1);
	else							lame_set_exp_nspsytune(gfp, 0);

	theErr = lame_init_params(gfp);
		if (theErr != noErr) HandleError("\pMP3 encoding failed; you may have chosen an invalid bitrate.", theErr, true);

	if (qt)
	{	// Output of QT Initial fillbuff get 16-bit PCM, same sample rate and channels
		
		// Hardcode intermediary settings
		qtInfo.outputFormat.format		= k16BitBigEndianFormat;
		qtInfo.outputFormat.sampleSize	= 16;
		qtInfo.outputFormat.numChannels	= info.outputFormat.numChannels;
		qtInfo.outputFormat.sampleRate	= info.outputFormat.sampleRate;
		qtInfo.outputFormat.flags		= kNoRealtimeProcessing;
		qtInfo.outputFormat.sampleCount	= 0;
		qtInfo.outputFormat.buffer		= 0L;
		qtInfo.outputFormat.reserved	= 0;
		
		info.inputFormat.format			= k16BitBigEndianFormat;
		info.inputFormat.sampleSize		= 16;
		info.inputFormat.numChannels	= info.outputFormat.numChannels;
		info.inputFormat.sampleRate		= info.outputFormat.sampleRate;
		info.inputFormat.flags			= 0;
		info.inputFormat.sampleCount	= 0;
		info.inputFormat.buffer			= 0L;
		info.inputFormat.reserved		= 0;

		// Setup component data
		qtInCompData.desc = qtInfo.inputFormat;
		qtInCompData.desc.buffer = (Byte *) *sourceHandle;
		qtInCompData.desc.flags = kExtendedSoundData;
		qtInCompData.recordSize = sizeof(ExtendedSoundComponentData);
		qtInCompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
		qtInCompData.bufferSize = 0;

		qtOutCompData.desc = info.inputFormat;
		qtOutCompData.desc.flags = kExtendedSoundData;
		qtOutCompData.recordSize = sizeof(ExtendedSoundComponentData);
		qtOutCompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
		qtOutCompData.bufferSize = 0;

		theErr = SoundConverterOpen(&qtInfo.inputFormat, &qtInfo.outputFormat, &qtInfo.outputSC); // Open SoundConverter (Duh)
			if (theErr != noErr) HandleError("\pCould not open Sound Converter", theErr, true);

		// Tell the sound converter we're cool with VBR formats
	//	SoundConverterSetInfo(qtInfo.outputSC, siClientAcceptsVBR, Ptr(true));															
		SoundConverterSetInfo(qtInfo.outputSC, scSoundVBRCompressionOK, Ptr(true));

		theErr = SoundConverterBeginConversion(qtInfo.outputSC);
			if (theErr != noErr) HandleError("\pCould not begin sound conversion", theErr, true);

		// Pass conversion atom to sound converter
		theErr = SoundConverterSetInfo(qtInfo.outputSC, siDecompressionParams, decomAtom);
			if (theErr != siUnknownInfoType && theErr != noErr)
			{	HandleError("\pProblem while setting up conversion atom", theErr, false);
				return;
			}
	}
	
	// Open SounderConverter (Duh)
	theErr = SoundConverterOpen(&info.inputFormat, &info.outputFormat, &info.outputSC); // Open SoundConverter (Duh)
		if (theErr != noErr) HandleError("\pCould not open Sound Converter", theErr, true);

	// Allocate memory
	if (qt)
	{	long targetBytes = 6144;
		do
		{	theErr = SoundConverterGetBufferSizes(info.outputSC, targetBytes, &info.inputBufferFrames, &info.inputBufferBytes, &info.outputBytes);
			targetBytes *= 2;
		} while (theErr == notEnoughBufferSpace && targetBytes < (MaxBlock() / 4));
	}
	else info.outputBytes = 6144;
	qtInfo.outputBytes = info.outputBytes;

	decomBuffer	= NewPtr(info.outputBytes);
	if (qt)
	{	qtOutBuffer	= NewPtr(info.outputBytes);
		qtOutCompData.desc.buffer = (Byte *) qtOutBuffer;
	}
	if (MemError()) HandleError("\pOut of Memory", MemError(), true);

	// Initialize Total Bytes to Zero
	info.totalBytesWritten = 0; 

	// Init converters
	theErr = SoundConverterBeginConversion(info.outputSC);
		if (theErr != noErr) HandleError("\pCould not begin sound conversion", theErr, true);
}


void Decoder::InitOutputMOV(void)
{
	OSErr theErr = noErr;	

	/* OUTPUT FORMAT */
	info.outputFormat = Prefs->movSettings;


	// QuickTime input weird stuff
	if (qt)
	{	// Output of QT Initial fillbuff get 16-bit PCM, same sample rate and channels
		
		// Hardcode intermediary settings
		qtInfo.outputFormat.format		= k16BitBigEndianFormat;
		qtInfo.outputFormat.sampleSize	= 16;
		qtInfo.outputFormat.numChannels	= info.outputFormat.numChannels;
		qtInfo.outputFormat.sampleRate	= info.outputFormat.sampleRate;
		qtInfo.outputFormat.flags		= kNoRealtimeProcessing;
		qtInfo.outputFormat.sampleCount	= 0;
		qtInfo.outputFormat.buffer		= 0L;
		qtInfo.outputFormat.reserved	= 0;
		
		info.inputFormat.format			= k16BitBigEndianFormat;
		info.inputFormat.sampleSize		= 16;
		info.inputFormat.numChannels	= info.outputFormat.numChannels;
		info.inputFormat.sampleRate		= info.outputFormat.sampleRate;
		info.inputFormat.flags			= 0;
		info.inputFormat.sampleCount	= 0;
		info.inputFormat.buffer			= 0L;
		info.inputFormat.reserved		= 0;

		// Setup component data
		qtInCompData.desc = qtInfo.inputFormat;
		qtInCompData.desc.buffer = (Byte *) *sourceHandle;
		qtInCompData.desc.flags = kExtendedSoundData;
		qtInCompData.recordSize = sizeof(ExtendedSoundComponentData);
		qtInCompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
		qtInCompData.bufferSize = 0;

		qtOutCompData.desc = info.inputFormat;
		qtOutCompData.desc.flags = kExtendedSoundData;
		qtOutCompData.recordSize = sizeof(ExtendedSoundComponentData);
		qtOutCompData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
		qtOutCompData.bufferSize = 0;

		theErr = SoundConverterOpen(&qtInfo.inputFormat, &qtInfo.outputFormat, &qtInfo.outputSC); // Open SoundConverter (Duh)
			if (theErr != noErr) HandleError("\pCould not open Sound Converter", theErr, true);
		theErr = SoundConverterBeginConversion(qtInfo.outputSC);
			if (theErr != noErr) HandleError("\pCould not begin sound conversion", theErr, true);

		// Pass conversion atom to sound converter
		theErr = SoundConverterSetInfo(qtInfo.outputSC, siDecompressionParams, decomAtom);
			if (theErr != siUnknownInfoType && theErr != noErr)
			{	HandleError("\pProblem while setting up conversion atom", theErr, false);
				return;
			}
	}

	// Open SoundConverter (Duh)
	theErr = SoundConverterOpen(&info.inputFormat, &info.outputFormat, &info.outputSC); 
		if (theErr != noErr) HandleError("\pCould not open Sound Converter", theErr, true);

	// Tell the sound converter we're cool with VBR formats
//	SoundConverterSetInfo(info.outputSC, siClientAcceptsVBR, Ptr(true));															
	SoundConverterSetInfo(info.outputSC, scSoundVBRCompressionOK, Ptr(true));

	// Set up the sound converters compression environment
	// Pass down siCompressionSampleRate, siCompressionChannels then siCompressionParams
	SoundConverterSetInfo(info.outputSC, siCompressionSampleRate, &info.outputFormat.sampleRate); // ignore errors
	SoundConverterSetInfo(info.outputSC, siCompressionChannels,   &info.outputFormat.numChannels);

	/*	Set up the compression environment by passing in the 'magic' compression params aquired from
		standard sound compression eariler. */
	if (Prefs->compressionParams)
	{	HLock(Prefs->compressionParams);
		theErr = SoundConverterSetInfo(info.outputSC, siCompressionParams, *Prefs->compressionParams);
			if (theErr != noErr) HandleError("\pCould not set compression parameters", theErr, false);
		HUnlock(Prefs->compressionParams);
	}
	
	// And so it begins
	theErr = SoundConverterBeginConversion(info.outputSC);
		if (theErr != noErr) HandleError("\pCould not begin conversion", theErr, false);


	/* NEW MOVIE */

	// New Movie - I'll find a better way to do this later
	/* ERROR */
	theErr = FSClose(output);
		if (theErr != noErr) HandleError("\pCould not delete... umm... stuff", theErr, false);
	theErr = CreateMovieFile(&outputSpec, 'TVOD', smSystemScript,
							  createMovieFileDeleteCurFile | createMovieFileDontCreateResFile, &output, &outMovie);
		if (theErr != noErr) HandleError("\pCould not create output movie", theErr, false);

	// New Track
	outTrack = NewMovieTrack(outMovie, 0, 0, kFullVolume);
	theErr = GetMoviesError();
		if (theErr != noErr) HandleError("\pCould not create track", theErr, false);
		
	// it's a sound track so add a SoundMedia
	// this is the sample rate we got back from our attempt to use siCompressionOutputSampleRate above
	/* ERROR */
	outMedia = NewTrackMedia(outTrack, SoundMediaType, info.outputFormat.sampleRate >> 16, NULL, 0);
	theErr = GetMoviesError();
		if (theErr != noErr) HandleError("\pCould not create media", theErr, false);

	// start media editing session
	theErr = BeginMediaEdits(outMedia);
		if (theErr != noErr) HandleError("\pCould not start media editing", theErr, false);
	
	// we need to get info about data/frame sizes 
	theErr = SoundConverterGetInfo(info.outputSC, siCompressionFactor, &compressionFactor);				
		if (theErr != noErr) HandleError("\pCould not set up compression factor", theErr, false);
		
	hSoundDescription = (SoundDescriptionV1Handle) NewHandleClear(sizeof(SoundDescriptionV1));	
		theErr = MemError();
		if (theErr != noErr) HandleError("\pCould not get memory for sound description handle", theErr, true);
	HLock((Handle)hSoundDescription);

	// Add all the details into the sound description
	(*hSoundDescription)->desc.descSize		 = sizeof(SoundDescriptionV1);
	(*hSoundDescription)->desc.dataFormat	 = (long)info.outputFormat.format;	   // compression format
	(*hSoundDescription)->desc.resvd1		 = 0;								   // must be 0
	(*hSoundDescription)->desc.resvd2		 = 0;							       // must be 0
	(*hSoundDescription)->desc.dataRefIndex	 = 0;								   // 0 - we'll let AddMediaXXX determine the index
	(*hSoundDescription)->desc.version		 = 1;								   // set to 1
	(*hSoundDescription)->desc.revlevel		 = 0;								   // set to 0
	(*hSoundDescription)->desc.vendor		 = 0;
	(*hSoundDescription)->desc.numChannels	 = info.outputFormat.numChannels;	   // number of channels
	(*hSoundDescription)->desc.sampleSize	 = info.outputFormat.sampleSize; 	   // bits per sample - everything but 8 bit can be set to 16
	(*hSoundDescription)->desc.compressionID = compressionFactor.compressionID;    // the compression ID (eg. variableCompression)
	(*hSoundDescription)->desc.packetSize	 = 0;								   // set to 0
	//(*hSoundDescription)->desc.sampleRate	 = theOutputSampleRate;		   		   // the sample rate
	(*hSoundDescription)->desc.sampleRate	 = info.outputFormat.sampleRate;	   // the sample rate
	// version 1 stuff
	(*hSoundDescription)->samplesPerPacket 	 = compressionFactor.samplesPerPacket; // the samples per packet holds the PCM sample count per audio frame (packet)
	(*hSoundDescription)->bytesPerPacket 	 = compressionFactor.bytesPerPacket;   // the bytes per packet

	// bytesPerFrame isn't necessarily calculated for us and returned as part of the CompressionFactor - not all codecs that
	// implement siCompressionFactor fill out bytesPerFrame - so we do it here - note that VBR doesn't deserve this treatment
	// but it's not harmful, the Sound Manager would do calculations itself as part of GetCompressionInfo()
	// It should be noted that GetCompressionInfo() doesn't work for codecs that need configuration with 'magic' settings.
	// This requires explicit opening of the codec and the siCompressionFactor selector for SoundComponentGetInfo()
	(*hSoundDescription)->bytesPerFrame 	 = compressionFactor.bytesPerPacket * info.outputFormat.numChannels;
	(*hSoundDescription)->bytesPerSample 	 = compressionFactor.bytesPerSample;							

    // the theCompressionParams are not necessarily present
	if (Prefs->compressionParams)
	{	// a Sound Description can't be locked when calling AddSoundDescriptionExtension so make sure it's unlocked
		HUnlock((Handle) hSoundDescription);
		theErr = AddSoundDescriptionExtension((SoundDescriptionHandle) hSoundDescription, Prefs->compressionParams, siDecompressionParams);	
			if (theErr != noErr) HandleError("\pCould not add sound description", theErr, false);
		HLock((Handle) hSoundDescription);
	}
        
	// VBR implies a different media layout, this will affect how AddMediaSample() is called below
    outputFormatIsVBR = ((*hSoundDescription)->desc.compressionID == variableCompression);


	/* BUFFERS AND SUCH */
	
	// Allocate memory
	long targetBytes = 6144;
	// FIXME
	/*do
	{	theErr = SoundConverterGetBufferSizes(info.outputSC, targetBytes, &info.inputBufferFrames, &info.inputBufferBytes, &info.outputBytes);
		targetBytes *= 2;
	} while (theErr == notEnoughBufferSpace && targetBytes < (MaxBlock() / 4));
*/	qtInfo.outputBytes = info.outputBytes = targetBytes;
	decomBuffer	= NewPtr(info.outputBytes);
	if (qt)
	{	qtOutBuffer	= NewPtr(info.outputBytes);
		qtOutCompData.desc.buffer = (Byte *) qtOutBuffer;
	}
	if (MemError()) HandleError("\pOut of Memory", MemError(), true);

	// Initialize Total Bytes to Zero
	info.totalBytesWritten = 0;


	/* SC FILLBUFFER CALLBACK SETUP */
		
	// if the source is VBR it means we're going to set the kExtendedSoundCommonFrameSizeValid
	// flag and use the commonFrameSize field in the FillBuffer callback
	//outputFormatIsVBR = (info.inputFormat.compressionID == variableCompression);
	
	// Stupid.
	#define kExtendedSoundCommonFrameSizeValid 1L << 3

	if (outputFormatIsVBR)
	{
		if (qt)	 qtOutCompData.extendedFlags |= kExtendedSoundCommonFrameSizeValid;
		else	 ac3CompData.extendedFlags |= kExtendedSoundCommonFrameSizeValid;
	}
}


void Decoder::WAVHeader(short theFile, Boolean stereo, long sampleRate, short sampleSize, Boolean real)
{
	OSErr theErr = noErr;
	long bytesWritten, fileLength;
	unsigned long bytesPerSec;
	unsigned short bytesPerSample, bitsPerSample;
	char buffer[255];
	char reverseBuf[4];
	
	// First identify file as RIFF/WAV
	bytesWritten = 12;
	strcpy(buffer, "RIFF....WAVE");
	if (real)
	{	GetEOF(theFile, &fileLength);
		fileLength -= 8;
		BlockMoveData(&fileLength, reverseBuf, sizeof(long));
		buffer[4] = reverseBuf[3];
		buffer[5] = reverseBuf[2];
		buffer[6] = reverseBuf[1];
		buffer[7] = reverseBuf[0];
	} else
	{	buffer[4] = 0;
		buffer[5] = 0;
		buffer[6] = 0;
		buffer[7] = 0;
	}
	if (noErr != FSWrite(theFile, &bytesWritten, buffer))
		HandleError("\pCould not write to WAV header... blah.", 0, false);
	
	// Write the format chunk
	strcpy(buffer, "fmt ....................");
	buffer[4] = 16;
	buffer[5] = 0;
	buffer[6] = 0;
	buffer[7] = 0;
	
	// Format [8-9]
	buffer[8] = 1;
	buffer[9] = 0;
	
	// Write number of channels [10-11]
	if (stereo)		buffer[10] = 2;
	else			buffer[10] = 1;
	buffer[11] = 0;
	
	// Write sample rate [12-15]
	BlockMoveData(&sampleRate, reverseBuf, sizeof(sampleRate));
	buffer[12] = reverseBuf[1];
	buffer[13] = reverseBuf[0];
	BlockMoveData(reverseBuf, &sampleRate, sizeof(short));
	buffer[14] = 0;
	buffer[15] = 0;
	
	// Bytes/Sample [20-21]
	if (stereo == false && sampleSize == 8)		bytesPerSample = 1;
	else if (stereo && sampleSize == 16)		bytesPerSample = 4;
	else										bytesPerSample = 2;
	buffer[20] = bytesPerSample;
	buffer[21] = 0;
	
	// Bytes/Second [16-19]
	bytesPerSec = bytesPerSample * sampleRate;
	BlockMoveData(&bytesPerSec, &reverseBuf, sizeof(bytesPerSec));
	buffer[16] = reverseBuf[3];
	buffer[17] = reverseBuf[2];
	buffer[18] = reverseBuf[1];
	buffer[19] = reverseBuf[0];
	
	// Bits/Sample [22-23]
	bitsPerSample = sampleSize;
	buffer[22] = bitsPerSample;
	buffer[23] = 0;
	
	bytesWritten = 24;
	FSWrite(theFile, &bytesWritten, buffer);
	
	// write the data chunk header
	strcpy(buffer, "data....");
	if (real)
	{	GetEOF(theFile, &fileLength);
		fileLength -= 36;
		BlockMoveData(&fileLength, reverseBuf, sizeof(long));
		buffer[4] = reverseBuf[3];
		buffer[5] = reverseBuf[2];
		buffer[6] = reverseBuf[1];
		buffer[7] = reverseBuf[0];
	}
	else
	{	buffer[4] = 0;
		buffer[5] = 0;
		buffer[6] = 0;
		buffer[7] = 0;
	}

	bytesWritten = 8;
	FSWrite(theFile, &bytesWritten, buffer);
}


#pragma mark -


int Decoder::SeekSync(int* fl, int* sr, int* br)
{
	unsigned char*	start;
	int				length = 0;

	do
	{	// Look for possible header "0xbb"
		do
		{	FillBuffer(&start, 1);
		} while(start[0] != 0x0b && state > kStateIdle);
	
		// Move back one byte
		if (Prefs->useRefs)	FSSetForkPosition(inputRef, fsFromMark, -1);
		else				SetFPos(input, fsFromMark, -1);

		// Test for syncframe
		FillBuffer(&start, 7);
		length = a52_syncinfo(start, fl, sr, br);
	
		// Move back six bytes
		if (Prefs->useRefs)	FSSetForkPosition(inputRef, fsFromMark, -6);
		else				SetFPos(input, fsFromMark, -6);

	} while(length <= 0 && state > kStateIdle);

	// Move back one byte
	if (Prefs->useRefs)	FSSetForkPosition(inputRef, fsFromMark, -1);
	else				SetFPos(input, fsFromMark, -1);
	
	return length;
}



void Decoder::WriteFrame(unsigned char* frame, long frameSize)
{
	OSErr theErr = noErr;

	// Encode to MP3 and write encoded data
	if (Prefs->wrapper == kWrapperMP3)
	{
		bytesWritten = lame_encode_buffer_interleaved(gfp, (short*) frame, frameSize/4, (unsigned char*) outputBuffer, kOutputBufferSize);
			if (bytesWritten < 0) HandleError("\pLAME Encoding Error", bytesWritten, true);
		theErr = FSWrite(output, &bytesWritten, outputBuffer);
			if (theErr != noErr) HandleError("\pCould not write to output", theErr, true);
		aiffEOF += bytesWritten;
		Progress((long) (aiffEOF/1024));
		return;
	}
	
	// Write to QuickTime file
	else if (Prefs->wrapper == kWrapperMOV)
	{
		TimeValue durationPerMediaSample;
		
		if (outputFormatIsVBR)
		{	numberOfMediaSamples = info.actualOutputFrames;
			durationPerMediaSample = compressionFactor.samplesPerPacket;
		} else
		{	numberOfMediaSamples = info.actualOutputFrames * compressionFactor.samplesPerPacket;
			durationPerMediaSample = 1;
		}
			
		theErr = AddMediaSample(outMedia,
								(char**) &frame,
								0,
								frameSize,
								durationPerMediaSample,
								(SampleDescriptionHandle) hSoundDescription,
								(long) numberOfMediaSamples,
								0, NULL);
			if (theErr != noErr) HandleError("\pCould not add media sample", theErr, true);

		aiffEOF += frameSize;
		Progress((long) (aiffEOF/1024));
		return;
	}

	// Write Output Samples
	bytesWritten = frameSize;
	theErr = FSWrite(output, &bytesWritten, frame);

	if (theErr != noErr)
		HandleError("\pCould not write to output", theErr, true);
	aiffEOF += bytesWritten;
	info.totalBytesWritten += (unsigned long) bytesWritten;
	Progress((long) (aiffEOF/1024));

	// Check to see if we're nearing the over 2G range, and create a new file if we are
	if (aiffEOF > kMaxOutputSize)
	{	if (Prefs->wrapper == kWrapperAIFF && !Prefs->channelSplit)
		{	theErr = (SetFPos(output, fsFromStart, 0));
				if (theErr != noErr) HandleError("\pSetFPos failed for AIFF Header", theErr, true);
			theErr = SetupAIFFHeader(output,	info.outputFormat.numChannels,	info.outputFormat.sampleRate,
													info.outputFormat.sampleSize,	info.outputFormat.format,
					info.totalBytesWritten, 	info.totalBytesWritten / info.outputFormat.sampleSize * info.outputFormat.numChannels);
				if (theErr != noErr && theErr != fnOpnErr) HandleError("\pCould not set up AIFF header", theErr, true);
		} FSClose(output);
		Str255 theString;
		PLstrcpy(theString, outputSpec.name);
		GetOutputFile(theString);
		theErr = FSpOpenDF(&outputSpec, fsWrPerm, &output);
			if (theErr != noErr) HandleError("\pCould not open data fork of new output file", theErr, true);
		info.totalBytesWritten = 0; aiffEOF = 0;
		if (Prefs->wrapper == kWrapperAIFF && !Prefs->channelSplit == false)
		{	theErr = SetupAIFFHeader(output,	info.outputFormat.numChannels,	info.outputFormat.sampleRate,
												info.outputFormat.sampleSize,	info.outputFormat.format,	0, 0);
				if (theErr != 0) HandleError("\pCould not set up dummy AIFF header", theErr, true);
			SetFPos(output, fsFromLEOF, 0);
			GetEOF(output, &aiffEOF);
	}	}
}



void Decoder::Progress(long feedbackSize)
{
	Str255	theString;
	Str31	theString2;

	// Update window
	if (TickCount() - ticksCurrent > kUpdateFrequency)
	{	NumToString(feedbackSize, theString);
 		PLstrcat(theString, "\p K");
		SetControlData(amountText, kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
		if (qt)
		{	float val = (float) time;
			amountDone = (long) ((val / duration) * kProgressBarMax);
		} else
			amountDone = (long) (inputEOF / inputFileSize);

		// Set progess bar value
		if (Prefs->useProgressBar)
			SetControlValue(progressBar, amountDone); 

		// Use percent (progress bar disabled)
		else
		{	long percVal = (amountDone * 100 / kProgressBarMax);
			long ptVal = ((amountDone * 1000) / kProgressBarMax) - (percVal*10);
			if (percVal > 100)
			{	percVal = 100;	ptVal = 100;	}
			PLstrcpy(theString, titleText);
			PLstrcat(theString, "\p (");
			NumToString(percVal, theString2);
			PLstrcat(theString, theString2);
			PLstrcat(theString, "\p.");
			NumToString(ptVal, theString2);
			PLstrcat(theString, theString2);
			PLstrcat(theString, "\p%)");
			SetControlData(decodeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);
		}

		// Print time remaining for decode
		ticksCurrent = TickCount();
		Time(ticksCurrent - ticksStart, amountDone);

		if (WaitNextEvent(everyEvent, &Event, 1, nil))
			HandleEvent(&Event);
	}

#ifdef CPU_TEST
	interfaceTime += TickCount() - interfaceTicks;
#endif				
}


void Decoder::Cleanup(void)
{
	OSErr theErr;
	Str31 theString;

	// QuickTime Output Cleanup
	if (Prefs->wrapper == kWrapperMOV)
	{
		theErr = EndMediaEdits(outMedia);
			IfError("\pProblem ending media edit", theErr);
		
		theErr = InsertMediaIntoTrack(outTrack, 0, 0, GetMediaDuration(outMedia), fixed1);
			IfError("\pProblem inserting media into track", theErr);

		theErr = AddMovieResource(outMovie, output, NULL, NULL);
			IfError("\pCould not add movie resource", theErr);	

		// Close stuff
		DisposeTrackMedia(outMedia);
		DisposeMovieTrack(outTrack);
		DisposeMovie(outMovie);
		DisposeHandle((Handle) hSoundDescription);
	}

	// MP3 Cleanup
	else if (Prefs->wrapper == kWrapperMP3)
	{	// Flush and write MP3 buffers
		bytesWritten = lame_encode_flush(gfp, (unsigned char*) outputBuffer, kOutputBufferSize);
			 if (bytesWritten < 0) HandleError("\pMP3 encoding error on last frame", bytesWritten, true);
		else if (bytesWritten > 0) theErr = FSWrite(output, &bytesWritten, outputBuffer);
		
		// Write final VBR tag
		if (lame_get_bWriteVbrTag(gfp) == 1)
		{	FILE *thefile = NULL;
			char *str = (char*) vec_malloc(2048);
			if (str != NULL)
			{	FSClose(output);
				FSpGetFullPath(&outputSpec, (StringPtr) str);
				thefile = fopen((char*) &str[1], "r+");
				if (thefile != NULL)
					lame_mp3_tags_fid(gfp, thefile);
				fclose(thefile);
			}
			vec_free(str);
		}
		
		// Done!
		lame_close(gfp);
	}

	// Close the Sound Converter
	theErr = SoundConverterClose(info.outputSC);
		if (theErr != 0) HandleError("\pSoundConverterClose failed", theErr, false);
	if (qt)
	{	theErr = SoundConverterClose(qtInfo.outputSC);
			if (theErr != 0) HandleError("\pNorm SoundConverterClose failed", theErr, false);	
	}
	
	// Create the real AIFF Header now that we are done; set file position to the beginning first
	if (Prefs->wrapper == kWrapperAIFF)
	{	theErr = (SetFPos(output, fsFromStart, 0));
			if (theErr != noErr) HandleError("\pSetFPos failed for AIFF Header", theErr, true);
		theErr = SetupAIFFHeader(output,	info.outputFormat.numChannels,	info.outputFormat.sampleRate,
											info.outputFormat.sampleSize,	info.outputFormat.format,
											info.totalBytesWritten, 		info.totalBytesWritten / info.outputFormat.sampleSize * info.outputFormat.numChannels);
			if (theErr != noErr && theErr != fnOpnErr) HandleError("\pCould not set up AIFF header", theErr, true);

	// WAV Header
	} else if (Prefs->wrapper == kWrapperWAV)
	{	theErr = (SetFPos(output, fsFromStart, 0));
			if (theErr != noErr) HandleError("\pSetFPos failed for WAV Header", theErr, true);
		WAVHeader(output,	info.outputFormat.numChannels == 2, 	info.outputFormat.sampleRate,
							info.outputFormat.sampleSize,			true);
	}

	// Cleanup QuickTime
	if (qt)
	{	DisposeTrackMedia(media);
		DisposeMovieTrack(track);
		DisposeMovie(movie);
		CloseMovieFile(qtRef);
		DisposeHandle(sourceHandle);
		DisposePtr((Ptr) decomAtom);
		DisposePtr(qtOutBuffer);
		DisposeSoundConverterFillBufferDataUPP(qtFillBufferUPP);
	}

	// Cleanup A52Lib
	else
	{	a52_free(a52_state);
	}

	// Dispose Memory
	DisposePtr(decomBuffer);
	DisposeSoundConverterFillBufferDataUPP(fillBufferUPP);

	// Close files
	if (Prefs->useRefs)	FSCloseFork(inputRef);
	else				FSClose(input);
	FSClose(output);

	// Flush the volume
	FlushVol(NULL, outputSpec.vRefNum);

	// Interface Cleanup
	PLstrcpy(theString, "\p");
	SetControlData(amountText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(detailsText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(timeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(decodeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlValue(progressBar, 0);
	DrawControls(window);
}

#pragma mark -

inline char SplitGetChannelMask(long channels, Boolean LFE)
{
	char mask = 0;
	
	switch (channels)
	{
		case A52_MONO:
		case A52_CHANNEL:
		case A52_CHANNEL1:
		case A52_CHANNEL2:
			mask = 1 << kCenter;
			break;
		
		case A52_STEREO:
		case A52_DOLBY:
			mask = 1 << kLeft
				 | 1 << kRight;
			break;
		
		case A52_3F:
			mask = 1 << kLeft
				 | 1 << kCenter
				 | 1 << kRight;
			break;
		
		case A52_2F1R:
			mask = 1 << kLeft
				 | 1 << kRight
				 | 1 << kLeftS;
			break;

		case A52_3F1R:
			mask = 1 << kLeft
				 | 1 << kCenter
				 | 1 << kRight
				 | 1 << kLeftS;
			break;

		case A52_2F2R:
			mask = 1 << kLeft
				 | 1 << kRight
				 | 1 << kLeftS
				 | 1 << kRightS;
			break;

		case A52_3F2R:
			mask = 1 << kLeft
				 | 1 << kCenter
				 | 1 << kRight
				 | 1 << kLeftS
				 | 1 << kRightS;
			break;
	}
	
	if (LFE) mask |= 1 << kLFE;
	
	return mask;
}


void Decoder::PrepChannel(int channel, int sampleRate)
{
	OSErr theErr;
	Str255 theString;
	long theSR;
	
	switch(sampleRate)
	{	case 32000:	theSR = rate32khz;
		case 44100: theSR = rate44khz;
		case 48000: theSR = rate48khz;
	}
	
	// Find the output file
	PLstrcpy(theString, inputSpec.name);
	GetName(theString, 10+channel);
	GetName(theString, Prefs->wrapper);
	GetOutputFile(theString);
	
	// Open it up
	split.fs[channel] = outputSpec;
	theErr = FSpOpenDF(&split.fs[channel], fsWrPerm, &split.ref[channel]);
		IfError("\pCould not open output channel file", theErr);
	
	// Add any headers we need
	if (Prefs->wrapper == kWrapperAIFF)
	{	theErr = SetupAIFFHeader(split.ref[channel], 1, theSR, 16, 'twos', 0, 0);
			IfError("\pCould not set up AIFF header", theErr);
	}
	else if (Prefs->wrapper == kWrapperWAV)
	{	WAVHeader(split.ref[channel], false, theSR, 16, false);
	}
}


void Decoder::SplitAC3(void)
{
	OSErr			theErr;
	int				length = 0;
	sample_t		level = 32767.0f;
	unsigned char*	start;
	float*			fltPtr;
	char*			shortBuf;
	short*			shortPtr;
	char			channelMask, numChannels = 0;

	// Yay Interface!
	Str255 theString;
	PLstrcpy(theString, "\pSplitting ");
	PLstrcat(theString, inputSpec.name);
	SetControlData(decodeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1);

	// File I/O
	for (int i=0; i<6; i++)
		split.ref[i] = 0;
		
	// Get work buffer
	shortBuf = NewPtr(256*2);

	// A52Lib setup
	a52_state = a52_init(0);

	// Start decoding
	ticksStart = TickCount();
	while (state > kStateIdle)
	{
		// Seek to sync word and parse sync info
		length = SeekSync(&flags, &sample_rate, &bit_rate);

		// Decode frame
		flags |= A52_ADJUST_LEVEL;
		FillBuffer(&start, length); // Meesa like code re-use!
		a52_frame(a52_state, start, &flags, &level, 0);
	
		// Dynamic Range Compression
		if (Prefs->a52nodynrng)
			a52_dynrng(a52_state, NULL, NULL);
		
		// Do per-channel file stuff
		channelMask = SplitGetChannelMask(flags & A52_CHANNEL_MASK, flags & A52_LFE);
		for (int i=0; i<6; i++)
		{	if (split.ref[i] == 0 && channelMask & (1 << i))
				PrepChannel(i, sample_rate);
		}
		
		// Get number of channels
		numChannels = 0;
		for (int i=0; i<6; i++)
		{	if (channelMask & (1 << i))
				numChannels++;
		}
		
		// Get samples
		for (int i=0; i<6; i++)
		{
			a52_block(a52_state);
		
			// Get each channel
			for (int j=0; j<numChannels; j++)
			{
				shortPtr = (short*) shortBuf;		
				fltPtr = a52_samples(a52_state);
	
				// Convert to short
				for (int k=0; k<256; k++)
					shortPtr[k] = (short) (fltPtr[k+256*j]);
				
				// Write buffer data to channel file
				SplitWrite(shortBuf, j, channelMask);
			}
		}
		
		Progress((long) (inputEOF/1024));
	}
	
	// Do cleanup
	a52_free(a52_state);
	DisposePtr(shortBuf);

	// Write final AIFF/WAV headers
	for (int i=0; i<6; i++)
	{
		if (split.ref[i] == 0)
			continue;
	
		int theSR;	
		switch(sample_rate)
		{	case 32000:	theSR = rate32khz;
			case 44100: theSR = rate44khz;
			case 48000: theSR = rate48khz;
		}

		if (Prefs->wrapper == kWrapperAIFF)
		{	long bytes;
			GetEOF(split.ref[i], &bytes);
			bytes -= kSizeOfAIFFHeader;
			theErr = (SetFPos(split.ref[i], fsFromStart, 0));
				IfError("\pSetFPos failed for AIFF Header", theErr);
			theErr = SetupAIFFHeader(split.ref[i], 1, theSR, 16, 'twos', bytes, bytes/16);
				IfError("\pCould not set up AIFF header", theErr);
		}
		else if (Prefs->wrapper == kWrapperWAV)
		{	theErr = (SetFPos(split.ref[i], fsFromStart, 0));
				IfError("\pSetFPos failed for WAV Header", theErr);
			WAVHeader(split.ref[i], false, theSR, 16, true);
		}
	}

	// Close Files
	for(int i=0; i<6; i++)
	{	if (split.ref[i] != 0)
			FSClose(split.ref[i]);
	}
	if (Prefs->useRefs)	FSCloseFork(inputRef);
	else				FSClose(input);

	// Interface Cleanup
	PLstrcpy(theString, "\p");
	SetControlData(amountText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(detailsText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(timeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlData(decodeText,	kControlEntireControl, 'text', PLstrlen(theString), (theString)+1); 	
	SetControlValue(progressBar, 0);
	DrawControls(window);
}


void Decoder::SplitWrite(char* data, int channel, char channelMask)
{
	OSErr theErr;
	long bytesWritten = 256 * 2;
	char channelNum;
	
	// Find the channel we need
	for (int i=0, c=-1; i<6; i++)
	{	if (channelMask & (1 << i))
			c++; // Tee-hee
		if (c == channel)
		{	channelNum = i;
			break;
		}
	}
	
	theErr = FSWrite(split.ref[channelNum], &bytesWritten, data);
		IfErrorFatal("\pCould not write channel data to file", theErr);
}